import h5py as h5
import sys

def decode(b_array):
    result = []
    for value in b_array:
        result.append(value.decode("utf-8"))
    return result


class PdrHDF(object):
    """
      Access to a pdr HDF5 file
    """
    def __init__(self, filename):
        self.chemistry_path = '/Local quantities/Chemistry'
        self.conditions_path = '/Local quantities/Gas conditions'
        self.position_path = '/Local quantities/Positions'
        self.masses_path = '/Parameters/Species mass'
        self.network_path = '/Parameters/Network'

        # hdf file
        self.f = h5.File(filename, 'r')
        self.destructions = {}
        self.formations = {}
        self.gasConditions = {}
        self.densities = {}
        self.masses = {}
        self.network = ""
        self.positionVar = {}
        self.rateUnit = ""
        self.__checkValidity()
        self.__buildSet()

    def __checkValidity(self):
        """
            Verify the the Metadata directory exists in the HDF file
        """
        try: 
            self.metadata = self.f['/Metadata/Metadata']
        except Exception as e:
            print(e)
            raise Exception("Metadata directory is missing in HDF5 file.")

    def close(self):
        """
            close hdf file
        """
        self.f.close()

    def __buildSet(self):        
        self.metadata = self.f['/Metadata/Metadata'].asstr()[:]
        i=0
        # build list of reactions and quantities
        for value in self.metadata:
            i = i+1
            # d_value = decode(value)
            d_value = value
            if d_value[1].startswith('Destruction rate'):
                if self.rateUnit == "" :
                    self.rateUnit = d_value[6]
                if d_value[1] not in self.destructions:
                    self.destructions[d_value[1]] = []

                self.destructions[d_value[1]].append(d_value)

            elif d_value[1].startswith('Formation rate'):
                if d_value[1] not in self.formations:
                    self.formations[d_value[1]] = []
                self.formations[d_value[1]].append(d_value)

            elif d_value[1] == "Gas conditions":
                self.gasConditions[str(d_value[3])] = d_value

            elif d_value[1] == "Positions":
                if len(self.positionVar.keys()) == 0:
                    self.positionVar = {'name': d_value[4],
                                        'unit': d_value[6],
                                        'short_name': d_value[3]}

            elif d_value[1].startswith("Densities"):
                self.densities[str(d_value[4])] = d_value

            elif "mass_of" in d_value[3]:
                qty = d_value[0] + '/' + d_value[1]
                self.masses[d_value[1].replace("Mass ", "")] = float(self.f[qty][0])

        if self.network_path in self.f.keys():
            self.getNetwork()

    def getAllSpecies(self):
        return self.f[self.chemistry_path].keys()

    def getConditionsValues(self, main_pos_vars, quantity_name):
        result = []
        for key in sorted(self.gasConditions.keys()):
            if key == quantity_name:
                value = self.gasConditions[key]
                for i in range(0, len(main_pos_vars)):
                    qty = str(value[0]) + '/' + str(value[1])
                    result.append(self.f[qty][i].tolist()[int(value[2])])

        return result

    def getTemperaturesMainPosVars(self):
        """
          returns values that stay constant for a given file :
            main_pos_var and temperatures

          returns dict[main_pos_vars, temperatures]

        """
        result = {}
        main_pos_vars = []

        for value in self.f[self.position_path]:
            main_pos_vars.append(value[0])

        result['temperatures'] = self.getConditionsValues(main_pos_vars, 'tgas')
        result['main_pos_vars'] = main_pos_vars
        return result

    def getFormationBySpecies(self, species):
        # field can be missing in shock hdf file
        try:
            return self.formations['Formation rates ' + species]
        except KeyError:
            return []

    def getDestructionBySpecies(self, species):
        # field can be missing in shock hdf file
        try:
            return self.destructions['Destruction rates ' + species]
        except KeyError:
            return []

    def getData(self, meta, main_pos_var):
        data = self.f[meta[0] + '/' + meta[1]]
        return data[main_pos_var][int(meta[2])]

    def getGasConditions(self, main_pos_var):
        result = {}
        for key in sorted(self.gasConditions.keys()):
            value = self.gasConditions[key]
            entry = {}
            entry['name'] = value[4]

            if value[6] != "no unit":
                entry['unit'] = value[6]
            else:
                entry['unit'] = ''

            # numpy has its own list type, integer stored as string
            qty = str(value[0]) + '/' + str(value[1])
            entry['value'] = self.f[qty][main_pos_var]\
                                 .tolist()[int(value[2])]
            result[value[4]] = entry
        return result

    def getNetwork(self):
        """
            concatenate content of Network table into a string and return it
        """
        if self.network == "":
            result = []
            try:
                for line in self.f[self.network_path]['Network'][0]:
                    result.append("%s \n"%line)
                self.network = ''.join(result)
            except KeyError as e:
                print("Network dataset is not defined in hdf5 file")
                print(e)
                self.network = 'Network is not available in the HDF5 file'

        return self.network

    def getAbundance(self, species, main_pos_var):
        sp = self.chemistry_path + '/' + species + '/Densities ' + species
        return self.f[sp][main_pos_var][0]

    def getAbundances(self, species):
        result = []
        for value in self.f[self.chemistry_path + '/' + species + '/Densities ' + species][:]:
            result.append(value[0])
        return result

    # if abundance is always cm-3, this should be removed
    # to save memory and iterations
    def getAbundanceUnit(self, species):
        # return "cm-3"
        return self.densities["n(" + species + ")"][6]

    def getTotalDestructionRate(self, species, main_pos_var):
        sp = self.chemistry_path + '/' + species \
                                 + '/Total dest rates ' + species
        return self.f[sp][main_pos_var][0]

    def getTotalDestructionRates(self, species):
        sp = self.chemistry_path + '/' + species \
                                 + '/Total dest rates ' + species
        # HDF5 list in not json serializable
        result = [ e[0] for e in self.f[sp]]
        return result

    def getTotalFormationRate(self, species, main_pos_var):
        sp = self.chemistry_path + '/' + species \
                                       + '/Total form rates ' + species
        return self.f[sp][main_pos_var][0]

    def getTotalFormationRates(self, species):
        sp = self.chemistry_path + '/' + species \
                                       + '/Total form rates ' + species
        # HDF5 list in not json serializable
        result = [ e[0] for e in self.f[sp]]
        return result

    def getFormationRates(self, species):
        sp = self.chemistry_path + '/' + species \
                                       + '/Formation rates ' + species
        return self.f[sp]

    def getDestructionRates(self, species):
        sp = self.chemistry_path + '/' + species \
                                       + '/Destruction rates ' + species
        return self.f[sp]

    def getMass(self, species):
        try:
            return self.masses[species]
        except KeyError:
            return 0

    def getMasses(self):
        return self.masses

    def getPositionVarName(self):
        return self.positionVar
