# -*- coding: utf-8 -*-

__author__ = "Nicolas Moreau"
__copyright__ = "Copyright 2020"
__credits__ = ["Nicolas Moreau"]
__license__ = "GPL"
__version__ = "1.1"
__maintainer__ = "Nicolas Moreau"
__email__ = "nicolas.moreau@obspm.fr"
__status__ = "Production"


import os
import os.path
import json
import collections
import uuid
from bottle import route, run, request, default_app,\
    BaseTemplate, response, abort, static_file, BaseRequest
from analyzer import Analyzer
import config as config


def enable_cors(fn):
    """ enable cross domain ajax requests"""
    def _enable_cors(*args, **kwargs):
        # set CORS headers
        response.headers['Access-Control-Allow-Origin'] = '*'
        response.headers['Access-Control-Allow-Methods'] = 'GET, POST, PUT, OPTIONS'
        response.headers['Access-Control-Allow-Headers'] = 'Origin, Accept, Content-Type, X-Requested-With, X-CSRF-Token'

        if request.method != 'OPTIONS':
            # actual request; reply with the actual response
            return fn(*args, **kwargs)

    return _enable_cors

def check_directory(directory):
    """  creates directory if it does not exist

    :param directory: name of a directory to create if it does not exist
    :returns: name of directory
    """
    if os.path.exists(directory) is False:
        try:
            print("directory '%s' is missing, try creating it " % directory)
            os.mkdir(directory)
            print("directory '%s' created " % directory)
        except Exception as e:
            print("error while creating '%s' directory" % directory)
            print(e)

    return directory


def save_data(filename, string_data):
    """create a file a write a string in it

    :param filename: name of the file to create,
                     if it exists it will be replaced
    :param string_data: content to write in the file

    :returns: True if the written file content is identical
                   to the string parameter, or False if not

    """
    f = open(os.path.join(config.SAVE_DIRECTORY, filename), "w")
    f.write(string_data)
    f.close()

    f = open(os.path.join(config.SAVE_DIRECTORY, filename), "r")
    content = f.read()
    f.close()

    # check content of written file
    if content == string_data:
        return True

    return False


def build_main_pos_var_export(main_pos_var, data):
    """writes a file containing rates for all main_pos_vars for a species

    :param main_pos_var:  name of position variable to navigate data (
                          AV or position )
    :param data:          a dictionnary contaning data
                          data['reactions'] list of reactions
                          data['main_pos_vars'][main_pos_var_value][reaction_name][rate] values

    :returns: a string containing the formatted data

    """
    col_index = 1
    string = "#[%s] %s     " % (col_index, main_pos_var)

    for reaction in data['reactions']:
        col_index += 1
        string += "[%s] %s      " % (col_index, reaction)
    string += "\n#[1]      "

    col_index = 1
    for unit in data['units']:
        col_index += 1
        string += "[%s] %s      " % (col_index, unit)
    string += "\n#"

    for i in range(col_index):
        string += ("[%s]" % str(i + 1)).rjust(20)
    string += "\n"
    for key in collections.OrderedDict(sorted(data["main_pos_vars"].items())):
        string += ("%2.5e" % key).rjust(20)
        for reaction in data['reactions']:
            if reaction in data['main_pos_vars'][key]:
                string += ("%2.10e" % data['main_pos_vars'][key][reaction]['value']).rjust(20)
            else:
                string += "_".rjust(20)

        string += "\n"

    return string


class AnalyzerManager():
    """
     Manages Analyzer object instances
     Each instance contains one hdf file
    """
    def __init__(self):
        self.analyzers = {}
        self.currentAnalyzer = None

    def getAnalyzer(self, client_id):
        """ returns Analyzer corresponding to the given client id
        :param client_id: id of the client process
        :returns: Azalyzer object
        """
        return self.analyzers[client_id]

    def addAnalyzer(self, client_id, filename):
        """ creates a new Analyzer object for the given client id
        :param client_id: id of the client process
        :param filename: name of the file to open
        """
        try:
            self.analyzers[client_id] = self.instantiateAnalyzer(filename)
        except Exception as e:
            raise e

    def instantiateAnalyzer(self, filename):
        """returns a Analyzer object for the hdf5 file filename

        :param filename: name of hdf5 file
        :returns: Analyzer object
        """
        analyzer = Analyzer(config.DATA_DIRECTORY)
        analyzer.loadHdf(filename)
        return analyzer


"""
    urls
"""
base_url = "/chemistry-analyzer"

@route(base_url+'/session/<session>/saverates/species/<species>/threshold/<threshold>/filename/<filename>',
       name='ratespecies')
@enable_cors
def rate_species(session, species, threshold, filename):
    """ writes locally rates for all reactions of given species
        at all main_pos_vars if at least one rate is above threshold and
    returns a dictionary containing the name of the generated files
    :param session: id of the client process
    :param species: species name
    :param threshold: threshold value
    :param filename: suffix for file name

    :returns: a dictionary of file names

    """
    global analyzerManager
    main_pos_var = analyzerManager.getAnalyzer(session).getPositionVarName()
    data = analyzerManager.getAnalyzer(session) \
                          .getAllReactions(species, float(threshold))
    save_data("formation_" + filename,
              build_main_pos_var_export(main_pos_var['short_name'],
                                        data['formation']))
    save_data("destruction_" + filename,
              build_main_pos_var_export(main_pos_var['short_name'],
                                        data['destruction']))
    return json.dumps({"formation": "formation_" + filename,
                       "destruction": "destruction_" + filename})


@route(base_url + "test")
def test():
    return "test"


@route(base_url + '/opensession', name='opensession')
@enable_cors
def open_session():
    """ generates and returns an id for this session
        to identify connection with client

    :returns: random uuid

    """
    return json.dumps({'sessionid': uuid.uuid4().hex})



@route(base_url + '/session/<session>/mainposvar/<main_pos_var:int>/species/<species>',
       name='mainposvarspecies')
@enable_cors
def main_pos_var_species(session, main_pos_var, species):
    """ returns list of reactions for a species at a given main_pos_var
    :param session: id of the client process
    :param main_pos_var: main_pos_var value
    :param species: species name

    :returns: a dictionary of reactions

    """
    global analyzerManager
    return json.dumps(analyzerManager.getAnalyzer(session) \
                          .getReactions(species, main_pos_var))


@route(base_url + '/session/<session>/abundances/species/<species>',
       name='abundancesspecies')
@enable_cors
def abundances(session, species):
    """ returns abundances of the given species at all main_pos_vars
    :param session: id of the client process
    :param species: species name

    :returns: a list of abundance at each main_pos_var

    """
    global analyzerManager
    return json.dumps(analyzerManager.getAnalyzer(session)
                                     .getAbundances(species))


@route(base_url + '/session/<session>/network',
       name='network')
@enable_cors
def network(session):
    """ returns network in given hdf file
    :returns: string

    """
    global analyzerManager
    return json.dumps(analyzerManager.getAnalyzer(session).getNetwork())


@route(base_url + '/session/<session>/positionvarname',
       name='positionvarname')
@enable_cors
def positionvarname(session):
    """ returns network in given hdf file
    :returns: string

    """
    global analyzerManager
    return json.dumps(analyzerManager.getAnalyzer(session)
                                     .getPositionVarName())


@route(base_url + '/session/<session>/gasconditions/mainposvar/<main_pos_var:int>',
       name='gasconditions')
@enable_cors
def gas_conditions(session, main_pos_var):
    """ returns gas conditions at given main_pos_var value
    :param session: id of the client process
    :param main_pos_var: main_pos_var value

    :returns: a dictionary containing gas conditions
    """
    global analyzerManager
    return json.dumps(analyzerManager.getAnalyzer(session)
                                     .getGasConditions(main_pos_var))


@route(base_url + '/session/<session>/allmainposvars',
       name='main_pos_vars')
@enable_cors
def mainposvars(session):
    """ returns a list of all main_pos_var positions
    :param session: id of the client process

    :returns: a list of values
    """
    global analyzerManager
    return json.dumps(analyzerManager.getAnalyzer(session)
                                     .getTemperaturesMainPosVars())


@route(base_url + '/session/<session>/allspecies',
       name='species')
@enable_cors
def species(session):
    """ returns a list of all species
    :param session: id of the client process

    :returns: a list of species
    """
    global analyzerManager
    # convert KeysViewHDF5 to list for json export
    return json.dumps(analyzerManager.getAnalyzer(session).getSpecies())



@route(base_url+'/allfiles',
       name='files')
@enable_cors
def files():
    """ returns a list of files contained in the data directory (file name must contain _c string)

    :returns: a list of file names
    """
    result = []
    for entry in os.listdir(config.DATA_DIRECTORY):
        if "_c" in entry:
            result.append(entry)
    return json.dumps(result)



@route(base_url + '/session/<session>/changefile/file/<filename>',
       name='changefile')
@enable_cors
def change_file(session, filename):
    """ loads a new file
    :param session: id of the client process
    :param filename: name of the file to load

    :returns: 200 status if file is loaded, 204 if not found

    """
    global analyzerManager

    try:
        analyzerManager.addAnalyzer(session, filename)

        if analyzerManager.getAnalyzer(session).hdf is not None:
            response.status = 200
            return {}
        else:
            response.status = 500
    except Exception as e:
        print("error")
        response.status = 500
        return json.dumps({"status" : "error", "message" : str(e)})


@route(base_url + '/session/<session>/positionvar')
def positionvar(session):
    """ returns the content of a file into the save directory
    :param filename: name of the file to read

    :returns: file content, 404 error if not found
    """
    global analyzerManager
    return json.dumps(analyzerManager.getAnalyzer(session)
                                     .getMainPositionVariable())


@route(base_url + '/savefile', name='savefile',
       method="POST")
@enable_cors
def save_file():
    """ saves a string into a file in save directory
    :param filename: name of the file to write
    :param string_data: content to write

    :returns: an empty list if the file has been written,
              HTTP 500 error if a problem occured

    """
    filename = request.forms.get('filename')
    string_data = request.forms.get('string_data')
    write_status = save_data(filename, string_data)

    # check content of written file
    if write_status is True:
        response.status = 200
        return []
    else:
        abort(500, "Server interal error : File was not correctly written.")


@route(base_url + '/existfile/<filename>',
       name='existfile', method="GET")
@enable_cors
def file_exists(filename):
    """ checks if a file already exists in the save directory
    :param filename: name of the file to test

    :returns: True if the file exists, False if not
    """
    result = {'status': os.path.isfile(os.path.join(config.SAVE_DIRECTORY, filename))}
    return result


@route(base_url + '/static/<filename:path>')
def static_content(filename):
    """ returns the content of a file into the save directory
    :param filename: name of the file to read

    :returns: file content, 404 error if not found
    """
    return static_file(filename, root=config.SAVE_DIRECTORY)


@route(base_url, name='home', method="GET")
@enable_cors
def home():
    """ home page

    :returns: string message
    """
    result = "chemistry analyzer server default page"
    return result


def runserver(port, debug):
    app = default_app()
    BaseTemplate.defaults['get_url'] = app.get_url
    run(host='127.0.0.1', port=port, debug=debug, server='cheroot')

analyzerManager = AnalyzerManager()
check_directory(config.DATA_DIRECTORY)
check_directory(config.SAVE_DIRECTORY)

def main():
    # increase max post size to 8 Mb to save big files
    BaseRequest.MEMFILE_MAX = 8 * 1024 * 1024
    runserver(config.APPLICATION_PORT, True)


if __name__ == "__main__":
    main()
