# -*- coding : utf-8 -*-

import sys, os
import json
import httplib
import base_test as test

    
class TestServer(test.Test):
  
  def __init__(self):
    super(TestServer, self).__init__() 
    self.port = "12345"
    self.session = None
    self.testFile = "test_c_1.hdf5"
    self.testSaveFile = "test.csv"
    self.avCount = None
    
  def doTests(self):
    self.testHome()
    self.testFiles()
    self.testOpenSession()
    self.testChangeFile()
    self.testGetAvs()
    self.testSpecies()
    self.testAvSpecies('H')
    self.testAvSpecies('H::')
    self.testAvSpecies('HCN+')
    self.testAvSpecies('e-')
    self.testAbundances("H")
    self.testAbundances("H::")
    self.testAbundances("HCN+")
    self.testAbundances("e-")
    self.testGasConditions()
    self.testFileExist()
    self.testFileDoesNotExist()
    self.testRateSpecies("HCN+")
    self.testExistingStaticContent()
    self.testNotExistingStaticContent()
    return self.getLogString()
    
  def baseTest(self, doTest, source=None) :
    if source is not None : 
      self.resultAppend("do test : %s "% source)    
    try : 
      doTest()
    except AssertionError as e:
      self.addProblem()
      self.resultAppend("assertion error : %s "%e)
      #~ self.resultAppend(str(e))           
    except httplib.ResponseNotReady as e : 
      self.addProblem()
      self.resultAppend("connection error : %s "%e)
    except httplib.HTTPException as e :
      self.addProblem()
      self.resultAppend("connection error : %s "%e)
    except Exception as e : 
      self.addProblem()
      self.resultAppend("error : %s "%e)       
    

  """ test default url
    
  """
  def testHome(self):   
    
    conn = httplib.HTTPConnection("localhost:"+self.port)
    conn.request("GET", "/pdrchemistry")
    response = conn.getresponse()
    status = response.status
    content = response.read()
    
    def test1():
      assert(200 == status), "request status is not 200 : %s" % status
      
    def test2():
      assert("chemistry analyzer server default page" == content), "default page text is not correct : %s" % content
      
    self.baseTest(test1, "testHome")   
    self.baseTest(test2)   

      
  def testOpenSession(self): 
    conn = httplib.HTTPConnection("localhost:"+self.port)
    conn.request("GET", "/pdrchemistry/opensession")
    response = conn.getresponse()
    status = response.status
    content = json.loads(response.read())
    
    def test1() :
      assert("sessionid" in content), "no session id key in response: %s" % content
      
    def test2():
      assert(len(content["sessionid"]) == 32), "uuid format is not correct : %s" % content["sessionid"]
      
    self.baseTest(test1, "testOpenSession")        
    self.baseTest(test2)  
    self.session = content["sessionid"]    
      
  def testFiles(self):    
    conn = httplib.HTTPConnection("localhost:"+self.port)
    conn.request("GET", "/pdrchemistry/allfiles")
    response = conn.getresponse()
    status = response.status
    content = json.loads(response.read())
    
    def test1():
      assert(len(content) > 0), "no file in data directorty"
    
    def test2():
      assert(self.testFile in content), "test file is missing in data directory"
      
    self.baseTest(test1, "testFiles")       
    self.baseTest(test2)   
      
    for value in content : 
      def testloop():
        assert("_c_" in value), "file name is not correct : %s" % value        
      self.baseTest(testloop)     
      
      
  def testChangeFile(self):
    conn = httplib.HTTPConnection("localhost:"+self.port)
    
    conn.request("GET", "/pdrchemistry/session/"+self.session+"/changefile/file/unknownfile")
    response = conn.getresponse()
    status = response.status
    
    def test1() :
      assert(status == 204), "incorrect status code for file not found : %s" % status     
      
    self.baseTest(test1, "testChangeFile")  
      
    conn.request("GET", "/pdrchemistry/session/"+self.session+"/changefile/file/"+self.testFile)
    response = conn.getresponse()
    status = response.status
    content = json.loads(response.read())      
      
    def test2() :
      assert(len(content) == 0), "incorrect response : %s" % content
    
    def test3():
      assert(status == 200), "test file not found"      

    self.baseTest(test2)  
    self.baseTest(test3)  
      
      
  def testGetAvs(self):
    conn = httplib.HTTPConnection("localhost:"+self.port)
    
    conn.request("GET", "/pdrchemistry/session/"+self.session+"/allavs")
    response = conn.getresponse()
    status = response.status
    content = json.loads(response.read()) 
    
    def test1() :       
      assert(status == 200), "incorrect status code : %s" % status  
    
    def test2() :          
      assert(len(content) > 0), "incorrect response : %s" % content 
      
    def test3() :   
      assert("temperatures" in content), 'temperature list is missing'
      
    def test4() :   
      assert("avs" in content), 'av list is missing'
      
    def test5() :   
      assert(len(content["temperatures"]) > 0 ), 'temperature list is empty'
      
    def test6() :   
      assert(len(content["avs"]) > 0 ), 'av list is empty'
      
      
    self.baseTest(test1, "testGetAvs")  
    self.baseTest(test2)  
    self.baseTest(test3)  
    self.baseTest(test4)  
    self.baseTest(test5)  
    self.baseTest(test6)  
    self.avCount = len(content["avs"])

    
  def testSpecies(self):
    conn = httplib.HTTPConnection("localhost:"+self.port)
    
    conn.request("GET", "/pdrchemistry/session/"+self.session+"/allspecies")
    response = conn.getresponse()
    status = response.status
    content = json.loads(response.read())  
    
    def test1():      
      assert(status == 200), "incorrect status code : %s" % status  
      
    def test2():
      assert(len(content) > 0 ), "species list is empty" 
      
    def test3():
      assert('H' in content ), "H is not in species list" 
      
    def test4():
      assert('H+' in content ), "H+ is not in species list" 
     

    self.baseTest(test1, "testSpecies")  
    self.baseTest(test2)  
    self.baseTest(test3)  
    self.baseTest(test4)  
    
    
  def testAvSpecies(self, species):
    conn = httplib.HTTPConnection("localhost:"+self.port)
    conn.request("GET", "/pdrchemistry/session/"+self.session+"/av/0/species/"+species)
    response = conn.getresponse()
    status = response.status
    content = json.loads(response.read())  
    
    def test1() :      
      assert(status == 200), "incorrect status code : %s" % status          
      
    def test2() :
      assert(species in content), "species not found in result array "
      
    def test3() :
      assert("abundance" in content[species]), 'abundance value is missing in result array'
      
    def test4() :      
      assert("abundance_unit" in content[species]), 'abundance unit value is missing in result array'
      
    def test5() :      
      assert("abundances" in content[species]), 'abundance values for different av is missing in result array'
      
    def test6() :      
      assert(len(content[species]["abundances"]) ==  self.avCount ), 'abundance values for different av is different from number of av values : %s instead of %s' % (len(content['H']["abundances"]), self.avCount)

    def test7() :      
      assert('creation' in content[species]), 'list of formation reactions is missing'
    
    def test8() :      
      assert('destruction' in content[species]), 'list of destruction reactions is missing'
      
    self.baseTest(test1, "testAvSpecies "+species)  
    self.baseTest(test2)  
    self.baseTest(test3)  
    self.baseTest(test4)  
    self.baseTest(test5)  
    self.baseTest(test6)  
    self.baseTest(test7)  
    self.baseTest(test8)        
      
    formation = content[species]['creation'][0]
    def test1():
      assert("reaction" in formation), "reaction is missing in formation_reaction[0]"
      
    def test2():
      assert("reactants" in formation), "reactants are missing in formation_reaction[0]"    
        
    def test3():
      assert("reactants_abundance_units" in formation), "reactants abundance unit is missing in formation_reaction[0]"
    
    def test4():
      assert("proportion" in formation), "proportion is missing in formation_reaction[0]"
      
    def test5():
      assert("%" in formation['proportion']), "proportion value is not correct in formation_reaction[0]"
    
    def test6():
      assert("rate" in formation), "rate is missing in formation_reaction[0]"      
    
    def test7():
      assert("products" in formation), "products are missing in formation_reaction[0]"
      
    def test8():
      assert("reactants_abundances" in formation), "reactants abundances are missing in formation_reaction[0]"
      
    def test9():
      assert("unit" in formation), "unit is missing in formation_reaction[0]"
      
    def test10():
      assert("name" in formation["reactants"][0]), "reactant name are missing in %s" % formation["reactants"][0]
      
    def test11():
      assert("mass" in formation["reactants"][0]), "reactant mass are missing in %s" % formation["reactants"][0]
      
    def test12():
      assert("name" in formation["products"][0]), "product name are missing in %s" % formation["products"][0]
      
    def test13():
      assert("mass" in formation["products"][0]), "product mass are missing in %s" % formation["products"][0]
       
    self.baseTest(test1)  
    self.baseTest(test2)  
    self.baseTest(test3)  
    self.baseTest(test4)  
    self.baseTest(test5)  
    self.baseTest(test6)  
    self.baseTest(test7)  
    self.baseTest(test8)   
    self.baseTest(test9)  
    self.baseTest(test10) 
    self.baseTest(test11) 
    self.baseTest(test12) 
    self.baseTest(test13) 
        
    destruction = content[species]['destruction'][0]
    
    def test1():
      assert("reaction" in destruction), "reaction is missing in destruction_reaction[0]"
      
    def test2():
      assert("reactants" in destruction), "reactants are missing in destruction_reaction[0]"
      
    def test3():
      assert("reactants_abundance_units" in destruction), "reactants abundance unit is missing in destruction_reaction[0]"
    
    def test4():
      assert("proportion" in destruction), "proportion is missing in destruction_reaction[0]"
      
    def test5():  
      assert("%" in destruction['proportion']), "proportion value is not correct in destruction_reaction[0]"
    
    def test6():
      assert("rate" in destruction), "rate is missing in destruction_reaction[0]"      
    
    def test7():
      assert("products" in destruction), "products are missing in destruction_reaction[0]"
      
    def test8():
      assert("reactants_abundances" in destruction), "reactants abundances are missing in destruction_reaction[0]"
      
    def test9():
      assert("unit" in destruction), "unit is missing in destruction_reaction[0]"     
      
    def test10():
      assert("name" in destruction["reactants"][0]), "reactant name are missing in %s" % destruction["reactants"][0]
      
    def test11():
      assert("mass" in destruction["reactants"][0]), "reactant mass are missing in %s" % destruction["reactants"][0]
      
    def test12():
      assert("name" in destruction["products"][0]), "product name are missing in %s" % destruction["products"][0]
      
    def test13():
      assert("mass" in destruction["products"][0]), "product mass are missing in %s" % destruction["products"][0]
      
    self.baseTest(test1)  
    self.baseTest(test2)  
    self.baseTest(test3)  
    self.baseTest(test4)  
    self.baseTest(test5)  
    self.baseTest(test6)  
    self.baseTest(test7)  
    self.baseTest(test8)   
    self.baseTest(test9)   
    self.baseTest(test10)   
    self.baseTest(test11)   
    self.baseTest(test12)   
    self.baseTest(test13)   
    
    
  def testAbundances(self, species):
    conn = httplib.HTTPConnection("localhost:"+self.port)      
    conn.request("GET", "/pdrchemistry/session/"+self.session+"/abundances/species/"+species)
    response = conn.getresponse()
    status = response.status
    content = json.loads(response.read())        
    
    def test1() :
      assert(status == 200), "incorrect status code : %s" % status          
      
    def test2() :
      assert(len(content) == self.avCount), "abundance values count is not equal to av values count : %s instead of %s" % (len(content), self.avCount)      

    self.baseTest(test1, "testAbundances "+species)  
    self.baseTest(test2)  
    

    
  def testGasConditions(self):
    conn = httplib.HTTPConnection("localhost:"+self.port) 
    conn.request("GET", "/pdrchemistry/session/"+self.session+"/gasconditions/av/1")
    response = conn.getresponse()
    status = response.status
    content = json.loads(response.read())  
    
    def test1() :      
      assert(status == 200), "incorrect status code : %s" % status   
      
    def test2() :    
      assert('nH' in content), "nH is missing in gas conditions"
      
    def test3() :    
      assert('nd(e-)' in content), "nd(e-) is missing in gas conditions"
      
    def test4() :    
      assert('Xray' in content), "Xray is missing in gas conditions"
      
    def test5() :    
      assert('Tg' in content), "Tg is missing in gas conditions"
      
    def test6() :    
      assert('zeta' in content), "zeta is missing in gas conditions"
      
    def test7() :    
      assert('ng(e-)' in content), "ng(e-) is missing in gas conditions"
      
    def test8() :    
      assert('UV' in content), "UV is missing in gas conditions"
      
    def test9() :    
      assert('Td(01)' in content), "Td(01) is missing in gas conditions"
      
    def test10() :    
      assert('nd(01)' in content), "nd(01) is missing in gas conditions"
      
    self.baseTest(test1, "testGasConditions")  
    self.baseTest(test2)
    self.baseTest(test3)
    self.baseTest(test4)
    self.baseTest(test5)
    self.baseTest(test6)
    self.baseTest(test7)
    self.baseTest(test8)
    self.baseTest(test9)
    self.baseTest(test10)
      
    nh = content['nH']
  
    def test1() :    
      assert('unit' in nh), "unit is missing in gas condition properties"
      
    def test2() :
      assert('name' in nh), "name is missing in gas condition properties"
      
    def test3() :
      assert('value' in nh), "value is missing in gas condition properties"


    self.baseTest(test1)  
    self.baseTest(test2)
    self.baseTest(test3) 
    
  def testFileDoesNotExist(self):
    conn = httplib.HTTPConnection("localhost:"+self.port)      
    conn.request("GET", "/pdrchemistry/existfile/testfile")
    response = conn.getresponse()
    status = response.status
    content = json.loads(response.read())        

    def test1() :
      assert(status == 200), "incorrect status code : %s" % status          
      
    def test2() :
      assert("status" in content), "status field is missing in response"
      
    def test3() :
      assert(content["status"] == False), "status field value is wrong, must be False when file not found"      

    self.baseTest(test1, "testFileDoesNotExist ")  
    self.baseTest(test2)  
    self.baseTest(test3)  
    
    
  def testFileExist(self):
    conn = httplib.HTTPConnection("localhost:"+self.port)      
    conn.request("GET", "/pdrchemistry/existfile/"+self.testSaveFile)
    response = conn.getresponse()
    status = response.status
    content = json.loads(response.read())        

    def test1() :
      assert(status == 200), "incorrect status code : %s" % status          
      
    def test2() :
      assert("status" in content), "status field is missing in response"
      
    def test3() :
      assert(content["status"] == True), "status field value is wrong, must be False when file not found"      

    self.baseTest(test1, "testFileExist ")  
    self.baseTest(test2)  
    self.baseTest(test3)  
    
  def testRateSpecies(self, species):
    
    conn = httplib.HTTPConnection("localhost:"+self.port)
    
    conn.request("GET", "/pdrchemistry/session/"+self.session+"/saverates/species/"+species+"/threshold/0.1/filename/rate_species.test")
    response = conn.getresponse()
    status = response.status
    content = json.loads(response.read())  
    
    def test1() :
      assert(status == 200), "incorrect status code : %s" % status          
      
    def test2() :
      assert("formation" in content), "formation key is missing in response object"
      
    def test3() :
      assert("destruction" in content), "destruction key is missing in response object"    
      
    def test4() :
      assert(os.path.exists("../save/formation_rate_species.test")), "file containing rates by av for formation reactions is missing"   
      
    def test5() :
      assert(os.path.exists("../save/destruction_rate_species.test")), "file containing rates by av for destruction reactions is missing"        
    
      
    self.baseTest(test1, "testRateSpecies ")  
    self.baseTest(test2)  
    self.baseTest(test3)        
    self.baseTest(test4)
    self.baseTest(test5)  
    
  def testExistingStaticContent(self):   
    
    conn = httplib.HTTPConnection("localhost:"+self.port)
    conn.request("GET", "/pdrchemistry/static/"+self.testSaveFile)
    response = conn.getresponse()
    status = response.status
    content = response.read()
    
    def test1():
      assert(200 == status), "request status is not 200 : %s" % status
      
    def test2():
      assert("test content" == content.strip()), "content found in test file is not correct : %s" % content
      
    self.baseTest(test1, "testExistingStaticContent")   
    self.baseTest(test2)   
    
    
  def testNotExistingStaticContent(self):   
    
    conn = httplib.HTTPConnection("localhost:"+self.port)
    conn.request("GET", "/pdrchemistry/static/notexistingfile")
    response = conn.getresponse()
    status = response.status
    content = response.read()
    
    def test1():
      assert(404 == status), "request status must be 404 when content not found : %s" % status
            
    self.baseTest(test1, "testNotExistingStaticContent")   
