# -*- coding: utf-8 -*-
import h5py as h5
import numpy as numpy
import time

from lib import hdfutil as hdfu

__author__ = "Nicolas Moreau"
__copyright__ = "Copyright 2012"
__credits__ = ["Nicolas Moreau"]
__license__ = "GPL"
__version__ = "1.0"
__maintainer__ = "Nicolas Moreau"
__email__ = "nicolas.moreau@obspm.fr"
__status__ = "Development"


'''
Notes :

PS: Table Metadata :
 Metadata contains :
PathHDF5 | DataSet | Col.Index | PubID | HName  | DataType | Unit | SKOS | UCD | utype | Description | ObjParent | Group |
   0     |    1    |     2     |   3   |   4    |    5     |  6   |  7   |  8  |  9    |   10        |    11     |  12   |
   
 Metadata_ObjType contains :
 PubID | Hname | SKOS | Description

'''



class PdrHDF(object):
    """
      Access to a pdr HDF5 file
    """
    def __init__(self, filename, treeManager=None):
        # hdf file
        print('opening hdf5 file {}...'.format(filename))
        self.f = h5.File(filename, 'r')
        """
        create a pivoted array with the metadata
         ------------------------------------------------------------------------------------
         The old versions of numpy do not treat automatically the fact that
         the metadata table is composed of string of variable lengths.
         Beside updating numpy, another solution consists in setting the type
         to "object" when creating the numpy array
         see http://stackoverflow.com/questions/16037824/how-to-convert-numpy-object-array-into-str-unicode-array
         note: setting the type to unicode or post-converting
         object type via astype('U') does not work either
         with the old version of numpy
         ------------------------------------------------------------------------------------
        """
        
        self.xQuantitiesPath = "/Local quantities/Positions"
        # separator in path
        self.separator = "#"
        # show only local quantites directory to user
        # ~ self.selectableData = "/Local quantities"
        self.treeManager = treeManager
        
        # self.metadata = self.f['/Metadata/Metadata'][:]
        # self.metadata = numpy.empty(self.f['/Metadata/Metadata'].shape,
        #         h5.special_dtype(vlen = str))
        # self.f['/Metadata/Metadata'].read_direct(self.metadata)

        # self.metadata = self.f['/Metadata/Metadata'].asstr()
        # this makes idat faster than the version with just .asstr() because
        # it loads the dataset in memory. Since the metadta dataset is not
        # that big, it's ok.
        self.metadata = self.f['/Metadata/Metadata'].asstr()[:]


        # indexes all the data
        st = time.time()
        init_result = self._buildIndex()
        print("time to build index: {} s".format((time.time() - st)))
        self.xQuantities = init_result["xquantities"]
        self.index = init_result["pointers"]

    def close(self):
        """
            close hdf file
        """
        self.f.close()

    def getByGroup(self, group):
        """
            returns the object located at group position
        """
        try:
            return self.f[group]
        except KeyError:
            print("group : " + group + " not found")

    def getDefaultColumns(self):
        result = []
        for key in self.index['Positions']:
            result.append(key)
        return result

    def _buildIndex(self):
        """
            creates the index
        """
        result = {}
        # map of Pointer objects
        pointers = {}  
        # quantities that can be used as x axis on result plot
        result["xquantities"] = [] 

        for line in self.metadata:
            pf = hdfu.PointerFactory()
            pf.withPath(line[0]).withDataset(line[1])
            pf.withName(line[4]).withColumn(line[2])
            pf.withUnit(line[6]).withUcd(line[8])
            pf.withSkos(line[7]).withUtype(line[9])
            pf.withDescription(line[10]).withObjParent(line[11])
            pf.withGroup(line[12]).getPointer()            
            
            # pointer is identified by the human readable name of a quantity
            index_name = line[4]
            if isinstance(index_name, str) is False:
                raise Exception("%s is not a string " % index_name)
            pointers[index_name] = pf.getPointer()

            if self.treeManager is not None : 
                self.treeManager.addToTree(line, pointers, self.separator)
            
            if line[0] + '/' + line[1] == self.xQuantitiesPath:
                result['xquantities'].append(line[4])

            result["pointers"] = pointers
            # ~ if self.treeManager is not None : 
                # ~ result["tree"] = self.treeManager.tree[0]

        return result

    def extractQuantity(self, p):
        """
            extract a quantity according to the pointer p
        """
        
        def _adapt_str_ds(ds):
            if h5.check_string_dtype(ds.dtype) != None:
                return ds.asstr()[:]
            return ds[:]

        ds = self.f["%s/%s" % (p.path, p.dataset)]
      
        # 1D array, returned directly
        if len(ds.shape) != 1:
            ds = _adapt_str_ds(ds)[:, int(p.column)]

        return _adapt_str_ds(ds)


    def getCode(self):
        """
            return version of the code (pdr) as string or empty
            string if property is not found
        """
        if "Code version" in self.index:
            return self.extractQuantity(self.index["Code version"])[0]
        else:
            return ""


