import copy
import shutil
from io import StringIO
import traceback
#import numpy as numpy

from lib import votable as vot


def _getValue(value, vtype):
  if vtype == "double":
      return "%.15e"%value
  elif vtype == "float":
      return "%2.10e"%value
  else:
      return value

#def _h5unicode(l):
#    return [c.decode() for c in l]

class PointerFactory(object):
    """
    creates a Pointer object
    """
    def __init__(self):
        self.pointer = Pointer()

    def withPath(self, path):
        self.pointer.path = path
        return self

    def withDataset(self, dataset):
        self.pointer.dataset = dataset
        return self

    def withName(self, name):
        self.pointer.name = name
        return self

    def withColumn(self, column):
        self.pointer.column = column
        return self

    def withUnit(self, unit):
        self.pointer.unit = unit
        return self

    def withUcd(self, ucd):
        self.pointer.ucd = ucd
        return self

    def withSkos(self, skos):
        self.pointer.skos = skos
        return self

    def withUtype(self, utype):
        self.pointer.utype = utype
        return self

    def withDescription(self, description):
        description = description.strip()
        if description != '' and description != "none":
            self.pointer.description = description.strip()
        return self

    def withObjParent(self, parent):
        self.pointer.objParent = parent
        return self

    def withGroup(self, group):
        self.pointer.group = group
        return self

    def getPointer(self):
        return self.pointer

    def getErrorPointer(self, quantity):
        self.pointer = Pointer()
        self.pointer.name = quantity
        self.pointer.unit = "N/A"
        self.pointer.description = "error"
        return self.pointer


class Pointer(object):
    """
        Points on data in a dataset
    """
    def __init__(self):
        # dataset path
        self.path = None
        # dataset name
        self.dataset = None
        # quantity human readable name
        self.name = None
        # column position in dataset
        self.column = None
        self.ucd = None
        self.skos = None
        self.utype = None
        self.unit = None
        self.description = None
        self.objParent = None
        self.group = None


class TreeManager(object):
    """
    Class returning a data structure conforming to the fancytre library
    https://wwwendt.de/tech/fancytree/demo/index.html#

    """
    def __init__(self, hdffile):
        self._tree = [{'title': '', 'key': 1, 'folder': True, 'children': []}]
        self._cache = {}  # cache for already explored paths
        self._nodecount = 1
        self._hdffile = hdffile
        self._pointers = hdffile.index
        self._separator = hdffile.separator
        self._fillTree()
        
    @property
    def tree(self):
        return self._tree[0]
        
    def _incremNodeCount(self):
        self._nodecount = self._nodecount + 1

    def isCached(self, path):
        """
          Returns True is a path is in the cache
          @param path   path to test
        """
        if path in self._cache:
            return True
        return False

    def _getTooltip(self, pointed):
        """
          Returns the description of a pointer object to be used as a tooltip
          @param pointed a string used as a key to search into pointers map
          @param pointers a map of apointer objects
        """
        try:
            return self._pointers[pointed].description.replace("<![CDATA[", "").replace("]]>", "")
        except KeyError:
            return ""
        except AttributeError:
            return ""

    def _getDataType(self, branch):
        if "/Local quantities" in branch:
            return "local"
        return "integrated"

    def _recursTree(self, values, tree, path):
        """
          recurse on values until the leaf and append the lead to the list

        """
        local_list = copy.deepcopy(values)
        # search part of path already created in tree
        for value in local_list:
            for element in tree:
                if element['title'] == value:
                    local_list.pop(0)
                    if len(local_list) > 0:
                        self._recursTree(local_list, element["children"],
                                         path)
                    return True

        # current element still contains element, is a folder and has children
        if len(local_list) > 1:
            self._incremNodeCount()
            children = []
            tree.append({"title": local_list[0],
                         "key": self._nodecount,
                         "folder": True,
                         "children": children})
            local_list.pop(0)
            self._recursTree(local_list, children, path)
        # element is a leaf
        else:
            self._incremNodeCount()
            # quantity names are unique in data, used as key for tree leaves
            tree.append({"title": local_list[0],
                         "key": local_list[0],
                         "type": self._getDataType(path),
                         "tooltip": self._getTooltip(local_list[0])})
            self._cache[path] = tree

    def _addValue(self, metaline):
        """
          add content of values list into path
          each element of the list will be a new node in the tree if
          it does not exist already
          the content of path variable is used as an index in the cache
          to directly get the list of leaves at this position

          @param values list describing path in the tree
          @param path string containing path in the tree without the final leaf
        """
        
        path = metaline[0] + "/" + metaline[1]
        line = (path).replace("/", self._separator)\
                                  .split(self._separator)
        line.append(metaline[4])

        # if a path is already cached, append the node to it
        if path in self._cache:
            self._incremNodeCount()
            self._cache[path].append({"title": line[len(line) - 1],
                                     "unit": metaline[6],
                                     "key": line[len(line) - 1],
                                     "type": self._getDataType(path)})
        else:
            local_list = copy.deepcopy(line)
            for value in local_list:
                for element in self._tree:
                    if element['title'] == value:
                        local_list.pop(0)
                        self._recursTree(local_list,
                                         element["children"],
                                         path)
                        return True

            self._incremNodeCount()
            self._tree.append({"title": local_list[0],
                              "key": self._nodecount,
                              "unit": metaline[6],
                              "folder": True,
                              "children": []})

    def _fillTree(self):
        for line in self._hdffile.metadata:
            self._addValue(line)


class ExtractFormatter(object):
    """
      write extracted data in a file
    """
    def __init__(self, header = {}, separator=',', isdouble=False):
        # a list of pointers
        self.pointers = []
        self.export_names = []
        # a list of columns
        self.columns = []
        self.headerColumns = []
        # separator between values
        self.separator = separator
        # precision is .6 if False, .15 if True
        self.isDouble = isdouble
        self.header = header

    def addColumn(self, pointer, column, export_name = None):
        """
            add a column of data and its pointer
        """

        self.pointers.append(pointer)
        self.columns.append(column)
        self.export_names.append(export_name if export_name != None else pointer.name)

    def clearColumns(self):
        self.pointers = []
        self.columns = []

    def _getEmptyData(self, length):
        result = []
        for i in range(0, length):
            result.append('Not_In_HDF5')
        return result

    def _getArraySize(self):
        length = 0
        for i in range(0, len(self.columns)):
            if len(self.columns[i]) > 0:
                length = len(self.columns[i])
                break
        return length

    def getBufferedText(self, replace_ws=False):
        """
            write all data
        """
        try:
            f = StringIO()
            precision = "float"
            self._columnsAreWritable()
            data = []
            widths = []
            length = self._getArraySize()
            
            for entry in self.header.values():
                # 20201215: WARNING !!!
                # All the comment lines before the true 'header' (see
                # line1, line2, line3 below) MUST start with '# ' The
                # space after the '#' is very important to distinguish
                # those 'human targeted' header info from the machine
                # oriented one (le column names with their indice and unit)
                f.write("# %s : %s \n" % (entry['label'], entry['value']))
            for i in range(len(self.columns)):
                if len(self.columns[i]) == 0:
                    self.columns[i] = self._getEmptyData(length)
            
            for i in range(0, length):
                values = [column[i] for column in self.columns]
                towrite = []
                for value in values:
                    if not isinstance(value, str):
                        value = _getValue(value, precision)
                    widths.append(len(value))
                    towrite.append(value.strip())
                data.append(self.separator.join(towrite))

            if length > 1:
                # common case, several lines, several columns
                line1 = []
                line2 = []
                line3 = []
                for idx, pointer in enumerate(self.pointers):
                    if replace_ws:
                        line1.append('[' + str(idx + 1) + ']' +
                                        pointer.name.replace(" ", replace_ws))
                        line2.append('[' + str(idx + 1) + ']' +
                                        pointer.unit.replace(' ', replace_ws))
                    else:
                        line1.append('[' + str(idx + 1) + ']' + pointer.name)
                        line2.append('[' + str(idx + 1) + ']' + pointer.unit)

                    line3.append(('[' + str(idx + 1) + ']').rjust(widths[idx]))
                    # 20201215: WARNING !!!
                    # The 'machine' header lines below (line1, line2,
                    # line3) must start with a '#' withtout any space
                    # after (that is the text direclty after #).
                    header = "#" + self.separator.join(line1) + "\n#" + \
                                self.separator.join(line2) + "\n#" + \
                                self.separator.join(line3)[1:] + "\n"

                f.write(header)
                f.write("\n".join(data))
            else:
                # file with only one line and several columns
                # are plotted as one column with several lines
                f.write("#{:>6}{:>24}{:>26}  {}\n".format("index",
                                                    "value",
                                                    "unit",
                                                    "quantity"))
                for idx, value in enumerate(data[0].split(self.separator)):
                    f.write("{:>7}{:>24} # {:>23}  {}\n".format(idx,
                                                        value,
                                                        self.pointers[idx].unit,
                                                        self.export_names[idx]))
            return f
        except Exception as e:
            raise e

    def getVoTable(self):
        try:
            self._columnsAreWritable()
            table = vot.Votable()
            datatype = 'float'
            builder = vot.InfoBuilder()
            for key, value in self.header.items():
                table.addInfo(builder.withName(value['label'])
                                     .withValue(value['value'])
                                     .getInfo())
                builder.reset()
            builder = vot.FieldBuilder()
            for i in range(0, len(self.columns)):
                field = self.pointers[i]
                link = vot.LinkBuilder()\
                          .withContentRole('type')\
                          .withHref(field.skos)\
                          .getLink()
                table.addField(builder.withName(field.name)
                                      .withUnit(field.unit)
                                      .withUcd(field.ucd)
                                      .withUtype(field.utype)
                                      .withDatatype(datatype)
                                      .withLink(link)
                                      .getField())
                table.addColumn(self.columns[i])
                builder.reset()            
            return table

        except Exception as e:
            raise e

    def _columnsAreWritable(self):
        """
            check data validity
        """
        if len(self.columns) != len(self.pointers):
            raise Exception("columns and pointers do not match")

        if len(self.columns) > 0:
            size = self._getArraySize()
            for i in range(1, len(self.columns)):
                if len(self.columns[i]) > 0 and len(self.columns[i]) != size:
                    raise Exception('all the columns do not have the same size')
            return True
        raise Exception("no data")


def update_header(hdf5, header):
    try:
        # _c_ files contains the code version
        header["code version"] = {"value": hdf5.getCode(),
                                    "label": "Produced by"}
    except KeyError as e:
        # _s_ files does not contain the information
        header["code version"] = {  "value": "unknown", 
                                    "label": "Produced by"}
        

def export_dat_as_stringIO(hdf5, header,
                    columns, separator="   ", isdouble=False):
    factory = PointerFactory()
    w = ExtractFormatter(header, separator=separator, isdouble=isdouble)
    # add code version to header
    update_header(hdf5, header)

    for i in range(0, len(columns)):
        try:
            col_export_name = None
            if isinstance(columns[i], str):
                column = columns[i]
            else:
                column, col_export_name = columns[i]

            w.addColumn(hdf5.index[column],
                        hdf5.extractQuantity(hdf5.index[column]),
                        col_export_name)
        except Exception as e:
            print(e)
            pointer = factory.getErrorPointer(column)
            w.addColumn(pointer, [])
    
    try:
        return  w.getBufferedText()
    except Exception as e:
        print(e)
        raise e


def export_dat_as_file(hdf5, header,
                    columns, outputfile,
                    separator="   ", isdouble=False):
    """
        export selected data into text file
    """        
    
    try:
        # content to save in a StringIO
        buffered_result = export_dat_as_stringIO(hdf5, header, columns, separator, isdouble)
        with open(outputfile, "w") as f:
            buffered_result.seek(0)
            shutil.copyfileobj(buffered_result, f)

    except Exception as e:
        traceback.print_exc()
        print(e)
        raise e




def export_votable_as_StringIO(hdf5, header, columns, isdouble=False):
    """
        export selected data into votable
    """
    w = ExtractFormatter(header, isdouble=isdouble)
    # add code version to header
    update_header(hdf5, header)
    rejected = []
    for i in range(0, len(columns)):
        try:
            column = columns[i]
            w.addColumn(hdf5.index[column],
                        hdf5.extractQuantity(hdf5.index[column]))
        except Exception as e:
            rejected.append(column)

    if len(rejected) > 0:
        traceback.print_exc()
        raise Exception("Columns were rejected for extraction")
        #export_errors(outputfile, rejected)

    try:
        table = w.getVoTable()
        return table.toStringIO()
    except Exception as e:
        raise e

