import threading
import collections
import time
import re
import os.path

class ThreadSafeDict(collections.OrderedDict) :
    def __init__(self, * p_arg, ** n_arg) :
        dict.__init__(self, * p_arg, ** n_arg)
        self._lock = threading.Lock()

    def __enter__(self) :
        self._lock.acquire()
        return self

    def __exit__(self, type, value, traceback) :
        self._lock.release()



class ThreadSafeContainer():
    def __init__(self, max_size, life_time):
        """Controls content of a ThreadSafeDict by limiting its maximum size
        and setting a mamximum age to entries

        :param max_size: maximum number of elements in container
        :type max_size: int

        :param life_time: lifetime of an entry in the dict, in seconds
        :type life_time: int
             
        """
        self._dict = ThreadSafeDict()
        self._maxSize = max_size
        self._lifeTime = life_time

    def add(self, key, value):
        """Adds a key/value pair in dict
        if the dict is already full, oldest value is removed

        :param key: key of new entry
        :type key: string

        :param value: content to save in dict
        :type : value: object

        """

        # remove old entries
        self._purge_by_age()

        # remove oldest item if container is full
        if len(self._dict) >= self._maxSize:
            self._dict.popitem(last=False)

        self._dict[key] = {"value" : value, "time": time.time()}

    def get(self, key):
        """Returns value corresponding to key

        :param key: key for which a value is searched
        :type key: string

        """

        # remove old entries
        self._purge_by_age()
        try : 
            return self._dict[key]["value"]
        except KeyError as e:
            print("Entry {} not found in ThreadSafeContainer".format(key))
            raise e

    def _purge_by_age(self):
        """Removes entries that are older than maximum life time

        """
        now = time.time()

        # contains entries removed because too old
        toremove = []
        for key in self._dict.keys():
            age_of_entry = now - self._dict[key]["time"]
            if age_of_entry > self._lifeTime :
                toremove.append(key)
        
        for key in toremove:
            self._dict.pop(key, None)


def secure_filename(filename):
    """
    stolen from Werkzeug.utils.

    20210719: importing Werkzeug in macos (macport) causes the loading of
    extension in sqlite to crash. So we just pick the only one function we
    needed from it and included it here.
    """
    
    _filename_ascii_strip_re = re.compile(r'[^A-Za-z0-9_.-]')

    if isinstance(filename, str):
        from unicodedata import normalize
        filename = normalize('NFKD', filename).encode('ascii', 'ignore')
        filename = filename.decode('ascii')

    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, ' ')

    filename = str(_filename_ascii_strip_re.sub('', '_'.join(
                   filename.split()))).strip('._')

    return filename
