import math
import numpy as np

from idatplot import utils as iu

_qid = iu.qid


def _format_val(v):
    try:
        return float(v)
    except:
        return v


def format_data(data):
    h, u, d = data
    nd = [[_format_val(e) for e in r] for r in d]
    return (h, u, nd)



def _normalize_data(cf, data, axis):
    """ normalize the columns in data (<axis>) by the column with id
    `<axis>normalize` 
    
        ex of plot/y1normalize: 'n(H2)'        <--- as named in the conf file
       
    `axis` is either 'y1' or 'y2'
    
    """
    
    h, u, d = data
   
    cyidx = []
    cyncidx = None
    cfparam = axis + 'normalize'
    if cfparam in cf:
        ync = _qid(cf[cfparam])
        if ync in h:
            print('normalizing {} with {}'.format(axis, ync))
            cyidx = [h.index(y) 
                    for y in set([_qid(y) for y in cf[axis]]) & set(h)]
            cyncidx = h.index(ync)
        else:
            print("quantity {} doesn't exist, aborting {} \
                    normaization".format(ync, cfparam))
            return data

    nd = []
    for r in d:
        nr = [e for e in r]
        if cyidx and cyncidx is not None:
            # print("normalizing with {}({}):{}".format(cf[cfparam], cyncidx, r[cyncidx]))
            nr = [e/r[cyncidx] if i in cyidx else e for i, e in enumerate(nr)]
        nd.append(nr)

    return (h, u, nd)



def _log_large_data_range(cf, data, thresold = 10):

    def _log_data(data, qlist):
        h, u, d = data
        nd = []
        idxl = [h.index(q) for q in qlist]
        for r in d:
            nd.append([np.log10(e) if i in idxl else e 
                for i, e in enumerate(r)])

        return (h, u, nd)

    def _process_data(data, qlist):
        
        # identify max ymin/ymax range on the data (quantities in qlist)
        axpminy = float("Inf")
        axpmaxy = - float("Inf")
    
        yp = [iu.dcol(d, h, y) for y in qlist]
    
        for yd in yp:
            axpminy = min(axpminy, 
                    min([y for y in yd if y > 0], default = axpminy))
            axpmaxy = max(axpmaxy,
                        max([y for y in yd if y > 0], default = axpmaxy))
        
        drange = (math.floor(np.log10(axpmaxy)) + 1 
                - math.floor(np.log10(axpminy)))

        # print("drange: {}".format(drange)) 
        if drange >= thresold:
            return (True, _log_data(data, qlist), drange)
    
        return (False, data, drange)

    h, u, d = data
    ndata = (h, u, [r for r in d])

    if '_pp_log_large_data_range' not in cf:
        cf['_pp_log_large_data_range'] = []

    for a, qlist in [(a, [_qid(ae) for ae in cf.get(a, [])]) for a in ('y1', 'y2')]:

        # we don't do any auto y log if user defined y boundaries
        if (any([p in cf for p in (a+'min', a+'max')]) 
                or a not in cf.get('log', [])):
                print('{} auto log skipped: boundaries set in conf or log not asked'.format(a))
                continue

        sqlist = set(qlist) & set(h)
        if not sqlist:
            continue
        processed, ndata, _ = _process_data(ndata, sqlist)
        if processed: 
            print('{} automatically applied log (value range too large)'.format(a))
            cf['_pp_log_large_data_range'].append(a)

    return ndata


def _data_fact(cf, data, axes):
    h, u, d = data

    x, y1, y2 = axes
    y1f = [y for y in y1 if y[3] != 1]
    y2f = [y for y in y2 if y[3] != 1]

    nd = []
    if y1f or y2f:
        yfidx = [(h.index(_qid(y)), y[3]) for y in y1f + y2f]
        for r in d:
            nr = [e for e in r]
            for i, fact in yfidx:
                nr[i] = fact * nr[i]
            nd.append(nr)
    else:
        nd = d
        
    return (h, u, nd)


def init_pprocs(cf):
    pprocs_map = (
            (None, lambda d, axes: format_data(d)), 
            (None, lambda d, axes: _data_fact(cf, d, axes)),
            ('y1normalize', lambda d, axes: _normalize_data(cf, d, 'y1')),
            ('y2normalize', lambda d, axes: _normalize_data(cf, d, 'y2')),
            (None, lambda d, axes: _log_large_data_range(cf, d)))
            
    return [p for k, p in pprocs_map if k is None or k in cf]

def apply_pprocs(d, axes, pprocs):
    
    if not pprocs:
        return d

    nd = pprocs[0](d, axes)
    return apply_pprocs(nd, axes, pprocs[1:])


