# -*- coding: utf-8 -*-

import argparse
import re
import io
import os
import datetime as dt
import collections
import functools

import numpy as np
import matplotlib as mpl

mpl.use('Agg')

from matplotlib import pyplot as plt

plt.style.use('default') # Reset  settings to defaults in case of crappy local user settings

from matplotlib import pyplot as plt

from idatplot import utils as iu
from idatplot import dataprocs as pdp
from idatplot.extract import extract as pe

ismstyle = {
    ### Custom matplotlib parameters ###
    "font.family"           : "sans-serif",
    "font.size"             : 16,
    "mathtext.default"      : "regular",
    "lines.linewidth"       : 2,     # line width in points
    "axes.formatter.limits" : (-6, 6),
    "axes.labelsize"        : 16,
    "axes.titlesize"        : 20,
    "xtick.top"             : True,   # draw ticks on the top side
    "xtick.bottom"          : True,   # draw ticks on the bottom side
    "xtick.major.size"      : 8,      # major tick size in points
    "xtick.minor.size"      : 5,      # minor tick size in points
    "xtick.major.width"     : 2,    # major tick width in points
    "xtick.minor.width"     : 1,    # minor tick width in points
    "xtick.labelsize"       : 16, # fontsize of the tick labels
    "xtick.direction"       : "in",    # direction: in, out, or inout
    "xtick.major.top"       : True,   # draw x axis top major ticks
    "xtick.major.bottom"    : True,   # draw x axis bottom major ticks
    "xtick.minor.top"       : True,   # draw x axis top minor ticks
    "xtick.minor.bottom"    : True,   # draw x axis bottom minor ticks
    "ytick.left"            : True,   # draw ticks on the left side
    "ytick.right"           : True,  # draw ticks on the right side
    "ytick.major.size"      : 8,      # major tick size in points
    "ytick.minor.size"      : 5,      # minor tick size in points
    "ytick.major.width"     : 2,    # major tick width in points
    "ytick.minor.width"     : 1,    # minor tick width in points
    "ytick.labelsize"       : 16, # fontsize of the tick labels
    "ytick.direction"       : "in",    # direction: in, out, or inout
    "ytick.major.left"      : True,   # draw y axis left major ticks
    "ytick.major.right"     : True,   # draw y axis right major ticks
    "ytick.minor.left"      : True,   # draw y axis left minor ticks
    "ytick.minor.right"     : True,   # draw y axis right minor ticks
    "legend.fontsize"       : 14
}

mpl.rcParams.update(**ismstyle)


# 2018.05.31 @ dlan

DEFAULT_SEP = '   '
COLORS = [
"#FF4040","#50789F","#60BF00","#FF9F40","#BF60BF","#DF2080","#60BFBF","#2080DF","#BF6000","#606060","#9F9F9F","#BFBFBF"]
AUTOXMIN_THRESOLD = 0.015
AUTOXMIN_SHIFT = - 0.1


DEFAULT_IMG_SIZE = (7, 6)  # <- ~ 640x480
# DEFAULT_IMG_SIZE = (12, 9)  # <- ~ 800x600
DEFAULT_IMG_RES = 240 # DPI
DEFAULT_OFF = ['png', 'pdf']

_qid = iu.qid
_ql = iu.qlabel
_dcol = iu.dcol 

def _dunit(h, u, c):
    return u[h.index(c)]


def _flatten(l):
    return [e for sl in l for e in sl]


def _gen_colors(cnt, clist = None, cmap = mpl.cm.terrain):
    if clist and cnt <= len(clist):
        # cm_subsection = [int(i) for i in np.linspace(0, len(clist) - 1 , cnt)]
        return clist[:cnt]
    else:
        cm_subsection = np.linspace(0, 0.9, cnt)
        return [cmap(c) for c in cm_subsection]


def _stripcol(data, c):
    """ strip column with id `c` from data """
    h, u, d = data
    sc_idx = h.index(_qid(c))
   
    sh = [e for i, e in enumerate(h) if i != sc_idx]
    su = [e for i, e in enumerate(u) if i != sc_idx]
    sd = [e for r in d for i, e in enumerate(r) if i != sc_idx]

    return (sh, su, sd)

def _get_output(of, default_off = DEFAULT_OFF):
    """ handle the format of the 'plot/output' config that can be:
    
    - filename
    - (filename, [file format 1, file format 2, ...])

    The second option allows to generate the plot in different format, for
    example png, pdf.

    If first option, the default file format (DEFAULT_OFF) are generated
    """
    off_l = [] # output file format list
    ofn = of
    
    if of:
        if isinstance(of, list):
            ofn, off_l = of

    if not off_l:
        off_l = default_off

    return (ofn, off_l)


def parse_data(data_filepath, filters = [], sep = DEFAULT_SEP):

    def _process_field_header(fh):
        fidx, fname = fh
        return (fname.strip(), int(fidx) - 1)

    def _parse_headers(l):
        rh = re.findall(r'\[([1-9][0-9]*)\]([^[]+)', l[1:]) 
        # l[1:] because we skip the leading # 
        return [_process_field_header(fh) for fh in rh]

    def _parse_data(l, idx = None):
        dl = [e.strip() for e in l.split(sep)]
        if idx:
            return [dl[i] for i in idx]
        return dl

    def _skip_human_header(f):
        l = f.readline()
        while l.startswith("# "):
            l = f.readline()
        return l

    def _check_notinhdf5(l):
        return [i for i, e in enumerate(l) if e != 'Not_In_HDF5']

    headers = []
    data = []

    with open(data_filepath) as df:
        hl = _skip_human_header(df)
        headers = _parse_headers(hl)
        units = _parse_headers(df.readline())
        df.readline() # I don't know what is the 3rd line...

        # filter Not_in_HDF5 columns/quantities
        dl =  _parse_data(df.readline())
        clean_idx = set(_check_notinhdf5(dl))

        if filters:
            fidx = set([i for i, h in enumerate(headers) if h[0] in filters])
            clean_idx = clean_idx & fidx

        headers = [headers[i] for i in clean_idx]
        units = [units[i] for i in clean_idx]

        data.append([dl[i] for i in clean_idx])
        for l in df:
            dl = _parse_data(l, clean_idx)
            data.append(dl)

    headers = [k for k, i in headers]
    units = [k for k, i in units]

    return (headers, units, data)


# plotting sub system
# -------------------

### Interesting read for sci plotting

# https://stackoverflow.com/questions/4700614/how-to-put-the-legend-out-of-matplotlib-plot/27355247#27355247


# https://stackoverflow.com/questions/37538236/matplotlib-set-x-y-label-in-one-line


def to_bytes(fig, format_, release = True):

   # return an image file of the plot as a byte stream
   bio = io.BytesIO()
   fig.savefig(bio, format = format_, dpi = fig.dpi)

   if release is True:
       plt.close(fig)
   # 20141006
   # used to "release" the Figure object and free the pyplot cache.
   # This has been revealed by warning message :
   # /usr/lib/python3/dist-packages/matplotlib/pyplot.py:412:
   # RuntimeWarning: More than 20 figures have been opened. Figures
   # created through the pyplot interface (`matplotlib.pyplot.figure`)
   # are retained until explicitly closed and may consume too much
   # memory. (To control this warning, see the rcParam
   # `figure.max_num_figures`).  max_open_warning, RuntimeWarning)

   return bio
   


def basic_plotter(runid, x, y1, y2 = [], cf = {}, data = None):

    # help for placing legend on top of the plotting area:
    # https://stackoverflow.com/questions/4700614/how-to-put-the-legend-out-of-matplotlib-plot#4701285

    legend_ncol = 4 # initial try for number of columns in legend
    retry = True

    while (retry):
        # print("try plotting with legend_ncol=",legend_ncol)
        # fig = plt.figure(figsize = (12, 9), dpi = 100)
        fig = plt.figure(figsize = DEFAULT_IMG_SIZE, dpi = DEFAULT_IMG_RES)
        ax = fig.add_subplot(111)

        lines = []
        q_cnt = len(y1) + len(y2)
        colors = _gen_colors(q_cnt, COLORS)
   
        xl, xu, xd = x
        xid, xlabel, *rest = xl

        # plot the first y axis
        for i, y in enumerate(y1, start = 0):
            yl, yu, yd = y
            yid, ylabel, ylinestyle, _ = yl
            lines += ax.plot(xd, yd, label = "{}".format(ylabel),
                       color = colors[i],
                       linestyle = ylinestyle,
                       linewidth = 2)

        ax.set_ylabel(cf.get('y1label', 'n/a'))
    
        # plot the second y axis if available, y2
        if y2:
            ax2 = ax.twinx()
        
            for i, y in enumerate(y2, start = 0):
                yl, yu, yd = y
                yid, ylabel, ylinestyle, _ = yl
                lines += ax2.plot(xd, yd,
                            linestyle = ylinestyle,
                            label = "{}".format(ylabel),
                            color = colors[len(y1) + i],
                            linewidth = 2)
            ax2.set_ylabel(cf.get('y2label', 'n/a'))

        llabels = [l.get_label() for l in lines]
        legd = fig.legend(lines, llabels, loc = 8,
                ncol = legend_ncol,
                framealpha = 0) # Franck doesn't like the frame !

        # ### set log scale

        if 'log' in cf:
            if 'y1' in cf['log']:
                if 'y1' not in cf.get('_pp_log_large_data_range', []):
                    ax.set_yscale('log')
                    minl = mpl.ticker.LogLocator(base = 10.0,
                            subs = (0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9))
                    ax.yaxis.set_minor_locator(minl)
                else:
                    ylab = cf.get('y1label', '')
                    ax.set_ylabel('log(y1)' if not ylab else ylab)

            if y2 and 'y2' in cf['log']:
                if 'y2' not in cf.get('_pp_log_large_data_range', []):
                    ax2.set_yscale('log')
                    minl = mpl.ticker.LogLocator(base = 10.0,
                            subs = (0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9))
                    ax2.yaxis.set_minor_locator(minl)
                else:
                    ylab = cf.get('y2label', '')
                    ax2.set_ylabel('log(y2)' if not ylab else ylab)

            if 'x' in cf['log']:
                ax.set_xscale('log')

        # ### x, y1, y2 limits

        dxmin = next(x for x in sorted(xd) if x > 0)
        dxmax = max(xd)
        if cf.get('xmin', dxmin) == 'auto':
            xmin = iu.auto1xmin(cf, data,
                    x = xid,
                    axis_filter = [_qid(yl) for yl,_,_ in (y1 + y2)],
                    thresold = AUTOXMIN_THRESOLD,
                    shift = AUTOXMIN_SHIFT)
        else:
            xmin = float(cf.get('xmin', dxmin))
   
        xl = ax.set_xlim(xmin or dxmin, float(cf.get('xmax', dxmax)))

        ylim = (float(cf['y1min']) if 'y1min' in cf else None,
                float(cf['y1max']) if 'y1max' in cf else None)
        y1l = ax.set_ylim(ylim)

        if y2:
            y2lim = (float(cf['y2min']) if 'y2min' in cf else None,
                     float(cf['y2max']) if 'y2max' in cf else None)
            y2l = ax2.set_ylim(y2lim)

        # ### global figure settings

        ax.set_xlabel(cf.get('xlabel', xlabel))
        # ax.set_xlabel('{}'.format(xlabel))
        # ttl = ax.set_title(cf.get('title', 'no title').format(runid))
        ttl = ax.set_title(
                re.sub(r'MODELNAME', runid, cf.get('title', 'no title')))

        ttl.set_position([.5, 1.025])

        fig.tight_layout()
        # 1 line 0.15, 2: 0.153, 3: 0.156...
        fig.subplots_adjust(bottom = 0.2 + 0.04 * ((q_cnt - 1) // legend_ncol))

        # ### Check wether legend is overflowing
        
        plt.draw()
        renderer = fig.canvas.renderer
        box1 = fig.get_window_extent(renderer)
        box2 = legd.get_window_extent(renderer)
        
        if ((box2.bounds[0]<0. or box2.bounds[0]+box2.bounds[2]>box1.bounds[0]+box1.bounds[2]) and legend_ncol > 1) :
            legend_ncol -= 1
            retry = True
        else :
            retry = False

    return (fig, ax)


def make_plots(cf, runid, data, plotter = basic_plotter):

    for plotcf in cf['plot']:

        x = plotcf['x']
        y1 = plotcf['y1']
        y2 = plotcf['y2']

        # Discard any plot having one of its Yaxis quantity all negative
        h, u, d = pdp.format_data(data)
        cy1 = [y for y in y1 if _qid(y) in h]
        cy2 = [y for y in y2 if _qid(y) in h]
        
        if not (cy1 + cy2):

            print("not any valid y quantity in ({}), passing...".format(
                ', '.join([_ql(q) for q in (y1 + y2)])))
            continue

        y1p = [(y, _dunit(h, u, _qid(y)), _dcol(d, h, _qid(y))) for y in cy1]
        y2p = [(y, _dunit(h, u, _qid(y)), _dcol(d, h, _qid(y))) for y in cy2]
        ny = np.array([c for _, _, c in y1p] + [c for _, _, c in y2p])
        if not np.any(ny[ny >= 0]):
            print("all the quantities ({}) have negative values, passing..."\
                    .format(', '.join([_qid(y) for y in (cy1 + cy2)])))
            continue

        pprocs = pdp.init_pprocs(plotcf)
        pdata = pdp.apply_pprocs(data, (x, y1, y2), pprocs)
        h, u, d = pdata
        
        cy1 = [y for y in y1 if _qid(y) in h]
        cy2 = [y for y in y2 if _qid(y) in h]

        xp = (x, _dunit(h, u, _qid(x)), _dcol(d, h, _qid(x)))
        y1p = [(y, _dunit(h, u, _qid(y)), _dcol(d, h, _qid(y))) for y in cy1]
        y2p = [(y, _dunit(h, u, _qid(y)), _dcol(d, h, _qid(y))) for y in cy2]

        fig, ax = plotter(runid, xp, y1p, y2p, plotcf, pdata)

        yield (fig, ax, plotcf.get('output', None))
   

def api_client(cfstrio, datafile, outdir, txt = False):

    def _preprocess_quantities(cf):
   
        def _process_quantity(q):
            """
            q can be either:
            
            1. qid
            2. (qid, q user name)
            3. (qid, q user name)
            4. (qid, q user name, line_style)
            4. (qid, q user name, line_style, factor)
            """

            default_linestyle = '-'
            default_fact = 1
    
            rmap = {
                0: lambda x: (default_linestyle, default_fact),
                1: lambda x: (x[0], default_fact),
                2: lambda x: (x[0], x[1])
                }
    
            if isinstance(q, str):
                return (q, q, default_linestyle, default_fact)
        
            yid, ylabel, *rest = q
            ylinestyle, yfact = rmap[len(rest)](rest)
    
            return (yid, ylabel, ylinestyle, yfact)

        def _complete(y, nexty):
            return y + [_process_quantity(nexty)]

        for plotcf in cf['plot']:
            plotcf['x'] = _complete([], plotcf['x'])[0]
            plotcf['y1'] = functools.reduce(_complete, plotcf.get('y1', []), [])
            plotcf['y2'] = functools.reduce(_complete, plotcf.get('y2', []), [])
    
        return cf

    
    def _extract_hdf5(datafile, quantities, outdir, ofname = None):
        if not ofname:
            dfroot, _ = os.path.splitext(os.path.basename(datafile))
            ofname = dfroot
            
        header = {
                'hdffile': {
                    'label': 'Model name',
                    'value': ofname
                    },
                'date': {
                    'label': 'Extraction date',
                    'value': dt.datetime.now().strftime('%d-%m-%Y')
                    }
                }

        ofname += '.dat'
        ofpath = outdir + '/' + ofname

        # build .txt file with hdf5 extracted data
        print('creating {} from hdf5 file...'.format(ofpath))
        pe.extract(datafile,
                list(quantities),
                ofpath,
                header,
                sep = DEFAULT_SEP)
        return ofpath

    dfroot, _ = os.path.splitext(os.path.basename(datafile))
    _q = iu.plot_quantities
    cf = dict(iu.load_conf(cfstrio))
    cf = _preprocess_quantities(cf)

    # trick to keep only once each quantity and keep it ordered like in the
    # yaml conf file (previously we used a set which doesn't keep the
    # order...)
    oc = collections.OrderedDict([(_qid(q), None) 
        for pcf in cf['plot'] for q in _q(pcf)])
    quantities = oc.keys()

    if not txt:
        for pcf in cf['plot']:
            ofn, _ = _get_output(pcf.get('output', ''))
            ofname, _ = os.path.splitext(os.path.basename(ofn))
            ofname = re.sub(r'MODELNAME', dfroot, ofname)
            df = _extract_hdf5(datafile, quantities, outdir,
                    ofname = ofname if ofname else None)
        datafile = df

    data = parse_data(datafile, filters = quantities)
    # from now, the data is considered quite clean (cleaning done in parse_data) 
    # This means, for example, that not existing/invalid quantities are
    # discarded, i.e are not in the `data` object.

    h, _, _ = data

    iq = set(quantities) - set(h)
    if iq:
        print('**Invalid quantities** (discarded): {}'.format(', '.join(iq)))
   
    if set([_qid(pcf['x']) for pcf in cf['plot']]) & iq:
        print('some x quantities are not in the hdf5, quitting...')
        return
    
    if not h:
        print('Not any valid quantity in the conf file, quitting...')
        return

    figs_l = []
    for fig, ax, of in make_plots(cf, dfroot, data):

        ofn, off_l = _get_output(of)
        if not ofn:
            ofn = dfroot

        # ofn = ofn.format(dfroot)
        ofn = re.sub(r'MODELNAME', dfroot, ofn)

        bin_figs = {}

        for off in off_l:
            bf = to_bytes(fig, format_ = off)

            bin_figs[off] = bf

            plot_path = outdir + '/' + ofn + '.' + off
            print("writing {}".format(plot_path))
            with open(plot_path, 'wb') as f:
                f.write(bf.getvalue())

        figs_l.append((fig, bin_figs))

    return figs_l 


def cmdline_client(cffile, datafile, outdir, txt = False):
    return api_client(iu.load_stringio(cffile), datafile, outdir, txt)



def main():
    
    parser = argparse.ArgumentParser()
    parser.add_argument('-txt', action='store_true')
    parser.add_argument('cffile')
    parser.add_argument('datafile')
    parser.add_argument('outdir')
   
    args = parser.parse_args()
   
    if not args.txt and not iu.valid_h5file(args.datafile):
        print('invalid hdf5 file: {}'.format(args.datafile))
        return


    cmdline_client(args.cffile, args.datafile, args.outdir, args.txt)


if __name__ == '__main__':
    main()
