import math
import io
import functools
import os

import yaml 
import numpy as np
import matplotlib as mpl
import h5py


def valid_h5file(filepath):
    try:
        f = h5py.File(filepath, 'r')
    except:
        return False
    return True

def load_stringio(filepath):
    siof = io.StringIO()
    with open(filepath) as _f:
        siof.write(_f.read())
    return siof

def load_conf(cf_stringio, components = ('extract', 'plot')):

    def _unns(document, ns):
        return {k.replace(ns+'/', ''):v for k, v in document.items()}

    def _components(cfd, c):
        for d in cfd:
            for k, v in d.items():
                if k.startswith(c):
                    yield _unns(d, c)
                break

    cfdocuments = list(yaml.safe_load_all(cf_stringio.getvalue()))

    return [(c, list(_components(cfdocuments, c))) 
            for c in components]


# ### extract utilities

# #### processor for specie command (scommand)

def _process_pdr_lines_scommand(cfdb, c):
    """
    :c lines <specie> <wl_inf> <wl_sup> <name format>
            example
            :c lines h2 13 56.5 f2

            with
            fX: format:
                    f0: preflabel
                    f1: I(H2 v=0,J=18->v=0,J=16 angle 60 deg)
                    f2: I(H2 3.5469 micrometres) angle 60 deg
                    f3: Line intensity of H2 3.5469 micrometres angle 60 deg
    """

    def _valid_pdr_lines_command(c):
        return len(c.split()) == 6

    def _get_lines(dbfile, specie, wlinf, wlsup, fmt):
        import sqlite3
        
        sqlqtpl = """
        select s.wl, s.preflabel, a.label 
        from species s, alt_labels a
        where s.id=a.line 
            and a.format = ?
            and s.wl >= ? and s.wl <= ?
        order by s.wl
        """
        res = []     

        conn = sqlite3.connect(dbfile)
        with conn:
            c = conn.execute(sqlqtpl, (fmt, wlinf, wlsup))
            res = c.fetchall()

        return res

    if not _valid_pdr_lines_command(c):
        return None

    _, cmd, specie, wlinf, wlsup, name_fmt = c.split()
   
    appath = os.path.dirname(os.path.realpath(__file__))
    dbfile = '/'.join([appath, cfdb['idatplot/pdr_lines_db']])
    res = _get_lines(dbfile, specie, wlinf, wlsup, name_fmt)
    
    return [(pl, l) for wl, pl, l in res]
    # return [pl for wl, pl, l in res]

def _is_scommand(s):
    return s.startswith(":c")

def process_ecf_quantities(cfdb, ql):

    def _process_scommand(cfdb, c):
        _, cmd, *_ = c.split()
        uc_processors = cfdb['idatplot/scommand_processors']
        if cmd in uc_processors:
            return globals()[uc_processors[cmd]](cfdb, c)

    def _complete(y, nexty):
        if not _is_scommand(nexty):
            return y + [nexty]
 
        return y + _process_scommand(cfdb, nexty)

    return functools.reduce(_complete, ql, [])




# ### common plot utils

def plot_quantities(plot_cf):
    q = [plot_cf['x']] + plot_cf['y1'] + plot_cf.get('y2', [])
    for a in ('y1', 'y2'):
        if a + 'normalize' in plot_cf:
            q += [plot_cf[a + 'normalize']]
   
    return q

def qid(q):
    if isinstance(q, str):
        return q.strip()
    return q[0].strip()

def qlabel(q):
    return q[1]

def dcol(d, h, c):
    col_idx = h.index(c)
    return [row[col_idx] for row in d]





# ### data utils

def max1stvar(plotcf, data, thresold, x = 'AV', axis_filter = []):
  
    """ returns (xt, <quantity id>, <quantity y value at xt>) 
    
        <quantity id> is the first quantity (relative to x) to have a value variation greater than
        *thresold*. The (x, y) point coresponding to the reaching of the
        *thresold* is given by (xt, <quantity y value at xt>).
    """



    def _v(v):
        return float(v)

    def _prepare_data(pcf, data):
        """ if user asked for logscale, use the log of the values of the
        log-scaled axes to compute auto xmin """

        h, u, d = data
        # 20190202: ugly hack: the first point is often messy... so skip it.
        nd = np.array(d[1:])
        # nd = np.array(d)

        y1, y2 = [[qid(q) for q in pcf.get(a, [])] for a in ('y1', 'y2')]
        cy1 = [y for y in y1 if y in h]
        cy2 = [y for y in y2 if y in h]

        cy1il = list(set([h.index(y) for y in cy1]))
        cy2il = list(set([h.index(y) for y in cy2]))
        for a, ail in zip(('y1', 'y2'), (cy1il, cy2il)):
            if (a in pcf['log'] 
                    and a not in pcf.get('_pp_log_large_data_range')):
                print("log10 {} values before computing autoxmin".format(a))
                nd[:, ail] = np.array(list(map(np.log10, nd[:, ail])))

        return (cy1, cy2, (h, u, nd.tolist()))


    def _ybound(data, boundf, qil):
        """ bound function 'boundf' is either np.amin or np.amax
            'qil' is a list of columns index to consider (and only them)
        """
        h, u, d = data
        dnpa = np.array(d)
        yb = boundf(dnpa, axis = 0)
        yb = [b for i, b in enumerate(yb) if i in qil]
        # print("boundf: {}".format(boundf(yb)))

        return boundf(yb)

    cy1, cy2, pdata = _prepare_data(plotcf, data)

    h, u, d = pdata
    xi = h.index(x)

    if axis_filter:
        cy1il = [h.index(q) for q in cy1 if q in axis_filter]
        cy2il = [h.index(q) for q in cy2 if q in axis_filter]
    else:
        cy1il = [h.index(y) for y in cy1]
        cy2il = [h.index(y) for y in cy2]
    
    cy1il = set(cy1il)
    cy2il = set(cy2il)

    r0 = d[0]


    my1 = tuple()
    deltay1 = _ybound(pdata, np.nanmax, cy1il) - _ybound(pdata, np.nanmin, cy1il)
    cy1thresold = thresold * deltay1
    # print('deltay1: {}, cy1thresold: {}'.format(deltay1, cy1thresold))
    for i in range(len(d)):
        for j in cy1il:
            v = abs(_v(d[i][j]) - _v(r0[j]))
            if v > cy1thresold:
                my1 = (_v(d[i][xi]), h[j], _v(d[i][j]))
                break
        if my1:
            break

    my2 = tuple()
    if cy2il:
        deltay2 = _ybound(pdata, np.nanmax, cy2il) - _ybound(pdata, np.nanmin, cy2il)
        cy2thresold = thresold * deltay2
        # print('deltay2: {}, cy2thresold: {}'.format(deltay2, cy2thresold))
        for i in range(len(d)):
            for j in cy2il:
                v = abs(_v(d[i][j]) - _v(r0[j]))
                if v > cy2thresold:
                    my2 = (_v(d[i][xi]), h[j], _v(d[i][j]))
                    break
            if my2:
                break

    if not my2:
        return my1
    if not my1:
        return my2

    return sorted([my1, my2], key = lambda x: x[0])[0]


def auto1xmin(cf, data, x = 'AV', axis_filter = [], thresold = 0.05, 
        algo = max1stvar, shift = - 0.1):

    """ find the most relevant xmin based on data shape (move xmin where
    things start to happen. )
    
    Only the columns in axis_filter are considered to find the best xmin.

    `thresold` is the amount of variation from x1.
    """
    
    def _shift_xmin(plotcf, x, xminv, data, shift = -0.05):
        """ 
        This function solves an issue raised by Emeric (who also provided the
        algo to solve it) on 20190213

         "Le graphe commence dès qu'on a une variation de 5%, mais du coup il
         ne montre pas du tout la partie constante qui était avant.  
         
         [...] 
 
         On voit qu'effectivement, l'algo a tronqué le graphe juste là où les
         variations commençaient (donc c'est bien la partie intéressante).
         J'ai vérifié sur un certain nombre d'autres cas, l'endroit de coupure
         a l'air d'être toujours pertinent. 

         Par contre, pour l'utilisateur lambda, ça n'est pas du tout clair que
         c'était constant avant, il voit un graphe qui démarre à un endroit où
         la quantité varie déjà beaucoup.

         Donc en fait il faudrait faire commencer le graphe un petit peu avant
         ce point de démarrage des variations, pour montrer une petite portion
         où les quantités sont constantes."
        
        """

        def _tx(xmin, xmax, log = False):
            if log:
                l10 = np.log10
                sxmin = np.power(10, l10(xmin) + shift * (l10(xmax) - l10(xmin)))
                # Cette deuxième formule vient du fait qu'on veut ajouter 10%
                # en taille apparente à l'écran, donc en log et non en
                # linéaire
            else:
                sxmin = xmin + shift * (xmax - xmin)
            return sxmin if sxmin > 0 else xmin
            
        h, u, d = data
        xi = h.index(x)

        nd = np.array(d)
        xmaxv = max(nd[:,xi])
        if 'x' in plotcf['log']:
            return _tx(xminv, xmaxv, log = True)
        return _tx(xminv, xmaxv)

    xv = None
    mv = algo(cf, data, thresold, x, axis_filter = axis_filter)
    # print("mv: {}".format(mv))

    if mv:
        xv, q, qv = mv
        if shift is not None:
            axv = _shift_xmin(cf, x, xv, data, shift = shift)
            print('adjusted xmin: {} -> {}'.format(xv, axv))
            xv = axv

    return xv

# ### plot utils


def compute_log_ticks_loc(vmin, vmax, thresold = 12):

    if abs(vmin) == float("Inf") or abs(vmax) == float("Inf"):
        return (mpl.ticker.LogLocator(), mpl.ticker.LogLocator())
    
    nticks = (math.floor(np.log10(vmax)) + 1 
            - math.floor(np.log10(vmin)))
    
    if nticks <= thresold:
        majl = mpl.ticker.LogLocator(base = 10, numticks = nticks)
        minl = mpl.ticker.LogLocator(base = 10.0, 
            subs = (0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9), 
            numticks = nticks)
    else:
        majl = mpl.ticker.LogLocator()
        minl = mpl.ticker.LogLocator()

    return (majl, minl)

