import os
import re
import sqlite3
import argparse


conf = {
        "k_B": 1.3807E-16,    # cgs
        "h": 6.6261E-27,    # cgs
        "c":2.99792458E10, # cgs
        "cm_to_microns": 1E4,

        "species" : [
            ("H2", "line_h2.dat"),
            ("H", "line_h.dat"),
            ("H3+", "line_h3p.dat")
            ],

        "line_file_adapters": {
            'H2': '_h2_lines_adapter'
        }
    }

def _fnull(l):
    return [e for e in l if e.strip()]

def _h2_lines_adapter(cf, specie, lines_filepath):
    """
    PDR line_X.dat format:
     # data1                        data2                 data3        
     # n nu nl E(K) Aij(s-1) quant: vu; Ju; vl; Jl; info: (description); lambda (micrometres)
     # 0,1, 2, 3,   4,       5,     6,  7,  8,  9,  10,                  11
    """

    with open(lines_filepath) as lf:
        lf.readlines(4)
        for line in lf:
            if line[0].startswith('#'):
                continue

            data1, rest = re.split('quant:', line)
            data2, data3 = re.split('info:', rest)
            n, nu, nl, ek, aij = [float(e) for e in data1.split()]
            vu, ju, vl, jl = [int(e) for e in _fnull(data2.split(';'))]
            desc, lambdadesc  = [e.strip() for e in data3.split(';')]

            T = float(ek)
            E = cf['k_B']*T
            nu1 = E/cf['h']
            wl = cf['c']/nu1
            wl *= cf['cm_to_microns']
            
            preflabel = "I(%s v=%i,J=%i->v=%i,J=%i angle 00 deg)" % (specie, vu, ju, vl, jl)
            
            f1 = preflabel
            f2 = "I(%s %f micrometres) angle 00 deg"  % (specie, wl)
            f3 = "Line intensity of %s %s micrometres angle 00 deg"  % (specie, wl)
           
            yield (specie, wl, preflabel, 
                    [("f0", preflabel), ("f1", f1), ("f2", f2), ("f3", f3)])

def extract_lines(cf, specie, lines_filepath):
    adapters = cf['line_file_adapters']
    if specie in adapters:
        return globals()[adapters[specie]](cf, specie, lines_filepath)

    print("no adapter found for {} lines file".format(specie))
    return []

def build_sql_params(cf, species_lines):
    species_sql = [] # id | line | wl | preflabel
    alt_sql = [] # line | format | label

    for i, sl in enumerate(species_lines, 1):
        specie, wl, preflabel, altlabels = sl
        alt_sql += [(i, fmt, label) for fmt, label in altlabels]
        species_sql += [(i, specie, wl, preflabel)]

    return (species_sql, alt_sql)

def build_lines_db(cf, dbfile, lines_data_dir, species):
    def _create_db(conn):
        conn.execute('create table species(id integer, line text, wl number, preflabel text);')
        conn.execute('create table alt_labels(line, format text, label text);')
    
    conn = sqlite3.connect(dbfile)
    with conn:
        _create_db(conn)

    with conn:
        for specie, specie_filename in species:
            print("building db for {}...".format(specie))
            specie_filepath = '/'.join([lines_data_dir, specie_filename])
            ssql, asql = build_sql_params(cf, 
                    extract_lines(cf, specie, specie_filepath))
            conn.executemany("insert into species (id, line, wl, preflabel) values (?,?,?,?)", ssql)
            conn.executemany("insert into alt_labels(line, format, label) values (?,?,?)", asql)
 





def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('lines_data_dir')
    parser.add_argument('dbfile')
    args = parser.parse_args()
 
    if os.path.isfile(args.dbfile):
        os.remove(args.dbfile)
    
    build_lines_db(conf, args.dbfile, args.lines_data_dir, conf['species'])

main()
