import os
import os.path
import uuid
from datetime import datetime
import traceback
import codecs
from io import StringIO

from lib import hdf
from lib import hdfutil
from server.lib import config as conf
from server.lib import logger as logger
from idatplot import plot as pdrplot


def extract(job):
    job_config = job

    header = {"hdffile":
              {"value": job_config['hdffile'],
               "label": "Model name"},
              "date": {"value": datetime.now().strftime('%d-%m-%Y'),
                       "label": "Extraction date"}
              }

    if job_config['dataformat'] not in  ["text", "votable"]:
        raise Exception("Unknown data format")

    try : 
        hdffile = hdf.PdrHDF(os.path.join(conf.DATA_DIRECTORY, job_config['hdffile']))
        if job_config['dataformat'] == "votable":
            buffered_result = hdfutil.export_votable_as_StringIO(hdffile, header,  job_config['quantities'])
        else : 
            buffered_result = hdfutil.export_dat_as_stringIO(hdffile, header, job_config['quantities'])

        # point at beginning of buffer for later reading
        buffered_result.seek(0)
        return buffered_result
        
    except Exception as e:
        raise e


def plot(job_config): 
    """
        @param job job parameters
    """   
    try:        
        yaml_content = job_config['plotconfiguration']['yaml']
        output = job_config['output']

        hdf_file = os.path.join(conf.DATA_DIRECTORY, job_config['hdffile'])
        yaml_str = codecs.decode(codecs.encode(yaml_content, 
                                                "utf-8", 
                                                'backslashreplace'),
                                                'unicode-escape')  

        # generate img from StringIO
        figs_l = pdrplot.api_client(StringIO(yaml_str),
                hdf_file,
                conf.OUTPUT_DIRECTORY)

        return {
                "image": output + ".png", 
                "figures": figs_l
                } 

    except Exception as e:
        traceback.print_exc()
        return {"error" : "error in execute_plot : {}".format(str(e))}



def remove_file(path):
    try:
        os.remove(path)
    except IOError as e:
        print('error removing {}: {}'.format(path, e))
        pass


def clean_data():
    """
        Remove all content from data and output directories
    """
    if conf.CLEAN_OUTPUT:
        print('cleaning output directory...')
        for f in os.listdir(conf.OUTPUT_DIRECTORY):
            if not f.startswith("README"):
                remove_file(os.path.join(conf.OUTPUT_DIRECTORY, f))

    if conf.CLEAN_USERDATA:
        print('cleaning userdata directory...')
        for f in os.listdir(conf.DATA_DIRECTORY):
            if not f.startswith("README"):
                remove_file(os.path.join(conf.DATA_DIRECTORY, f))




if __name__ == "__main__":
    pass
