import base64
import re
import os.path
import traceback
import uuid
import time
import json
from io import StringIO
from server.lib import logger as logger
from lib import hdf as hdf
from lib import hdfutil as hdfutil
from lib import util as ismutil
from ismvoca.lib.impl import sqlite as ia
from server.lib.config import *
from server.lib import yamlbuilder as yaml
import server.lib.bottle as bottle
from server.lib import exporter


server_app = bottle.Bottle()
server_app.secret_key = b'\x05\xf5\x9fsH\x0e\xb3\x9f\xb7\xa6\x05\x8dQ\x06\x18\xd8'
server_app.config['UPLOAD_FOLDER'] =  DATA_DIRECTORY
bottle.TEMPLATE_PATH.insert(0, TEMPLATE_DIRECTORY)


# list of opened hdf files
session_opened_files = ismutil.ThreadSafeContainer(100, CACHE_LIFETIME)
# string of yaml plot configurations
session_yaml_files = ismutil.ThreadSafeContainer(25, CACHE_LIFETIME)
# StringIO of text extractions
session_dat_files = ismutil.ThreadSafeContainer(50, CACHE_LIFETIME)
# StriongIO of votable extractions
session_vot_files = ismutil.ThreadSafeContainer(25, CACHE_LIFETIME)

IVPROTO = None

"""
def get_session_data(s_uuid):
    global session_opened_files
    session_data = server_cache["opened"][s_uuid]
    print("GET SESSION DATA")
    print(session_data)
    return session_data
"""

#@cache.cached(timeout=360000000, key_prefix='autompletion_vocab')
def load_concepts_file(code):
    """
        load the concepts file and store result into local cache
    """
    print('loading the ism-voca dictionnary for code {}...'.format(code))

    try:
        code_ident = next(c for c in ALLOWED_CODES if c in code.lower())
    except:
        print('code "{}" not supported by idat'.format(code))
        return None

    print(code_ident, code_ident in ISMVOCA_DICTS)
    if code_ident in ISMVOCA_DICTS:
        ivdbpath = os.path.join(ISMVOCA_DICTS_DIRECTORY, 
                ISMVOCA_DICTS[code_ident])
    else:
        return None

    try:
        if os.path.isfile(ivdbpath):
            ivproto = ia.make_protocol({'cdb': ivdbpath})
            return ivproto

    except Exception as e:
        logger.append_exception(e)
        print(e)
        raise e

    return None



def bad_request(message):
    """Handles request error
    :param message: message explaining the error

    """
    bottle.response.content_type = 'application/json'
    bottle.response.status_code = 400
    return json.dumps({'message': message})

@server_app.route(BASE_URL+"/")
def index():
    """Homepage of the service

    :return: main page
    :rtype: string
    """
    return "ISM service"
    
@server_app.route(BASE_URL+"/projectload/<project>/<model>", method="GET")
def projectload(project, model):
    """Load a file from its name and origin project
       Store it in session a correspondance between the file and
       a session id
    
    Used in web mode

    :param project: name of the project containing the datafile
    :type project: string

    :param model: name of the file to open in project containing the datafile
    :type model: string    

    :return: a map describing the structure of the file and the session id
    :rtype: json
    """
    return load(model, project)

@server_app.route(BASE_URL+"/localload/<filename>", method="GET")
def localload(filename):
    """Load a file only from its name ( in data directory)
       Stores it in session a correspondance between the file and
       a session id
       
       Used in local mode

    :param filename: name of the hdf file to open
    :type filename: string
    :return: a map describing the structure of the file and the session id
    :rtype: json
    """
    return load(filename)        
        
def load(model, project=None):
    """Load a hdf file and returns a summary of its content
    if project == None, we are in local mode. If project value is provided, we
    are in online mode and data are searched in the raw directory of the project

    :param model:  name of a model
    :type model: string

    :param project:  name of a project
    :type model: string

    :return: quantities contained in this model
    :rtype: json

    """
    global IVPROTO 
    global session_opened_files
    if project is None:
        path = os.path.join(DATA_DIRECTORY, model)
        filename = model
    else:
        path = os.path.join(DATA_DIRECTORY, project, "raw", model, model+"_s_20.hdf5")
        filename = os.path.join(project, "raw", model, model+"_s_20.hdf5")
    try:
        hdf_file = hdf.PdrHDF(path)
        tree = hdfutil.TreeManager(hdf_file)   
        session_id = str(uuid.uuid4())

        
        IVPROTO = load_concepts_file(hdf_file.getCode())



        # saves opened file name as a session var
        session_opened_files.add(session_id, filename)

        hdf_file.close()
        bottle.response.content_type = "application/json"
        return json.dumps({'tree': tree.tree,
                            "sessionId": session_id,
                            "xQuantities": hdf_file.xQuantities,
                            "xDefault": "AV"})

    except OSError as e:
        print(e)
        return bad_request("Unable to open {}".format(filename))   
    except Exception as e:
        print(e)
        return bad_request(e)   

            
def allowed_file(filename):    
    return '.' in filename and \
        filename.rsplit('.', 1)[1].lower() in ALLOWED_EXTENSIONS


@server_app.route(BASE_URL+'/import', method='POST')
def upload_file():
    """Imports a file in the data directory for use in local mode

    """
    # check if the post request has the file part
    localfilename = bottle.request.forms.get("lfname", "").strip()
    if not localfilename:
        if 'file' not in bottle.request.files:
            return "No file to upload"

        file_ = bottle.request.files['file']
        if file_ and allowed_file(file_.filename):
            try:
                filename = ismutil.secure_filename(file_.filename) 
                file_.save(os.path.join(DATA_DIRECTORY, filename), overwrite=True)
                print('{} saved in userdata folder'.format(os.path.join(DATA_DIRECTORY, filename)))
                localfilename = filename        
            except Exception as e:
                print(e)
        else:
            return "file type is not allowed"

    return bottle.redirect("http://localhost:{}{}/ui/index?model={}".format(
        SERVER_PORT, BASE_URL, localfilename),
        code = 302)


@server_app.route(BASE_URL+'/ui/import', method='GET')
def upload_file_get():
    f_l = os.listdir(DATA_DIRECTORY)
    return bottle.template('import', 
            f_l = f_l, 
            service_uri = SERVICE_URI,
            ismwebsite_uri = ISM_WEBSITE,
            allowed_ext = ALLOWED_EXTENSIONS)


def export_as_file(request, exportformat):
    """Export some quantities from an opened file, writes them in session

    :param request: received request object
    :type request: object

    :param exportformat: format in which data will be exported 
    :type exportformat: str

    :return: extraction result and status
    :rtype: json
    """
    global session_dat_files
    session_id = request.json["sessionId"]
    try : 
        parameters = {"parameters" : request.json["quantities"],
                    "datatype": request.json["datatype"],
                    "files": (session_opened_files.get(session_id),)}
        config = build_extract_configuration(exportformat, parameters)
        buffered_file = exporter.extract(config)        
        session_dat_files.add(session_id, buffered_file)
        bottle.response.content_type = "application/json"
        # format is text or votable
        return json.dumps({"status": "OK", "result" : session_id, "format" : exportformat})
    except KeyError as e:
        return json.dumps({"status": "ERROR", "error" : "Your session has expired"})
    except Exception as e:
        return json.dumps({"status": "ERROR", "error" : str(e)})


@server_app.route(BASE_URL+"/export/data/text", method="POST")
def export_as_text():
    """Export data in text format

    :param json: data to export
    :type json: json
    """
    return export_as_file(bottle.request, "text")


@server_app.route(BASE_URL+"/export/data/votable", method="POST")
def export_as_votable():
    """Export data in votable format

    :param json: data to export
    :type json: json
    """
    return export_as_file(bottle.request, "votable")


@server_app.route(BASE_URL+"/autocomplete/<usersearch>", method="GET")
def autocomplete(usersearch):
    """Endpoint providing autocompletion
    :param usersearch: a text input for which we will provide suggestions
    :type usersearch: string
    """
    global IVPROTO

    if IVPROTO == None :
        raise Exception("The ism-voca protocol has not been initialized")

    try:
        ac = IVPROTO("search", usersearch, limit = SUGGEST_RESULT_LIMIT)
        result = [(cname, IVPROTO('preflabel', cid)) 
                for cname, cid in ac['results']]
    except Exception as e:
        # raise
        print(e)
        result = []

    bottle.response.content_type = "application/json"
    return json.dumps(result)


@server_app.route(BASE_URL+"/generate/plot", method="POST")
def plot():
    """Plot data

    :param json: data to export
    :type json: json
    """
    bottle.response.content_type = "application/json"
    try:
        plot_config = bottle.request.json["parameters"]
        session_id = bottle.request.json["sessionId"]

        if len(plot_config['y1']) > 0:
            config = build_plot_configuration(plot_config,
                                              (session_opened_files.get(session_id),), session_id)

            result = exporter.plot(config)
            figure, bfigures = result['figures'][0]
            bimg = base64.b64encode(
                    bfigures['png'].getvalue()).decode('ascii')

            return json.dumps({
                "status": "OK", 
                "pdf" : re.sub('.png', '.pdf', result['image']),
                "bimg": bimg
                })
        else:
            return json.dumps({"message": "Y1 list is empty",
                                "status": "ERROR"})
    # session_id does not exist in session_opened_files
    except KeyError as e:
        traceback.print_exc()
        if "figures" not in result :
            return json.dumps({"message": "An error occurred while generating images.", "status": "ERROR"})
        else : 
            return json.dumps({"message": "An error occurred, your session may have ended, please reload the page.", "status": "ERROR"})

    except Exception as e:
        logger.append_info(e)
        traceback.print_exc()
        return json.dumps({"message": "error in plot route : {}".format(str(e)), "status": "ERROR"})


def build_extract_configuration(exportformat, params):
    """Returns an export configuration object

    :param exportformat: format of the result file
    :type exportformat: str

    :param params: extraction parameters
    :type params: object

    :return: configuration object
    :rtype: dict
    """
    quantities = params["parameters"]
    datatype = params["datatype"]
    hdffile = params["files"][0]
    if "output_dir" in params :
        output_dir = params["output_dir"]
    else : 
        output_dir = None

    config = {
        'dataformat': exportformat,
        'datatype': datatype,
        'quantities': quantities,
        'output_path': output_dir,
        'hdffile': hdffile
    }
        
    return config


def build_plot_configuration(params, hdffiles, session_id):
    """Returns a plot configuration object

    :param params: extraction parameters
    :type params: object

    :param hdffiles: a list of hdffiles to process
    :type hdffiles: list

    :param session_id: user session_id
    :type session_id: string

    :return: configuration object
    :rtype: dict
    """


    def add_limit(conf_dict, limit, params):
        try:
            if params[limit].strip() != '':
                conf_dict[limit] = params[limit]
        except Exception as e:
            logger.append_exception(e)
    
    f = hdffiles[0]

    output = f.replace('.hdf5', "").replace('/', '_')

    if 'y2' in params:
        y2 = params['y2']
        for val in y2:
            val.append('--')            

    plotconfiguration = {
        "x": params['x'],
        "y1": params['y1'],
        "y2": y2,
        "xlabel": params['xlabel'],
        "y1label": params['y1label'],
        "y2label": params['y2label'],
        "log": params['log'],
        "title": params['title'],
        "output": output
    }

    add_limit(plotconfiguration, 'xmin', params)
    add_limit(plotconfiguration, 'xmax', params)
    add_limit(plotconfiguration, 'y1min', params)
    add_limit(plotconfiguration, 'y1max', params)
    add_limit(plotconfiguration, 'y2min', params)
    add_limit(plotconfiguration, 'y2max', params)

    yaml_content = yaml.generate_yaml(plotconfiguration)

    config = {
        'hdffile': f,
        'plotconfiguration': {
            'yaml' : yaml_content
        },
        'output': output         
    }
    
    # save yaml content in user session
    session_yaml_files.add(session_id, yaml_content)

    return config


"""
    !!! functions to use only in local environment
    !!! to remove when hosting on a web server
"""

# @server_app.route(BASE_URL+'/html/<filename>')
# def serve_html(filename):
#     if filename in ("help.html",):
#         return bottle.static_file(filename, root=CLIENT_DIRECTORY)
#     else:
#         return "Unknown page"

@server_app.route(BASE_URL+'/ui/<tpl>')
def serve_tpl(tpl):
    if tpl in ("help", "index"):
        print('serving {}...'.format(tpl))
        return bottle.template(tpl, 
                service_uri = SERVICE_URI,
                ismwebsite_uri = ISM_WEBSITE)
    else:
        return "Unknown page"

def get_file_response(byte_data, filename, content_type):
    """returns response for routes returning data

    :param byte_data: binary content to return 
    :type byte_data: byte

    :param filename: name of the file that will be showed to user
    :type filename: string

    :param content_type: content type of HTTP response
    :type content_type: string

    """
    return bottle.LocalResponse(
        body=byte_data,
        headers={
            "Content-Disposition": "attachment; filename={}".format(filename),
            "Content-Type": content_type,
        }
    )

def get_file_error(message):
    """returns error message for routes returning data

    :param message: error message
    :type message: string

    """
    return bottle.LocalResponse(
                body= message,
                headers={
                    "Content-Type": "text/plain",
                }
            )

@server_app.route(BASE_URL+'/generate/yaml/<sessionId>', method="GET")
def get_yaml_file(sessionId):
    global session_opened_files, session_yaml_files
    try : 
        hdf_filename = session_opened_files.get(sessionId).split(".")[0]
        file_content = session_yaml_files.get(sessionId).replace(hdf_filename, 'MODELNAME')

        byte_data = StringIO(file_content).read()
        
        return get_file_response(byte_data, YAML_TEMPLATE_NAME, "application/yaml")       

    except KeyError as e :
        return get_file_error("Data not found, your session may have expired.")
    

@server_app.route(BASE_URL+'/generate/text/<sessionId>', method="GET")
def get_text_file(sessionId):
    global session_opened_files, session_dat_files
    try : 
        hdf_filename = session_opened_files.get(sessionId).split(".")[0]
        byte_data = session_dat_files.get(sessionId).read()
        return get_file_response(byte_data, hdf_filename, "text/plain")
    except KeyError as e :
        return get_file_error("Data not found, your session may have expired.")
  

@server_app.route(BASE_URL+'/generate/votable/<sessionId>', method="GET")
def get_text_file(sessionId):
    global session_opened_files, session_dat_files
    try : 
        hdf_filename = session_opened_files.get(sessionId).split(".")[0]
        byte_data = session_dat_files.get(sessionId).read()
        return get_file_response(byte_data, hdf_filename, "application/xml")
    except KeyError as e :
        return get_file_error("Data not found, your session may have expired.")


@server_app.route(BASE_URL+'/output/<filename>')
def serve_static(filename):
    """Only pdf files are still written on disk

    """

    # filename starts by uuid
    parts = filename.split("_")

    # remove uuid from proposed file name
    if not any([ext in filename for ext in ('.pdf',)]):
        parts.pop(0)    

    headers = {
            'Content-Disposition': 'attachment; filename="{}"'.format("_".join(parts)),
            'Cache-Control': 'no-cache'
            }

    return bottle.static_file(
            filename, 
            root = OUTPUT_DIRECTORY,
            headers = headers)



@server_app.route(BASE_URL+'/ui/static/css/<filename>')
def css(filename):
    return bottle.static_file(filename, root=CLIENT_DIRECTORY + "/static/css")

@server_app.route(BASE_URL+'/ui/static/css/commons/<filename>')
def css_commmons(filename):
    return bottle.static_file(filename, root=CLIENT_DIRECTORY + "/static/css/commons")

@server_app.route(BASE_URL+'/ui/static/js/<filename>')
def js(filename):
    return bottle.static_file(filename, root=CLIENT_DIRECTORY + "/static/js")

@server_app.route(BASE_URL+'/ui/static/img/<filename>')
def js(filename):
    return bottle.static_file(filename, root=CLIENT_DIRECTORY + "/static/img")


def start():

    exporter.clean_data() 

    bottle.run(server_app, 
            reloader = True, #False,#True,
            debug = True,#False,
            host = SERVER_HOST, 
            port = SERVER_PORT, 
            server = "cheroot")   


if __name__ == '__main__':
    start()

