import re
import sqlite3
import sys
import argparse
import itertools

import sqlite_spellfix

from ismvoca.lib import utils as u
import ismvoca


def _sanitize_fts_uquery(uq):
	search_table_tokensep = r'[()\-]'
	return re.sub(search_table_tokensep, " ", uq)

def associate_ident(gconn, term):
    # print('associate ident: ', term)
    q = "select ident from dict_entries where value=?"
    ident = None
    with gconn() as conn:
        r = conn.execute(q, (term,)).fetchone()
        if r:
            ident = r[0]
        else:
            ident = None

    return ident

def associate_preflabel(gconn, tident):
    q = "select value from dict_entries where ident=? and att='pref'"

    with gconn() as conn:
        r = conn.execute(q, (tident,)).fetchone()
        if r:
            pref = r[0]
        else:
            pref = None

    return pref


def spellfix(gconn, search_query, splitq = False, top = 5):

    # base_spellfix = """
    # select :term{0} as term, word, editdist3(:term{0}, word) as ed3
    # from spellfix1data 
    # where word match :term{0}
    #       and ed3 in (select editdist3(:term{0}, word) as e3 
    #                   from spellfix1data where word match :term{0} 
    #                   order by e3 limit 1)
    # """

    base_spellfix = """
    select :term{0} as term, word, min(editdist3(:term{0}, word)) as ed3, matchlen
    from spellfix1data 
    where word match :term{0} and top=:top group by matchlen 
    """

    # base_spellfix1 = """
    #     SELECT :term{0} as term, word, editdist3(:term{0}, word) FROM spellfix1data
    #     WHERE word MATCH :term{0} and top=1
    # """


    spquery = base_spellfix
    # if top1:
    #     spquery = base_spellfix1

    terms = [search_query]
    if splitq != False:
        terms = search_query.split(splitq)

    # print('terms: ', terms)
    params = {"term{}".format(i): t for i, t in enumerate(terms, 1)}
    params['top'] = top
    # print("params:",params)
    query = " UNION ".join(
        [spquery.format(i + 1) for i in range(len(params) - 1)]
    )

    query += "order by ed3"
    
    # print("query: ", query)
    # correction_map = {}
    corrections_map = {}
    with gconn() as conn:
        c = conn.execute(query, params)
        # correction_map = dict(c)
        for t, w, ed, wl in c.fetchall():
            if t not in corrections_map:
                corrections_map[t] = []
            # print(w, ed, wl)
            corrections_map[t].append(w)

    print(search_query, corrections_map)
    # return [correction_map.get(t, t) for t in terms]
    return [" ".join(p) for p in itertools.product(*corrections_map.values())]


def search(gconn, search_query, limit = 20, add_sp = True, sptop = 5):
 
    def _query(gconn, q, limit = 10):
        # print(q, type(q))
        # fts_query = "SELECT description FROM fts4data2 WHERE fts4data2 MATCH ? limit ?"
        # fts_query = """
        # SELECT min(description) FROM fts4data2 WHERE fts4data2 MATCH ?
        # group by length(description) limit ?"""
        fts_query = """
        SELECT description FROM fts4data2 WHERE fts4data2 MATCH ?
        order by length(description) limit ?"""

        res = []
        with gconn() as conn:
            c = conn.execute(fts_query, 
	    (_sanitize_fts_uquery(q), limit))
            res = c.fetchall()
        return [r[0] for r in res]
   
    def _sample(l, cnt):

        lsize = len(l)
        if lsize <= cnt:
            return l
        step = int(lsize/cnt)
        return [l[i] for i in range(0, lsize, step)]

    res = _query(gconn, search_query + "*", limit - 2 * sptop)

    limremaining = limit - len(res)
    print('res before sp: ', res) 
    spres = []
    if add_sp:
        for correction in spellfix(gconn, search_query, top = sptop):
            print("correction0: ", correction)
            r = _query(gconn, correction)
            if r:
                spres += r[:]

        spres = _sample(spres, limremaining / 2)

        spres2 = []
        for correction in spellfix(gconn, search_query, 
                splitq = " ", top = sptop):
            print("correction: ", correction)
            r = _query(gconn, correction)
            if r:
                spres2 += r[:]

        spres2 = _sample(spres2, limremaining / 2)
       
        res += spres + spres2

    # res = sorted(res, key = lambda x: len(x))
    
    return {
        "terms": search_query,
        "results": [(t, associate_ident(gconn, t)) for t in u.rmdup(res)]
    }


def make_protocol(cf):


    def _getconn():

        conn = None

        def __getconn(conn):
            def _initconn(conn):
                conn.enable_load_extension(True)
                conn.load_extension(sqlite_spellfix.extension_path())
                # conn.load_extension(cf.get('spellfixpath', spellfixpath))
                return conn

            if cf.get('connpool', False):
                if conn == None:
                    conn = _initconn(
                            sqlite3.connect(
                                cf['cdb'],
                                check_same_thread = cf.get('check_same_thread', True)))
                return conn

            return _initconn(sqlite3.connect(cf['cdb']))

        return __getconn(conn)

    # import importlib.resources as pkg_resources
    # spellfixpath = ""
    # with pkg_resources.path('ismvoca.resources.sqlite', 'spellfix.so') as p:
    #     spellfixpath = str(p)

    _protocol = {
            'spellfix': (lambda q, splitq = False, top = 5:
                spellfix(_getconn, q, splitq, top = top)),
            'search': (lambda q, add_sp = True, sptop = 3,
                limit = 15:
                search(_getconn, q, limit, add_sp, sptop)),
            'ident': lambda term: associate_ident(_getconn, term),
            'preflabel': lambda tident: associate_preflabel(_getconn, tident)
            }

    def _proxy(f, *args, **kwargs):
        if f in _protocol:
            return _protocol[f](*args, **kwargs)
        return None

    return _proxy





