import time
import collections
import itertools
import re
import pickle
import os

import yaml


from ismvoca.lib import utils as u

DEFAULT_RADIX_CNT = 14
DEFAULT_KMATCH_CNT = 3


_dg = u.dg

# db layer
# -------

def load_concepts_db(fp, yl_adapter):

    """ 
    returns dict(chname:cident)
    """

    st = time.time()
    conceptsdb =  []

    with open(fp) as f:
        conceptsdb = [(
            b.get('ID', None), 
            [b.get('PREF', None)] + b.get('ALT', [])
            )
            for b in yl_adapter.safe_load_all(f) if b]
        conceptsdb = list(conceptsdb)

    _dg("time to load concepts db file: {}s".format(time.time() - st))
    
    # we return inverted hash index
    # WARNING !! This assumes a given hname map to only 1 ident...
    # return { hname: ident for ident, hnames in qnamesdb for hname in hnames}
    invidx = { hname: ident 
            for ident, hnames in conceptsdb if ident is not None 
            for hname in hnames if hname is not None}
    conceptsdb = None # purge memory

    # debug
    memfoot = 0
    import sys
    for k, v in invidx.items():
        memfoot += sys.getsizeof(k) + sys.getsizeof(v)
    print("memory for concepts db: {}".format(memfoot/1024/1024))

    return invidx


def load_concepts_meta_db(fp, yl_adapter, atts = None):

    """
    atts: list of the documents attributes to load. If None, all the atts will
    be loaded.
 
    returns dict(cident:dict(att:value))

    """
    st = time.time()
    cmeta_db =  []

    with open(fp) as f:
        for b in yl_adapter.safe_load_all(f):
            if atts:
                cmeta_l = [(a, b.get(a, None)) for a in attrs]
            else:
                cmeta_l = [(a, b[a]) for a in b.keys() if a != 'ID']
        cmeta_db.append((b.get('ID', None), cmeta_l))
        

    _dg("time to load the concept meta db file '{}': {}s".format(
        fp,
        time.time() - st))
    
    return { 
            ident: dict(m)
            for ident, m in cmeta_db if ident is not None 
            }


def cdb_hnames(conceptsdb):
    return list(conceptsdb.keys())


def build_nghash(concepts, hf = None):
    ngdb = {} 
    # import shelve
    # ngdb = shelve.open('ngdb.shelf', flag='n')
    st = time.time()
    cnt = 1
    try:
        for qn in concepts:
            for ng in u.ngrams(qn):
                if ng not in ngdb: ngdb[ng] = set([])
                ngdb[ng].add(qn)
                cnt += 1
    except Exception as e:
        print('exception while building nghash: ', e)
        print("len ngdb: {}, {}", len(ngdb), cnt)
        raise

    _dg('time to build ngdb: {}s'.format(time.time() - st))

    if hf is not None:
        with open(hf, 'w') as of:
            for k, v in ngdb.items():
                of.write('{}\t{}\n'.format(k, ', '.join(v)))

    return ngdb




# text processing
# ---------------


def likelyhood_ss_sort(mqname_l, qname, case_sensitive = False):
    # We compute p(mqname is substr(sqname match)) to identify even better
    # candidates
    mqname_lp = [1] * len(mqname_l)
    n = max(map(len, mqname_l))
    re_flags = re.IGNORECASE
    if case_sensitive:
        re_flags = 0
    cregex = re.compile(re.sub(r'(\\ )+', r'.{1,2}', re.escape(qname)),
            flags = re_flags)
    for i, v in enumerate(mqname_l):
        # we allow distances of 2 chars max in place of whitespace
        mo = cregex.search(v)
        if mo is not None:
            spos, _ = mo.span()
            # exact pattern appearing sooner is moore likely.
            mqname_lp[i] += (1 - spos / n)

    s = sum(mqname_lp)
    mqname_lp = [p/s for p in mqname_lp]
    
    return [k for k, v in sorted(
        zip(mqname_l, mqname_lp), 
        key = lambda x: x[1], 
        reverse = True)]


def likelyhood_freq_sort():
    
    def build_p(words):
        n = len(words)
        def p(word): 
            return words[word] / n
        return p

    def _words(text): return re.findall(r'\w+', text.lower())

    def words(text): return collections.Counter(_words(text))

    return None


# def candidates(word): 
#     return (known([word]) or known(edits1(word)) or known(edits2(word)) or [word])
# 
# 
# def correction(word): 
#     return max(candidates(word), key=P)



def spellfix(ngdb, qname, 
        radix_cnt,
        keep_match_cnt,
        shorter_candidates,
        case_sensitive,
        limit = 10):
  
    lsssort = mk_cf_likelyhood_ss_sort()
    st = time.time()
    qng = u.ngrams(qname)
    _dg(qng)

    # count how many ngrams each candidate have in common with qname
    mqnamec = collections.Counter(
            [qn for ng in qng if ng in ngdb for qn in ngdb[ng]])
    # We should include a filter on the len of the qn here (too much shorter
    # from the len of q should be discarded) ? We would keep any one bigger though.
    smqnamec = sorted(mqnamec.items(), key = lambda x: x[1], reverse = True)

    sbestmatch = itertools.groupby(smqnamec, lambda x: x[1])
    # we keep only the qnames having the most ngrams in common
    # `keep_match_cnt` allow to keep not only the candidates with the most
    # ngrams in common with the q to spell but also to pick candidates in less
    # good candidates (having -1 ngram in commoin, -2, etc... up to
    # -`keep-match-cnt`)
    mqname_l = []
    for i in range(keep_match_cnt):
        try:
            k, v = next(sbestmatch)
        except:
            break
        mqname_l += [qn for qn, cnt in v]

    if not mqname_l:
        return []

    # group by radix
    st = time.time()
    if radix_cnt is not None:
        def _group(k, cnt = radix_cnt):
            return k[0:cnt]
        # _dg([(k, list(v)) for k, v in itertools.groupby(mqname_l, _group)])
        mqname_l = sorted(set(mqname_l), key = len)
        mqname_l = [likelyhood_ss_sort(list(v), qname, case_sensitive)[0] 
                for k, v in itertools.groupby(mqname_l, _group)]

    ls_mdname_l = likelyhood_ss_sort(mqname_l, qname, case_sensitive)
    _dg(ls_mdname_l[0:10])
    print('time to group by radix: {}'.format(time.time() - st))

    # TODO here; add another likelyhood processing based on likelyhood of
    # substring based on the whole corpus of qnames, for example, H2 is more
    # likely than Fe
    # Caution: this will shuffle the 'sort by qname len' sorting done above...
    # which may be more relevant
    #
    # likelyhood_freq_sort(ls_mdname_l, qname)


    if shorter_candidates:
        mqname_l = sorted(set(mqname_l), key = len)
        lsl_mdname_l = (likelyhood_ss_sort(list(v), qname, case_sensitive)[0] 
            for k, v in itertools.groupby(mqname_l, lambda x:len(x)))
        ls_mdname_l = list(set(
            ls_mdname_l[0:int(limit/2)] + 
            u.take(lsl_mdname_l, int(limit/2))))
    else:
        ls_mdname_l = ls_mdname_l[0:limit]

        
    _dg(mqname_l[0:10])
    _dg('time to spellfix: {}s'.format(time.time() - st))
    return ls_mdname_l

def _sample_compress(sample, size):
    nc = []
    ssize = len(sample)
    inc = int(ssize / size + 1)
    for i in range(0, ssize, inc):
        nc.append(sample[i])

    return nc





# ### concepts filters 


def cf_startswith(conceptsdb, ngdb, q, candidates):
    # !!!!!!!!!!!!! O(n) !!
    return [c for c in candidates if c.startswith(q)]

def cf_contains(conceptsdb, ngdb, q, candidates, ignore_case = True):
    if ignore_case:
        ql = q.lower()
        return [c for c in candidates if ql in c.lower()]
    else:
        return [c for c in candidates if q in c]


def mk_cf_sample_len(cnt = 1):

    def _cf_sample_len(conceptsdb, ngdb, q, candidates):
        nc = []
        for g, v in itertools.groupby(sorted(candidates, key = len), 
                lambda x: len(x)):
            nc += _sample_compress(list(v), cnt)
        return nc

    return _cf_sample_len

def mk_cf_sample_compress(size):
    def _cf_sample_compress(conceptsdb, ngdb, q, candidates):
        return _sample_compress(candidates, size)

    return _cf_sample_compress


def mk_cf_likelyhood_ss_sort(
        strict_substr = True,
        case_sensitive = False,
        grouper = lambda x: x[1],
        group_filter = lambda x: sorted([q for q, p in x], key = len),
        group_filter_compress_size = 3,
        gfilter_1grp = False):

    def _cf_likelyhood_ss_sort(conceptsdb, ngdb, q, candidates):
        # We compute p(candidate|q substr location) to identify even better
        # candidates.
        # a candidates having subs q sooner are better
        clikelyhood = [1] * len(candidates)
        # n = max(map(len, candidates))

        re_flags = re.IGNORECASE
        if case_sensitive:
            re_flags = 0

        sre = re.escape(q)
        if not strict_substr:
            # we allow distances of 2 chars max in place of whitespace
            sre = re.sub(r'(\\ )+', r'.{1,2}', sre)

        cregex = re.compile(sre, flags = re_flags)

        for i, v in enumerate(candidates):
            mo = cregex.search(v)
            if mo is not None:
                spos, _ = mo.span()
                # exact pattern appearing sooner is moore likely.
                clikelyhood[i] = (1 - spos / len(v))

        s = sum(clikelyhood)
        clikelyhood = [p/s for p in clikelyhood]
        intc = sorted(zip(candidates, clikelyhood), 
                key = lambda x: x[1], 
                reverse = True)
        
        if grouper:
            nintc = []
            gintc = itertools.groupby(intc, grouper)
            for i, gr in enumerate(gintc):
                g, v = gr
                if i == 0 and not gfilter_1grp:
                    # it may be useful to not filter the first group (the most
                    # liklely candidate based on position of substring)
                    # because it may contain the candidates STARTING WITH the
                    # substring q (in this case p = 1 - 0/something = 1).
                    nintc += [q for q, p in v]
                else:
                    nintc += _sample_compress(group_filter(list(v)),
                        group_filter_compress_size)

            intc = nintc
        else:
            intc = [k for k, v in intc]

        return intc

    return _cf_likelyhood_ss_sort


def search(conceptsdb, ngdb, q, cfilters, limit = 10):
    """ return list of (hname, ident) for each concept """

    def _process_q(filters, conceptsdb, ngdb, q, candidates = []):
        if not filters:
            return candidates

        nc = filters[0](conceptsdb, ngdb, q, candidates)

        return _process_q(filters[1:], conceptsdb, ngdb, q, nc)

    candidates = _process_q(cfilters, conceptsdb, ngdb, q, cdb_hnames(conceptsdb))[:limit]

    return filter(None, [(c, conceptsdb.get(c, None)) for c in candidates])



def make_protocol(cf):

    bincdb = cf.get('bincdb', 'conceptsdb.pickle')

    if bincdb and os.path.exists(bincdb):
        with open(bincdb, 'rb') as dbf:
            conceptsdb = pickle.load(dbf)
    else:
        conceptsdb = load_concepts_db(cf['cdb'], yaml)
        with open(bincdb, "wb") as dbf:
            pickle.dump(conceptsdb, dbf)
 
    ngdb = build_nghash(cdb_hnames(conceptsdb), hf = None)

    _protocol = {
        'spellfix': (lambda q, 
            radix_cnt = DEFAULT_RADIX_CNT,
            keep_match_cnt = DEFAULT_KMATCH_CNT,
            shorter_candidates = False,
            case_sensitive = False,
            limit = 10: 
            spellfix(ngdb, q, radix_cnt, keep_match_cnt,
                shorter_candidates, case_sensitive, limit)),
        'search': (lambda q, cfilters, limit = 10:
            search(conceptsdb, ngdb, q, cfilters, limit))
            }

    def _proxy(f, *args, **kwargs):
        if f in _protocol:
            return _protocol[f](*args, **kwargs)
        return None

    return _proxy




