! Copyright (C) 2002 - Observatoire de Paris/Meudon
!                      Universite Paris 7
!                      CNRS (France)
!                      Contact : Jacques.Lebourlot@obspm.fr
!
! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License version 2
! as published by the Free Software Foundation.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

! JLB - 22 May 2008

MODULE PXDR_EXCITH2

  PRIVATE

  PUBLIC :: EQLH2V, COLH2V

CONTAINS

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
SUBROUTINE EQLH2V (ttry, iopt)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   USE PXDR_FLAGS
   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PXDR_AUXILIAR
   USE PXDR_STR_DATA
   USE PXDR_RF_TOOLS

   ! Computes equilibrium matrix of H2 levels and solves for detailed balance
   IMPLICIT NONE

   REAL (KIND=dp), INTENT (IN)                  :: ttry
   INTEGER, INTENT (IN)                         :: iopt

   REAL (KIND=dp), DIMENSION (:,:), ALLOCATABLE :: fr_des
   REAL (KIND=dp), DIMENSION (:),   ALLOCATABLE :: fr_des_m
   INTEGER                                      :: j_sp
   INTEGER                                      :: nused  ! number of H2 levels used in computation
   INTEGER                                      :: nrxi   ! number of reactions with H2
   INTEGER                                      :: mm     ! index of grain sizes
   INTEGER                                      :: nrendo ! index of endothermal reaction
   INTEGER                                      :: if_ac
   INTEGER                                      :: if_110h2
   INTEGER                                      :: k, k1, k2, k3
   INTEGER                                      :: lev, ij, iv, levu, iju, ivu
   INTEGER                                      :: levJ0 ! index of v=0, J=0
   INTEGER                                      :: levJ1 ! index of v=0, J=1
   REAL (KIND=dp)                               :: toto
   REAL (KIND=dp)                               :: ratefgr ! formation   rate on grains (old style)
   REAL (KIND=dp)                               :: ratefLH ! formation   rate by Langmuir-Hinshelwood
   REAL (KIND=dp)                               :: ratefER ! formation   rate by Eley-Rideal
   REAL (KIND=dp)                               :: ratefga ! formation   rate in gas phase
   REAL (KIND=dp)                               :: ratedes ! desorption  rate from grains
   REAL (KIND=dp)                               :: ratedga ! destruction rate in gas phase
   REAL (KIND=dp)                               :: ratedph ! destruction rate by photons
   REAL (KIND=dp)                               :: rateads ! adsorption  rate on grains

   REAL (KIND=dp)                               :: endo_rxn, rat_l, taux, dumy
   REAL (KIND=dp)                               :: tt, t3  ! gas temperature (max 10 K, normalized to 300 K)
   REAL (KIND=dp), DIMENSION (0:14)             :: kvh2

                                                ! Scaling factor applied to Herraez-Aguilar C+ + H2 rates
                                                !         to match Zanchet et al. rates
   REAL (KIND=dp)                               :: scale_rot
   REAL (KIND=dp)                               :: scale_vib

   !-----------------------------------------------------------------------------
   ! Prepare some quantities
   !-----------------------------------------------------------------------------
   j_sp   = in_sp(i_h2)
   nused  = spec_lin(in_sp(i_h2))%use
   levJ0  = lev_h2(0,0)
   levJ1  = lev_h2(0,1)
   nrendo = nh2for + ncrion + npsion + nasrad + nrest + 1

   IF ( .NOT. ALLOCATED(fr_des)) THEN
      ALLOCATE (fr_des(1:nused,npbgr))
      ALLOCATE (fr_des_m(1:nused))
   ENDIF
   fr_des(:,:) = 0.0_dp
   fr_des_m(:) = 0.0_dp

   timh21(iopt) = 0.0_dp ! H2 form. rate by gas phase chemistry [cm-3 s-1]
   timh22(iopt) = 0.0_dp ! H2 dest. rate by gas phase chemistry [cm-3 s-1]
   timh23(iopt) = 0.0_dp ! H2 form. rate by LH or old style     [cm-3 s-1] - used to compute heating rate by H2 formation on grains
   timh24(iopt) = 0.0_dp ! H2 dest. rate by photons             [cm-3 s-1]
   timh25(iopt) = 0.0_dp ! H2 form. rate by Eley-Rideal process [cm-3 s-1] - used to compute heating rate by H2 formation on grains

   if_ac    = 0        ! 0 for accretion, 1 for rejection
   if_110h2 = 0        ! 0 for destruction of h2, 1 for collision partner
   !--- Rates of various processes in cm-3 s-1 ---
   ratefgr  = 0.0_dp   ! H2 formation rate on grains (physisorption)
   ratefLH  = 0.0_dp   ! H2 formation rate by Langmuir-Hinshelwood (physisorption)
   ratefER  = 0.0_dp   ! H2 formation rate by Eley-Rideal (chemisorbed + physisorbed)
   ratefga  = 0.0_dp   ! Total H2 formation rate in gas phase
   ratedes  = 0.0_dp   ! Total H2 desorption from grains (thermal + non-thermal)
   ratedph  = 0.0_dp   ! H2 photo-destruction rate in gas phase
   ratedga  = 0.0_dp   ! Destruction rate in gas phase (without specific cases)
   rateads  = 0.0_dp   ! H2 adsorption rate on grains

   create(1:nused) = 0.0_dp  ! formation   rate in each level      [cm-3 s-1]
   remove(1:nused) = 0.0_dp  ! destruction rate in each level      [cm-3 s-1]

   !----------------------------------------------------------------------------
   ! SUMMARY OF H2 Excitation prescriptions
   !----------------------------------------------------------------------------
   ! FORMATION:
   ! * Old style:               (  0)     ratefgr    pophm1
   ! * Langmuir-Hinshelwood:    (211)     ratefLH    pophm1
   ! * Eley-Rideal chemisor:    ( 20)     ratefER    pophm2
   ! * Eley-Rideal physisor:    (113)     ratefER    pophm2
   ! * CR-photons gr desorp:    (116/117) ratedes    pophm1
   ! * Therm. desorption:       (118)     ratedes    fr_des_m (scaled Boltzmann distribution at Tgr)
   ! * Gas phase                (DEFAULT) ratefga    2/3 J=0, 1/3 J=1
   !----------------------------------------------------------------------------
   ! DESTRUCTION
   ! * Photodissociations:      (  7)     ratedph    pdh2vJ
   ! * endothermal reactions:   (  6)     ratedga    rate specific to H2 levels
   ! * C+ + H2 -> CH+ + H       ( 61)     ratedga    rate specific to H2 levels
   ! * S+ + H2 -> SH+ + H       ( 62)     ratedga    rate specific to H2 levels
   ! * collisions with H/H2:    (110)     ratedga    rate specific to H2 levels
   ! * H2 sticking grains:      (113)     rateads    rate is the same for all levels
   ! * Gas phase:               (DEFAULT) ratedga    rate is the same for all levels
   !----------------------------------------------------------------------------

   !----------------------------------------------------------------------------
   ! H2 desorption from grains
   ! * controls H2 ortho / para conversion on grains
   ! * follows a Boltzman distribution at grains temperatures
   !----------------------------------------------------------------------------
   DO mm = 1, npbgr ! Loop on grain sizes (without PAH part)
      toto = 0.0_dp
      DO lev = 1, nused
         fr_des(lev,mm) = spec_lin(j_sp)%gst(lev) &
                        * EXP(-spec_lin(j_sp)%elk(lev) / tdust(mm,iopt))
         toto = toto + fr_des(lev,mm)
      ENDDO
      fr_des(:,mm) = fr_des(:,mm) / toto  ! normalization
   ENDDO
   ! Sum on grain sizes
   toto = 0.0_dp
   DO lev = 1, nused
      fr_des_m(lev) = SUM(fr_des(lev,1:npbgr) * wwmg(1:npbgr)) / SUM(wwmg(1:npbgr))
      toto = toto + fr_des_m(lev)
   ENDDO
   fr_des_m(:) = fr_des_m(:) / toto

   !----------------------------------------------------------------------------
   ! Fill create and remove matrixes
   !----------------------------------------------------------------------------

   nrxi = xmatk(i_h2)%nbt ! number of chemical reactions with H2

   DO k = 1, nrxi
      k1 = xmatk(i_h2)%fdt(k)%r1  ! index of reactif 1
      k2 = xmatk(i_h2)%fdt(k)%r2  ! index of reactif 2
      k3 = xmatk(i_h2)%fdt(k)%r3  ! index of reactif 3

      SELECT CASE (xmatk(i_h2)%fdt(k)%it)  ! Reaction type

      !===================================================================================
      !=== H2 coming from grains (formation and desorption)
      !===================================================================================

      !.........................................................................
      ! (0) H2 formation on grains - Old style / R = 3*10-17 * sqrt(T/100)
      !     Note: chemistry file must not have both type 0 and type 211
      !.........................................................................
      CASE (0)
         ! NOTE: previously we recomputed taux. It is not exactly the same value than
         !       xmatk(i_h2)%fdt(k)%k
         !taux    = gamm(nh2for) * densh(iopt) * (ttry / bet(nh2for))**alpha(nh2for) &
         !        * (g_ratio / 0.01_dp) / (istiny + ab(i_h))                ![cm+3 s-1]
         ratefgr = xmatk(i_h2)%fdt(k)%k * ab(k1) * ab(k2)                   ![cm-3 s-1]
         !--- Excitation: use pophm1
         DO lev = 1, nused
            ij = njl_h2(lev)
            iv = nvl_h2(lev)
            dumy = 0.0_dp
            DO levu = nused+1, nlevh2
               iju = njl_h2(levu)
               ivu = nvl_h2(levu)
               dumy = dumy + pophm1(ivu,iju) * cashm(levu,lev)
            ENDDO
            create(lev) = create(lev) + (dumy + pophm1(iv,ij)) * ratefgr
         ENDDO

      !.........................................................................
      ! (211) H2 formation on grains - Langmuir-Hinshelwood
      !       Note: chemistry file must not have both type 0 and type 211
      !       There are mm rxns to sum over grain size distribution
      !.........................................................................
      CASE (211) ! Formation on grains
         ratefLH = xmatk(i_h2)%fdt(k)%k * ab(k1) * ab(k2)                  ![cm-3 s-1]
         !--- Excitation: use pophm1
         DO lev = 1, nused
            ij = njl_h2(lev)
            iv = nvl_h2(lev)
            dumy = 0.0_dp
            DO levu = nused+1, nlevh2
               iju = njl_h2(levu)
               ivu = nvl_h2(levu)
               dumy = dumy + pophm1(ivu,iju) * cashm(levu,lev)
            ENDDO
            create(lev) = create(lev) + (dumy + pophm1(iv,ij)) * ratefLH
         ENDDO
         timh23(iopt) = timh23(iopt) + ratefLH

      !.........................................................................
      ! (20) H2 formation on grains - Eley-Rideal with chemisorbed species
      !.........................................................................
      CASE (20) ! H + H:: -> H2 (ER with chemisorbed species)
         ratefER = ratefER + xmatk(i_h2)%fdt(k)%k * ab(k1) * ab(k2)       ![cm-3 s-1]
         !--- Excitation ---
         ! create is filled after ratefer has been completed with ER formation
         ! involving physisorbed species (type 113)

      !.........................................................................
      ! (116 / 117) H2 desorption from grains by cosmic rays (116) and photons (117)
      !.........................................................................
      CASE (116, 117)  ! Various desorption processes
         taux = xmatk(i_h2)%fdt(k)%k * ab(k1)                             ![cm-3 s-1]
         DO lev = 1, nused
            iju = njl_h2(lev)
            ivu = nvl_h2(lev)
            create(lev) = create(lev) + taux * pophm1(ivu,iju)
         ENDDO
         ratedes = ratedes + taux                                         ![cm-3 s-1]

      !.........................................................................
      ! (118) H2 evaporation from grains
      !.........................................................................
      CASE (118)  ! Various desorption processes
         taux = xmatk(i_h2)%fdt(k)%k * ab(k1)                             ![cm-3 s-1]
         DO lev = 1, nused
            create(lev) = create(lev) + taux * fr_des_m(lev)
         ENDDO
         ratedes = ratedes + taux                                         ![cm-3 s-1]

      !===================================================================================
      !=== H2 destruction
      !===================================================================================

      !.........................................................................
      ! (7) H2 photodissociation
      !.........................................................................
      CASE (7)  ! Photo-dissociation
         ! Matrix evrdbc does not include dissociation any more
         !--- Same destruction rate for all levels
         !destru = destru + xmatk(i_h2)%fdt(k)%k*ab(k1)*ab(k2)*ab(k3)
         !--- or use specific rate per level
         DO lev = 1, nused
            remove(lev) = remove(lev) + pdh2vJ(lev,iopt) * ab(k1)
            ratedph     = ratedph     + pdh2vJ(lev,iopt) * ab(k1) * xreh2(lev,iopt) ![cm-3 s-1]
         ENDDO

      !.........................................................................
      ! (6) H2 destruction by endothermal reactions
      !.........................................................................
      CASE (6)   ! Energy dependent rxns (endothermal)
         DO lev = 1, nused
            endo_rxn    = MAX(bet(nrendo) - spec_lin(in_sp(i_h2))%elk(lev), 0.0_dp)
            rat_l       = gamm(nrendo) * (ttry/300.0_dp)**alpha(nrendo) * EXP(-endo_rxn/ttry)
            remove(lev) = remove(lev) + rat_l * ab(k1) * ab(k2)               ! [cm-3 s-1]
            ratedga     = ratedga     + rat_l * ab(k1) * ab(k2) * xreh2(lev,iopt)
         ENDDO
         nrendo = nrendo + 1

      !.........................................................................
      ! (61) CH+ formation - Destroy H2 in levels
      !.........................................................................
      CASE (61)
         IF (nused <= 8 ) THEN
            WRITE (*,*) ' Error in PXDR_EXCITH2: CH+ formation'
            STOP
         ENDIF

         !--- Mix Zanchet et al. (ApJ, 2013) and Herraez-Aguilar et al. (PCCP, 2014)
         ! Octavio Roncero suggested to be more confident in the quantum treatment by Zanchet et al.
         ! rather than the semi-classical treatment of Herraez-Aguilar.
         ! Herraez-Aguilar provides rates for H2 in v=0, J=0, 1, 3, 5, 7 + v=1, J=0, 1, 2, 3, 4
         ! Zanchet provides rates for v=0, J=0; v=1, J=0 and J=1; v=2, J=0
         ! Following Octavio Roncero's suggestions, we scale Herraez-Aguilar rates to match Zanchet rates
         ! We find two scaling laws, one for levels <= v=0, J=7 and on for above levels
         ! Note that the rotationnal level v=0, J=8 has the same shape than v=1 levels and so it is scaled by scale_vib.
         scale_rot = 2.4_dp * (ttry / 1000.0_dp)**(-0.5_dp)
         scale_vib = 1.4_dp * (ttry / 1000.0_dp)**(-0.1_dp)
         ! Compute Herraez-Aguilar rates and scale them to get Zanchet et al. values
         rat_l = 10.0_dp**(-(1640.07_dp * ttry**(-0.94_dp) + 9.83_dp)) * scale_rot ! v=0, J=0
         remove(1) = remove(1) + rat_l * ab(k1) * ab(k2)
         ratedga   = ratedga   + rat_l * ab(k1) * ab(k2) * xreh2(1,iopt)

         rat_l = 10.0_dp**(-(1682.81_dp * ttry**(-0.95_dp) + 9.87_dp)) * scale_rot ! v=0, J=1
         remove(2) = remove(2) + rat_l * ab(k1) * ab(k2)
         ratedga   = ratedga   + rat_l * ab(k1) * ab(k2) * xreh2(2,iopt)

         rat_l = 10.0_dp**(-(1400.00_dp * ttry**(-0.94_dp) + 9.80_dp)) * scale_rot ! v=0, J=2 - Interpolation
         remove(3) = remove(3) + rat_l * ab(k1) * ab(k2)
         ratedga   = ratedga   + rat_l * ab(k1) * ab(k2) * xreh2(3,iopt)

         rat_l = 10.0_dp**(-(1219.00_dp * ttry**(-0.93_dp) + 9.74_dp)) * scale_rot ! v=0, J=3
         remove(4) = remove(4) + rat_l * ab(k1) * ab(k2)
         ratedga   = ratedga   + rat_l * ab(k1) * ab(k2) * xreh2(4,iopt)

         rat_l = 10.0_dp**(-(1000.00_dp * ttry**(-0.92_dp) + 9.74_dp)) * scale_rot ! v=0, J=4 - Interpolation
         remove(5) = remove(5) + rat_l * ab(k1) * ab(k2)
         ratedga   = ratedga   + rat_l * ab(k1) * ab(k2) * xreh2(5,iopt)

         rat_l = 10.0_dp**(-(740.30_dp  * ttry**(-0.92_dp) + 9.74_dp)) * scale_rot ! v=0, J=5
         remove(6) = remove(6) + rat_l * ab(k1) * ab(k2)
         ratedga   = ratedga   + rat_l * ab(k1) * ab(k2) * xreh2(6,iopt)

         rat_l = 10.0_dp**(-(322.27_dp  * ttry**(-0.86_dp) + 9.72_dp)) * scale_rot ! v=0, J=6 - Interpolation
         remove(7) = remove(7) + rat_l * ab(k1) * ab(k2)
         ratedga   = ratedga   + rat_l * ab(k1) * ab(k2) * xreh2(7,iopt)

         rat_l = 10.0_dp**(-(21.41_dp   * ttry**(-0.52_dp) + 9.58_dp)) * scale_rot ! v=0, J=7
         remove(8) = remove(8) + rat_l * ab(k1) * ab(k2)
         ratedga   = ratedga   + rat_l * ab(k1) * ab(k2) * xreh2(8,iopt)

         rat_l = (4.66e-10_dp * ttry**(-0.29_dp) + 1.25e-10_dp) * scale_vib        ! v=0, J=8
         remove(9) = remove(9) + rat_l * ab(k1) * ab(k2)
         ratedga   = ratedga   + rat_l * ab(k1) * ab(k2) * xreh2(9,iopt)

         ! From v=1, J=0 (lev=10) and up to v=2, J=0 (lev=22) excluded use scaled Herraez-Aguilar rates
         DO lev = 10, min(l0h2, 21) ! lev = 10 is v=1, J=0 and lev = 22 is v=2, J=0
            rat_l = (8.0e-10_dp * ttry**(-0.29_dp) + 1.25e-10_dp) * scale_vib
            remove(lev) = remove(lev) + rat_l * ab(k1) * ab(k2)
            ratedga     = ratedga     + rat_l * ab(k1) * ab(k2) * xreh2(lev,iopt)
         ENDDO
         ! From v = 2, J=0 (lev = 22) use Zanchet et al. v=2, J=0 rate
         DO lev = min(22,l0h2), l0h2
            rat_l = (2.0e-9_dp * ttry**(-0.29_dp) + 1.8e-10_dp)
            remove(lev) = remove(lev) + rat_l * ab(k1) * ab(k2)
            ratedga     = ratedga     + rat_l * ab(k1) * ab(k2) * xreh2(lev,iopt)
         ENDDO
         nrendo = nrendo + 1

      !.........................................................................
      ! (62) SH+ formation - Destroy H2 in levels
      !.........................................................................
      CASE (62)  ! Formation of SH+

         !--- Old prescription copied from CH+
         !IF (nused <= 18 ) THEN
         !   WRITE (*,*) ' Error in PXDR_EXCITH2: SH+ formation'
         !   STOP
         !ENDIF
         !DO ii = 1, 18 ! Endothermal reaction with the 18 first levels of H2
         !   endo_rxn   = MAX(9860.0_dp - spec_lin(in_sp(i_h2))%elk(ii), 0.0_dp)
         !   rat_l      = 1.58e-10_dp * EXP(-endo_rxn/ttry)
         !   remove(ii) = remove(ii) + rat_l * ab(k1) * ab(k2)
         !   ratedga    = ratedga    + rat_l * ab(k1) * ab(k2) * xreh2(ii,iopt)
         !ENDDO
         !DO ii = 19, nused
         !   remove(ii) = remove(ii) + 1.6e-9_dp * ab(k1) * ab(k2)
         !   ratedga    = ratedga    + 1.6e-9_dp * ab(k1) * ab(k2) * xreh2(ii,iopt)
         !ENDDO

         !--- Rates from Zanchet et al. 2013b
         tt = MAX(10.0_dp,ttry)
         t3 = tt / 300.0_dp
         DO lev = 1, l0h2
            SELECT CASE (nvl_h2(lev))
            CASE (0)
               rat_l = 9.1324e-14_dp * (t3)**2.65_dp * EXP(-12772.2_dp/tt)
               remove(lev) = remove(lev) + rat_l * ab(k1) * ab(k2)
               ratedga     = ratedga     + rat_l * ab(k1) * ab(k2) * xreh2(lev,iopt)
            CASE (1)
               rat_l = 1.1731e-12_dp * (t3)**1.52_dp * EXP(-2575.74_dp/tt)
               remove(lev) = remove(lev) + rat_l * ab(k1) * ab(k2)
               ratedga     = ratedga     + rat_l * ab(k1) * ab(k2) * xreh2(lev,iopt)
            CASE (2)
               rat_l = (0.5_dp + 0.5_dp*TANH( 2.522e-4_dp*(4132.13_dp-tt))) * 2.0410e-12_dp *(t3)**1.006_dp    *EXP(22.591_dp/tt)  &
                     + (0.5_dp + 0.5_dp*TANH(-2.522e-4_dp*(4132.13_dp-tt))) * 3.4247e-10_dp *(t3)**(-0.201_dp) *EXP(-130.634_dp/tt)
               remove(lev) = remove(lev) + rat_l * ab(k1) * ab(k2)
               ratedga     = ratedga     + rat_l * ab(k1) * ab(k2) * xreh2(lev,iopt)
            CASE (3)
               rat_l = (0.5_dp + 0.5_dp*TANH( 4.8411e-4_dp*(3394.1_dp-tt))) * 1.9824e-10_dp * (t3)**(-0.106_dp) *EXP(-110.39_dp/tt)&
                     + (0.5_dp + 0.5_dp*TANH(-4.8411e-4_dp*(3394.1_dp-tt))) * 5.8290e-10_dp * (t3)**(-0.297_dp) *EXP(1.02_dp/tt)
               remove(lev) = remove(lev) + rat_l * ab(k1) * ab(k2)
               ratedga     = ratedga     + rat_l * ab(k1) * ab(k2) * xreh2(lev,iopt)
            CASE DEFAULT ! v>=4 use the rate for v=4
               rat_l = (0.5_dp + 0.5_dp*TANH( 1.508e-3_dp*(678.31_dp-tt)))  * 2.8479e-10_dp * (t3)**0.119_dp   * EXP(-7.257_dp/tt) &
                     + (0.5_dp + 0.5_dp*TANH(-1.508e-3_dp*(678.31_dp-tt)))  * 2.4097e-10_dp * (t3)**0.124_dp   * EXP(12.334_dp/tt)
               remove(lev) = remove(lev) + rat_l * ab(k1) * ab(k2)
               ratedga     = ratedga     + rat_l * ab(k1) * ab(k2) * xreh2(lev,iopt)
            END SELECT
         ENDDO

         nrendo = nrendo + 1

      !.........................................................................
      ! (110) Collisional destruction of H2 by H and H2 - rates depend on v
      !.........................................................................
      CASE (110)  ! Collisional destruction of H2
         IF (k1 == i_h .AND. k2 == i_h2) THEN
            SELECT CASE (F_H2_COLL_D)
            CASE (1)
               taux = xmatk(i_h2)%fdt(k)%k
               ratedga   = ratedga   - taux * ab(k1) * ab(k2)
               remove(:) = remove(:) - taux * ab(k1) * ab(k2)
            CASE (2)
               DO iv = 0, 14
                  kvh2(iv) = -a1h2v(iv) - a2h2v(iv) &
                           / (ttry**a3h2v(iv) * (1.0_dp + a4h2v(iv) * ttry**a5h2v(iv)))
               ENDDO
               kvh2 = EXP(kvh2)
               DO lev = 1, l0h2
                  iv = nvl_h2(lev)
                  remove(lev) = remove(lev) + kvh2(iv) * ab(i_h) * ab(i_h2)
                  ratedga     = ratedga + kvh2(iv) * ab(i_h) * ab(i_h2) * xreh2(lev,iopt)
               ENDDO
            CASE DEFAULT
               PRINT *, " F_H2_COLL_D = ", F_H2_COLL_D, " not yet implemented"
            END SELECT
         ELSE IF (k1 == i_h2 .AND. k2 == i_h2 .AND. if_110h2 == 0) THEN
            if_110h2 = 1
            SELECT CASE (F_H2_COLL_D)
            CASE (1)
               taux = xmatk(i_h2)%fdt(k)%k
               ratedga   = ratedga   - taux * ab(k1) * ab(k2)
               remove(:) = remove(:) - taux * ab(k1) * ab(k2)
            CASE (2)
               DO iv = 0, 14
                  kvh2(iv) = 1.30_dp * (1.0_dp + 0.023_dp*iv &
                           + 1.93e-5_dp * iv**4.75_dp + 2.85e-24_dp * iv**21.60_dp) &
                           * EXP(-(7.47_dp - 0.322_dp*iv) * 1.0e3 / ttry) * 1.0e-10_dp
               ENDDO
               DO lev = 1, l0h2
                  iv = nvl_h2(lev)
                  remove(lev) = remove(lev) + kvh2(iv) * ab(i_h2) * ab(i_h2)
                  ratedga     = ratedga + kvh2(iv) * ab(i_h2) * ab(i_h2) * xreh2(lev,iopt)
               ENDDO
            CASE DEFAULT
               PRINT *, " F_H2_COLL_D = ", F_H2_COLL_D, " not yet implemented"
            END SELECT
         ENDIF

      !.........................................................................
      ! (113) H2 sticking on grains
      !.........................................................................
      CASE (113)  ! sticking on grains and Eley-Rideal on physisorbed sites
         ! ATTENTION: besoin de verifier les signes dans les + / - taux
         taux = xmatk(i_h2)%fdt(k)%k                  ! xmatk is negative or positive
         IF (k1 == i_h2) THEN
            IF (if_ac == 0) THEN                      ! adsorption
               ! taux is negative
               ! * use -taux to add the contribution to rateads and remove
               rateads   = rateads   - taux * ab(k1)
               remove(:) = remove(:) - taux * ab(k1)
               if_ac = 1
            ELSE                                      ! rejection
               ! taux is positive
               ! * use -taux because it is a correction term to the previous IF
               rateads   = rateads   - taux * ab(k1) * ab(k2)
               remove(:) = remove(:) - taux * ab(k1) * ab(k2)
            ENDIF
         ELSE     ! Eley-Rideal formation on physisorbed species
            ! taux is positive
            ratefER = ratefER + taux * ab(k1) * ab(k2)
            !--- Excitation ---
            ! create is filled after ratefer has been completed with ER formation
            ! involving chemisorbed species (type 20)
         ENDIF

      !.........................................................................
      ! DEFAULT - H2 formation and destruction in gas chemical reactions
      !           other than photo-dissociation
      !.........................................................................
      CASE DEFAULT
         taux = xmatk(i_h2)%fdt(k)%k * ab(k1) * ab(k2) * ab(k3)         ! [cm-3 s-1]
         IF (taux > 0.0_dp) THEN
            !*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!
            ! NOTE: We may wish to add specific excitation for exothermic reactions
            !       Such excitation can be controled by pophm_ch
            !*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!
            ! Distribution of H2 formed in generic gas phase reactions
            ! Ad hoc hypothesis : 2/3 in J=0, 1/3 in J=1
            !     because rxn h3+ + x -> h2 + xh+ dominates (WHY ?)
            create(levJ0) = create(levJ0) + 2.0_dp / 3.0_dp * taux
            create(levJ1) = create(levJ1) + taux / 3.0_dp
            ratefga   = ratefga   + taux                                ! [cm-3 s-1]
         ELSE ! taux is negative
            remove(:) = remove(:) - taux                                ! [cm-3 s-1]
            ratedga   = ratedga   - taux                                ! [cm-3 s-1]
         ENDIF
      END SELECT
   ENDDO

   !======================================================================================
   ! Additionnal terms in create(:)
   !======================================================================================
   !--- Excitation by Eley-Rideal mechanisms (chemisorbed + physisorbed) : pophm2
   DO lev = 1, nused
      ij = njl_h2(lev)
      iv = nvl_h2(lev)
      dumy = 0.0_dp
      DO levu = nused+1, nlevh2
         iju = njl_h2(levu)
         ivu = nvl_h2(levu)
         dumy = dumy + pophm2(ivu,iju) * cashm(levu,lev)
      ENDDO
      create(lev) = create(lev) + (dumy + pophm2(iv,ij)) * ratefER
   ENDDO

   !--------------------------------------------------------------------------------------
   ! Store some useful data for PREP concerning H2 formation / destruction processes
   ! Note: timh23 and timh25 are used to compute heating rate by H2 formation on grains
   ! timh21(iopt) : H2 formation by gas phase chemistry                        [cm-3 s-1]
   ! timh22(iopt) : H2 destruction by gas phase chemistry                      [cm-3 s-1]
   ! timh23(iopt) : H2 formation by Langmuir-Hinshelwood process or old style  [cm-3 s-1]
   ! timh24(iopt) : H2 destruction by photons                                  [cm-3 s-1]
   ! timh25(iopt) : H2 formation by Eley-Rideal process                        [cm-3 s-1]
   !--------------------------------------------------------------------------------------
   timh21(iopt) = ratefga + ratedes      ! fabric_exo
   timh22(iopt) = ratedga + rateads      ! does not include photo-dissocation in gas phase
                                         ! adsorption must be included
   timh23(iopt) = timh23(iopt) + ratefgr ! one of the two term is zero
   timh24(iopt) = ratedph
   timh25(iopt) = ratefER

END SUBROUTINE EQLH2V

!---------------------------------------------------------------------

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
SUBROUTINE COLH2V (ttry, iopt, itn)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   USE PXDR_FLAGS
   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PXDR_AUXILIAR
   USE PXDR_STR_DATA

!---------------------------------------------------------------------
!      xnprot is no. density of protons (cm-3)
!      ttry is temperature of neutral fluid

!      called from sub. EQLH2V
!
!      collision rates for H2 + H come from Martin & Mandy.
!      New fits by Jacques Le Bourlot (july 1996)
!
!          or
!
!      from D. Flower (cf MNRAS)
!
!      Collision rates for H2 + He, H2 + H2o, H2 + H2p
!      come from D. Flower
!
!---------------------------------------------------------------------

   IMPLICIT NONE

   REAL (KIND=dp), INTENT (IN)                      :: ttry
   INTEGER, INTENT (IN)                             :: iopt
   INTEGER, INTENT (IN)                             :: itn

   REAL (KIND=dp), DIMENSION (30)                   :: gerlic
   INTEGER                                          :: nqc
   REAL (KIND=dp)                                   :: xchh2, fac
   REAL (KIND=dp)                                   :: decal, tdd, tdd2
   REAL (KIND=dp)                                   :: gmm, ga1, ga2
   REAL (KIND=dp)                                   :: ch2h_nr, ch2h_r, ch2he, ch2h
   REAL (KIND=dp)                                   :: ch2ph2, ch2oh2, cophp, c, d
   INTEGER                                          :: levu, levl, ndej, lev1, lev2
   INTEGER                                          :: ju, jl, ivu, ivl
   REAL (KIND=dp), DIMENSION (:,:), ALLOCATABLE     :: ctod
   REAL (KIND=dp), DIMENSION (:), ALLOCATABLE       :: tempnrj, g_stat
   INTEGER                                          :: j_sp

   ! With David's prescription for O/P balance (3/1)
   ! Data from Gerlich for ortho-para transfer with h+.
   DATA gerlic / 2.024e-10_dp, 9.575e-10_dp, 2.729e-10_dp, 7.910e-10_dp &
         , 2.246e-10_dp, 6.223e-10_dp, 1.828e-10_dp, 4.975e-10_dp, 1.484e-10_dp &
         , 3.000e-10_dp, 1.000e-10_dp, 3.000e-10_dp, 1.000e-10_dp, 3.000e-10_dp &
         , 1.000e-10_dp, 3.000e-10_dp, 1.000e-10_dp, 3.000e-10_dp, 1.000e-10_dp &
         , 3.000e-10_dp, 1.000e-10_dp, 3.000e-10_dp, 1.000e-10_dp, 3.000e-10_dp &
         , 1.000e-10_dp, 3.000e-10_dp, 1.000e-10_dp, 3.000e-10_dp, 1.000e-10_dp &
         , 3.000e-10_dp /

!  Up to now collisions with H2 and He are used only for v = 0
!  Educated guesses may obtain from collisions with H
!  WARNING: c: desexcitation, d: excitation
!            h2h: H2 + H
!            h2h2: H2 + H2
!            h2he: H2 + He
!            ophp: H2 + H+ (ortho - para)

   j_sp = in_sp(i_h2)
   ALLOCATE (tempnrj(1:l0h2))
   tempnrj(1:l0h2) = spec_lin(j_sp)%elk(1:l0h2) / ttry
   ALLOCATE (g_stat(1:l0h2))
   g_stat(1:l0h2)  = spec_lin(j_sp)%gst(1:l0h2)
   ALLOCATE (ctod(1:l0h2,1:l0h2))
   !ctod = 0.0_dp

   IF (ichh2 == 0) THEN
      nqc = 4
   ELSE
      nqc = 3
   ENDIF
   IF (ichh2 <= 1) THEN
      xchh2 = 0.0_dp
   ELSE
      xchh2 = 1.0_dp
   ENDIF

!$OMP PARALLEL PRIVATE(levl,fac)
!$OMP DO SCHEDULE(DYNAMIC)
   ! ctod is the Boltzmann factor used to compute
   !    excitation rates from desexcitation rates
   DO levu = 2, l0h2
      DO levl = 1, levu-1
         fac = (tempnrj(levu) - tempnrj(levl))
         ctod(levl,levu) = EXP(-fac) * g_stat(levu) / g_stat(levl)
      ENDDO
   ENDDO
!$OMP END DO
!$OMP END PARALLEL

!  xx  $OMP PARALLEL PRIVATE(ju,ivu,jl,ivl,levl,ndej,decal,tdd,tdd2,gmm,ch2h_nr,ch2h_r) &
!  xx  $OMP& PRIVATE(lev1,lev2,ga1,ga2,ch2h,ch2he,ch2ph2,ch2oh2,cophp,c,d)
!  xx  $OMP DO SCHEDULE(DYNAMIC)
   DO levu = 2, l0h2
      ju = njl_h2(levu)
      ivu = nvl_h2(levu)

      DO levl = 1, levu-1
         jl = njl_h2(levl)
         ivl = nvl_h2(levl)
         ndej = ABS(ju - jl)

         ! Fit of H2 + H rates. The reduced variable tdd is offset by a
         ! different value for reactive (resp. non-reactive) collisions.
         IF (MOD(ndej,2) == 0) THEN
            decal = 1.0_dp
         ELSE
            decal = 0.3_dp
         ENDIF
         tdd = decal + ttry * 1.0e-3_dp
         tdd2 = tdd*tdd

         ! Reactive and non-reactive collisions are mixed in Martin and Mandy
         gmm = q_h2_h_1(levl,levu) + q_h2_h_2(levl,levu) / tdd + q_h2_h_3(levl,levu) / tdd2
         IF (nqc == 4) THEN
            gmm = gmm +  q_h2_h_4(levl,levu) * tdd
         ENDIF

         ch2h_nr = 10.0_dp**gmm
         ch2h_r  = 0.0_dp

         ! Add reactive collisions (David's rule) if ichh2 = 3
         IF (ichh2 == 3 .AND. tau(iopt) <= 1.0_dp) THEN

            IF (MOD(ndej,2) == 0) THEN
               ch2h_r = 10.0_dp**gmm * EXP(-MAX(0.0_dp,(3900.0_dp - spec_lin(in_sp(i_h2))%elk(levu) &
                      + spec_lin(in_sp(i_h2))%elk(levl))/ttry))
            ELSE
               IF (jl > 0) THEN
                  lev1 = lev_h2(ivl,jl-1)
               ELSE
                  lev1 = lev_h2(ivl,jl+1)
               ENDIF
               lev2 = lev_h2(ivl,jl+1)
               IF (lev2 == 0) THEN
                  lev2 = lev_h2(ivl,jl-1)
               ENDIF
               IF (lev2 > l0h2) THEN
                  print *, " +++ Above upper H2 level!"
                  lev2 = levu
               ENDIF
               ga1 = q_h2_h_1(lev1,levu) + q_h2_h_2(lev1,levu) / tdd + q_h2_h_3(lev1,levu) / tdd2
               IF (nqc == 4) THEN
                  ga1 = ga1 + q_h2_h_4(lev1,levu) * tdd
               ENDIF
               ga2 = q_h2_h_1(lev2,levu) + q_h2_h_2(lev2,levu) / tdd + q_h2_h_3(lev2,levu) / tdd2
               IF (nqc == 4) THEN
                  ga2 = ga2 + q_h2_h_4(lev2,levu) * tdd
               ENDIF
               ch2h_r = (10.0_dp**ga1 + 10.0_dp**ga2) * 0.5 * EXP(-MAX(0.0_dp,(3900.0_dp - spec_lin(in_sp(i_h2))%elk(levu) &
                      + spec_lin(in_sp(i_h2))%elk(levl))/ttry))

               IF (MOD(ju,2) == 1) THEN
                  ch2h_r = ch2h_r / 3.0_dp
               ENDIF
            ENDIF
         ENDIF

         ! Ortho/Para transfer with H from Schofield
         ! plus David's rule (for a few delta v = 0)
         IF (ichh2 >= 2 .AND.  ivu == ivl .AND. ndej < 3) THEN
            IF (MOD(ju,2) == 1 .AND. MOD(ndej,2) == 1) THEN
               ch2h_r = 8.0e-11_dp * EXP(-3900.0_dp/ttry) / 3.0_dp
            ELSE
               ch2h_r = 8.0e-11_dp * EXP(-3900.0_dp/ttry)
            ENDIF
         ENDIF

         ch2h = ch2h_nr + xchh2 * ch2h_r

         ! He - H2
         gmm = q_h2_he_1(levl,levu) + q_h2_he_2(levl,levu) / tdd + q_h2_he_3(levl,levu) / tdd2
         ch2he = 10.0_dp**gmm

         ! pH2 - H2
         gmm = q_h2_ph2_1(levl,levu) + q_h2_ph2_2(levl,levu) / tdd + q_h2_ph2_3(levl,levu) / tdd2
         ch2ph2 = 10.0_dp**gmm

         ! oH2 - H2
         gmm = q_h2_oh2_1(levl,levu) + q_h2_oh2_2(levl,levu) / tdd + q_h2_oh2_3(levl,levu) / tdd2
         ch2oh2 = 10.0_dp**gmm

         ! c- Ortho to Para transfer with ions (Gerlich), only for v = 0
         ! IF (ivu == 0 .AND. ndej == 1) THEN
         IF (jl == ju-1) THEN
            cophp = gerlic(ju)
         ELSE
            cophp = 0.0_dp
         ENDIF

         ! Bringing it all back home
         c = ch2ph2 * ab(i_h2) * h2para + ch2oh2 * ab(i_h2) * h2orth &
           + ch2he * ab(i_he) + ch2h * ab(i_h)
         IF (i_hp /= 0) THEN
            c = c + cophp * ab(i_hp)
         ENDIF
         IF (i_h3p /= 0) THEN
            c = c + cophp * ab(i_h3p)
         ENDIF
         d = c * ctod(levl,levu)

         eqcol(levl,levu) = eqcol(levl,levu) - c
         eqcol(levu,levu) = eqcol(levu,levu) + c
         eqcol(levu,levl) = eqcol(levu,levl) - d
         eqcol(levl,levl) = eqcol(levl,levl) + d
      ENDDO
   ENDDO
!  xx  $OMP END DO
!  xx  $OMP END PARALLEL
   DEALLOCATE (tempnrj)
   DEALLOCATE (g_stat)
   DEALLOCATE (ctod)

END SUBROUTINE COLH2V

END MODULE PXDR_EXCITH2
