! Copyright (C) 2002 - Observatoire de Paris/Meudon
!                      Universite Paris 7
!                      CNRS (France)
!                      Contact : Jacques.Lebourlot@obspm.fr
!
! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License version 2
! as published by the Free Software Foundation.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


MODULE PXDR_RF_INIT

   USE PXDR_FLAGS
   USE PXDR_CONSTANTES
   USE PXDR_TRANSFER

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: RF_INIT, ADD_WL_LINES

!------------------------------------------------------------------------

CONTAINS

!**********************************************************************************************************
SUBROUTINE RF_INIT
!**********************************************************************************************************
! purpose : Initialise model dependant quantities related to Radiation Field (RF)
!   * Wavelength grid                    : rf_wl (need not have constant step)
!   * Incident ISRF (specific intensity) : rf_incIm and rf_incIp
!   * Dust absorption curve              : rf_extpou (Fitzpatrick fit)

!---------------------------Declaration of PARAMETERs-----------------------------------------------------
    USE PXDR_FLAGS
    USE PXDR_CONSTANTES
    USE PXDR_CHEM_DATA
    USE PXDR_PROFIL
    USE PXDR_AUXILIAR
    USE PXDR_STR_DATA
    USE PXDR_DRAINE_DUST
    USE PXDR_RF_TOOLS
    USE PXDR_TRANSFER
    USE PXDR_READCHIM
    USE PXDR_DUSTEM

    IMPLICIT NONE

    INTEGER                                    :: I_cut_L              !  Index of Lyman limit
    REAL (KIND=dp)                             :: wl_cut_D = 2000.0_dp !  Draine field cutoff wavelength
    REAL (KIND=dp)                             :: wl_cut_F             !  Fitzpatrick cutoff wavelength (Ang)
    INTEGER                                    :: I_cut_F              !  Index of Fitzpatrick cutoff position in grid
    REAL (KIND=dp)                             :: wl_cut_F2            !  WL cutoff below which grains cross section is constant
    INTEGER                                    :: I_cut_F2             !  Index of the wavelength defined above
    REAL (KIND=dp)                             :: wl_cut               !  Dummy WL cutoff
    INTEGER                                    :: I_cut                !  Index of dummy cutoff position in grid

    INTEGER                                    :: mm
    INTEGER                                    :: ii, i, j, k, l
    INTEGER                                    :: nb_pt                !  Number of points in a file
    INTEGER                                    :: np_before            !  Number of points before Lyman lim.
    INTEGER                                    :: add_1, iloc, iord    !  Used by RF_EXTAND
    REAL (KIND=dp)                             :: wl, rf
    REAL (KIND=dp)                             :: wl1, wl2, sig1, sig2
    TYPE (RF_TYPE)                             :: rf_tmp

    REAL(KIND=dp)                              :: wl0
    INTEGER                                    :: njl, nvl
    INTEGER                                    :: nwlg_incr

    ! Auxiliary variables for DUSTEM field

    REAL(KIND=dp)                              :: r1, r2, r3, r4
    REAL(KIND=dp)                              :: wlth, fact
    REAL(KIND=dp)                              :: absor, scater, extinct
    REAL(KIND=dp)                              :: Dr_int1, Dr_int2
    INTEGER                                    :: il1, il2, lev

   ! Roots of Legendre polynomials
   !    Taken from file: lgvalues-abscissa.php
   !    found at: https://pomax.github.io/bezierinfo/legendre-gauss.html
   ! Set in reverse (correct) order : bug found by Javier, 15 II 2007
   ! Select the one you need by changing the comment and the name in this file
   ! Do not forget to modify xmui_M in PXDR_RF_INIT.f90

    REAL (KIND=dp), DIMENSION(32)              :: xmui32    ! L = 63

    DATA xmui32 / 0.999305041735772139456905624346_dp, 0.996340116771955279346924500676_dp, &
                  0.991013371476744320739382383443_dp, 0.983336253884625956931299302157_dp, &
                  0.973326827789910963741853507352_dp, 0.961008799652053718918614121897_dp, &
                  0.946411374858402816062481491347_dp, 0.929569172131939575821490154559_dp, &
                  0.910522137078502805756380668008_dp, 0.889315445995114105853404038273_dp, &
                  0.865999398154092819760783385070_dp, 0.840629296252580362751691544696_dp, &
                  0.813265315122797559741923338086_dp, 0.783972358943341407610220525214_dp, &
                  0.752819907260531896611863774886_dp, 0.719881850171610826848940217832_dp, &
                  0.685236313054233242563558371031_dp, 0.648965471254657339857761231993_dp, &
                  0.611155355172393250248852971019_dp, 0.571895646202634034283878116659_dp, &
                  0.531279464019894545658013903544_dp, 0.489403145707052957478526307022_dp, &
                  0.446366017253464087984947714759_dp, 0.402270157963991603695766771260_dp, &
                  0.357220158337668115950442615046_dp, 0.311322871990210956157512698560_dp, &
                  0.264687162208767416373964172510_dp, 0.217423643740007084149648748989_dp, &
                  0.169644420423992818037313629748_dp, 0.121462819296120554470376463492_dp, &
                  0.072993121787799039449542941940_dp, 0.024350292663424432508955842854_dp /

    REAL (KIND=dp), DIMENSION(31)              :: xmui31  ! L = 61

    DATA xmui31 / 0.999259859308777029698408465036_dp, 0.996102296316267132885148430476_dp, &
                  0.990429971189290352428071061861_dp, 0.982255949097236649491103128010_dp, &
                  0.971600723371651806447179482540_dp, 0.958491172973927092029560475166_dp, &
                  0.942960401392328503826364811623_dp, 0.925047635636203755227300130757_dp, &
                  0.904798122521093465758485602428_dp, 0.882263012831897363072686624356_dp, &
                  0.857499231512070922818403320545_dp, 0.830569333604004851345931109477_dp, &
                  0.801541346103976371538505514495_dp, 0.770488596055419318990944694995_dp, &
                  0.737489525283156749860784276501_dp, 0.702627492222297055120972765060_dp, &
                  0.665990561335479446997546034082_dp, 0.627671280646885180726954764916_dp, &
                  0.587766447953087338002149550122_dp, 0.546376866300251095816100384524_dp, &
                  0.503607089344755955917020805932_dp, 0.459565157240113395207143978076_dp, &
                  0.414362323717126048129588750656_dp, 0.368112775046564529663021830953_dp, &
                  0.320933341594194004074101601003_dp, 0.272943202696726343188683627973_dp, &
                  0.224263585604165531668991393345_dp, 0.175017459249015628555687423904_dp, &
                  0.125329223615896808619648999500_dp, 0.075324395496234332763826939118_dp, &
                  0.025129291421820614727088275999_dp /

    REAL (KIND=dp), DIMENSION(30)              :: xmui30  ! L = 59

    DATA xmui30 / 0.999210123227436022034229585798_dp, 0.995840525118838173876746713377_dp, &
                  0.989787895222221717367278987016_dp, 0.981067201752598185618576799827_dp, &
                  0.969701788765052733721544098914_dp, 0.955722255839996107397231845830_dp, &
                  0.939166276116423249495419011610_dp, 0.920078476177627552856656862520_dp, &
                  0.898510310810045941937789329573_dp, 0.874519922646898315129308099912_dp, &
                  0.848171984785929632490515494994_dp, 0.819537526162145759368518108520_dp, &
                  0.788693739932264054569944799777_dp, 0.755723775306585686868842066602_dp, &
                  0.720716513355730399436021061014_dp, 0.683766327381355437222930239224_dp, &
                  0.644972828489477067813447896420_dp, 0.604440597048510363444208776311_dp, &
                  0.562278900753944539178272587486_dp, 0.518601400058569747417889348485_dp, &
                  0.473525841761707111108163053753_dp, 0.427173741583078389307452853530_dp, &
                  0.379670056576797977154952670522_dp, 0.331142848268448194252352965351_dp, &
                  0.281722937423261691690694860339_dp, 0.231543551376029338010344631347_dp, &
                  0.180739964873425417240876941262_dp, 0.129449135396945003146444164650_dp, &
                  0.077809333949536569419285507082_dp, 0.025959772301247798589170385400_dp /

    REAL (KIND=dp), DIMENSION(29)              :: xmui29  ! L = 57

    DATA xmui29 / 0.999155200407386606442737044185_dp, 0.995551476597290902602771526180_dp, &
                  0.989079008248442636499910255275_dp, 0.979755014694350309107816334393_dp, &
                  0.967606202502924090153300784621_dp, 0.952667557518869091442845510023_dp, &
                  0.934982137588259348480843234766_dp, 0.914600928564352540686505649898_dp, &
                  0.891582692022030176399882820813_dp, 0.865993794074807479275071232512_dp, &
                  0.837908013339373316352097920622_dp, 0.807406327913088141048997152207_dp, &
                  0.774576681749652745266140046497_dp, 0.739513731020042267846617234851_dp, &
                  0.702318571153908113479757301458_dp, 0.663098445332125266433016299754_dp, &
                  0.621966435263079111033984058745_dp, 0.579041135130225030489977117136_dp, &
                  0.534446309648847586399802417425_dp, 0.488310537216718463615552572737_dp, &
                  0.440766839186839565193703062214_dp, 0.391952296330753150371216941213_dp, &
                  0.342007653597995261248368267484_dp, 0.291076914311109189533025268314_dp, &
                  0.239306924966153454428948509085_dp, 0.186846951835761321374383881464_dp, &
                  0.133848250595466857022376847875_dp, 0.080463630214142729309847792381_dp, &
                  0.026847012365942355803338689517_dp /

    REAL (KIND=dp), DIMENSION(28)              :: xmui28  ! L = 55

    DATA xmui28 / 0.999094343801465584353153825091_dp, 0.995231226081069747216308760099_dp, &
                  0.988293715540161511089925163523_dp, 0.978301709140256383376990551538_dp, &
                  0.965285901905490183626194976055_dp, 0.949286479561962635646737602795_dp, &
                  0.930352880247496300547272723292_dp, 0.908543620420655490846107289957_dp, &
                  0.883926108327827540789012285809_dp, 0.856576433762748635402991386767_dp, &
                  0.826579132142881651672126974457_dp, 0.794026922893866498030008308486_dp, &
                  0.759020422705128902202413294446_dp, 0.721667834450188083522593787048_dp, &
                  0.682084612694470455501561477970_dp, 0.640393106807006894267938596063_dp, &
                  0.596722182770663320104135217138_dp, 0.551206824855534618754363514576_dp, &
                  0.503987718384381714195224401219_dp, 0.455210814878459578948831581325_dp, &
                  0.405026880927091278118866962603_dp, 0.353591032174954520969707327033_dp, &
                  0.301062253867220669053094233125_dp, 0.247602909434337203972966554971_dp, &
                  0.193378238635275258240184723013_dp, 0.138555846810376242012886580438_dp, &
                  0.083305186822435374440287355441_dp, 0.027797035287275437094061196775_dp /

    REAL (KIND=dp), DIMENSION(27)              :: xmui27  ! L = 53

    DATA xmui27 / 0.999026666867340983851071106984_dp, 0.994875117018338884918889995949_dp, &
                  0.987420637397343558552145720404_dp, 0.976686328857903237200010071988_dp, &
                  0.962707645785923583256993926905_dp, 0.945530975164995853763837022906_dp, &
                  0.925213359866651486256276060480_dp, 0.901822286284701580757467231388_dp, &
                  0.875435454065568939417909335723_dp, 0.846140515970772949425887510873_dp, &
                  0.814034785913567835469638304075_dp, 0.779224915346254021535948614841_dp, &
                  0.741826538809184316285766268754_dp, 0.701963889719172919385570976970_dp, &
                  0.659769387631983124692289822856_dp, 0.615383198331127370729937820162_dp, &
                  0.568952768195209429731624834177_dp, 0.520632334385933073327017138750_dp, &
                  0.470582412481382283683212136328_dp, 0.418969263255204528036102697463_dp, &
                  0.365964340372191181984332108165_dp, 0.311743720834468228882550390152_dp, &
                  0.256487520069997300077411214453_dp, 0.200379293606213569778602462829_dp, &
                  0.143605427316256153947089762899_dp, 0.086354518263248215285443177769_dp, &
                  0.028816748199341777656235847500_dp /

    REAL (KIND=dp), DIMENSION(26)              :: xmui26  ! L = 51

    DATA xmui26 / 0.998951111103950278090961440717_dp, 0.994477590929216029245043271800_dp, &
                  0.986446195651549840645322859118_dp, 0.974883884221744503140706470400_dp, &
                  0.959831826933086552532025153425_dp, 0.941343853641359056843579183851_dp, &
                  0.919486128916424539893596376542_dp, 0.894336890534495322520844088264_dp, &
                  0.865986162846067585244358517624_dp, 0.834535432326734534961726389621_dp, &
                  0.800097283430468324334758754226_dp, 0.762794995193744960279211163089_dp, &
                  0.722762099749983193676816676072_dp, 0.680141904227167702092238779737_dp, &
                  0.635086977695245924298124579289_dp, 0.587758604979579069902025781557_dp, &
                  0.538326209285827438375951167975_dp, 0.486966745698096077782458241346_dp, &
                  0.433864067718761670309086568210_dp, 0.379208269116093669246681766939_dp, &
                  0.323195003434807825501099066601_dp, 0.266024783605001827472971730599_dp, &
                  0.207902264156366059686466128122_dp, 0.149035508606949180488634040018_dp, &
                  0.089635244648900565488854691122_dp, 0.029914109797338766043680776071_dp /

    REAL (KIND=dp), DIMENSION(25)              :: xmui25  ! L = 49

    DATA xmui25 / 0.998866404420071050185459444974_dp, 0.994031969432090712585108200421_dp, &
                  0.985354084048005882309009625632_dp, 0.972864385106692073713344104606_dp, &
                  0.956610955242807942997745644157_dp, 0.936656618944877933780874947272_dp, &
                  0.913078556655791893089735642772_dp, 0.885967979523613048637540982467_dp, &
                  0.855429769429946084611362643935_dp, 0.821582070859335948356254110874_dp, &
                  0.784555832900399263905305196341_dp, 0.744494302226068538260536252682_dp, &
                  0.701552468706822251089546257884_dp, 0.655896465685439360781624864004_dp, &
                  0.607702927184950239180381796392_dp, 0.557158304514650054315522909626_dp, &
                  0.504458144907464201651459131849_dp, 0.449806334974038789147131467778_dp, &
                  0.393414311897565127394229253824_dp, 0.335500245419437356836988257291_dp, &
                  0.276288193779531990327645278521_dp, 0.216007236876041756847284532617_dp, &
                  0.154890589998145902071628620941_dp, 0.093174701560086140854450377640_dp, &
                  0.031098338327188876112328989666_dp /

    REAL (KIND=dp), DIMENSION(24)              :: xmui24  ! L = 47

    DATA xmui24 / 0.998771007252426118600541491563_dp, 0.993530172266350757547928750849_dp, &
                  0.984124583722826857744583600027_dp, 0.970591592546247250461411983801_dp, &
                  0.952987703160430860722960666026_dp, 0.931386690706554333114174380102_dp, &
                  0.905879136715569672822074835671_dp, 0.876572020274247885905693554805_dp, &
                  0.843588261624393530711089844520_dp, 0.807066204029442627082553043025_dp, &
                  0.767159032515740339253855437523_dp, 0.724034130923814654674482233494_dp, &
                  0.677872379632663905211851280676_dp, 0.628867396776513623995164933070_dp, &
                  0.577224726083972703817809238540_dp, 0.523160974722233033678225869138_dp, &
                  0.466902904750958404544928861651_dp, 0.408686481990716729916225495815_dp, &
                  0.348755886292160738159817937270_dp, 0.287362487355455576735886461317_dp, &
                  0.224763790394689061224865440175_dp, 0.161222356068891718056437390783_dp, &
                  0.097004699209462698930053955854_dp, 0.032380170962869362033322243152_dp /

    REAL (KIND=dp), DIMENSION(23)              :: xmui23  ! L = 45

    DATA xmui23 / 0.998663042133817981128268417399_dp, 0.992962348906174364073095444666_dp, &
                  0.982733669804166863477979579659_dp, 0.968021391853991942737748066372_dp, &
                  0.948892363446089795622213539437_dp, 0.925433798806753950977491357700_dp, &
                  0.897752711533941965701319951320_dp, 0.865975394866858062915888776817_dp, &
                  0.830246837066066053032385304791_dp, 0.790730057075274255189144147420_dp, &
                  0.747605359615666054000336505268_dp, 0.701069512020405697512122089427_dp, &
                  0.651334846201997715106477155817_dp, 0.598628289712715153177172533600_dp, &
                  0.543190330261802635270962691029_dp, 0.485273918388164662772320193708_dp, &
                  0.425143313282828397322146863429_dp, 0.363072877020995710123706978515_dp, &
                  0.299345822701870015483439256578_dp, 0.234252922206269768626056115574_dp, &
                  0.168091179467103528606796737428_dp, 0.101162475305584239515859657009_dp, &
                  0.033772190016052041519557831891_dp /

    REAL (KIND=dp), DIMENSION(22)              :: xmui22  ! L = 43

    DATA xmui22 / 0.998540200636774224936056720607_dp, 0.992316392138515808483338003614_dp, &
                  0.981151833077913966662749005350_dp, 0.965099650422493139394398050580_dp, &
                  0.944239509118194099203254653027_dp, 0.918675259984175774323001086694_dp, &
                  0.888534238286043202338367910190_dp, 0.853966595004710378728302241293_dp, &
                  0.815144539645135010487437635752_dp, 0.772261479248755899017758549232_dp, &
                  0.725531053660717002606964936720_dp, 0.675186070666122365333699038564_dp, &
                  0.621477345903575847802424699122_dp, 0.564672453185470768424636819382_dp, &
                  0.505054391388202317982798352291_dp, 0.442920174525411483834826593647_dp, &
                  0.378579352014707132511764672772_dp, 0.312352466502785812236545163573_dp, &
                  0.244569456928201251507302434741_dp, 0.175568014775516785746507749381_dp, &
                  0.105691901708653247117305848915_dp, 0.035289236964135359058196704631_dp /

    REAL (KIND=dp), DIMENSION(21)              :: xmui21  ! L = 41

    DATA xmui21 / 0.998399618990062415022868126847_dp, 0.991577288340860919792361253281_dp, &
                  0.979342508063748193708982566543_dp, 0.961759365338204488746927157853_dp, &
                  0.938923557354988178533173433626_dp, 0.910959724904127452583847879209_dp, &
                  0.878020569812172742711985103479_dp, 0.840285983261816900925439995086_dp, &
                  0.797962053255487413232705952546_dp, 0.751279935689480489568472175915_dp, &
                  0.700494590556171213741569929055_dp, 0.645883388869247833957496303718_dp, &
                  0.587744597485109322840711345474_dp, 0.526395749931192287592867558792_dp, &
                  0.462171912070421929759075029931_dp, 0.395423852042975057677093954594_dp, &
                  0.326516124465411512197156589375_dp, 0.255825079342879083966414791019_dp, &
                  0.183736806564854550852755639381_dp, 0.110645027208519868349122545003_dp, &
                  0.036948943165351775813095980038_dp /

    REAL (KIND=dp), DIMENSION(20)              :: xmui20  ! L = 39

    DATA xmui20 / 0.998237709710559200349622702421_dp, 0.990726238699457006453054352221_dp, &
                  0.977259949983774262663370283713_dp, 0.957916819213791655804540999453_dp, &
                  0.932812808278676533360852166845_dp, 0.902098806968874296728253330868_dp, &
                  0.865959503212259503820781808355_dp, 0.824612230833311663196320230666_dp, &
                  0.778305651426519387694971545506_dp, 0.727318255189927103280996451755_dp, &
                  0.671956684614179548379354514961_dp, 0.612553889667980237952612450231_dp, &
                  0.549467125095128202075931305530_dp, 0.483075801686178712908566574245_dp, &
                  0.413779204371605001524879745804_dp, 0.341994090825758473007492481179_dp, &
                  0.268152185007253681141184344809_dp, 0.192697580701371099715516852065_dp, &
                  0.116084070675255208483451284408_dp, 0.038772417506050821933193444025_dp /

    REAL (KIND=dp), DIMENSION(19)              :: xmui19  ! L = 37

    DATA xmui19 / 0.998049930535687619812847076552_dp, 0.989739454266385571944425891106_dp, &
                  0.974846328590153507640885540946_dp, 0.953466330933529595670542153117_dp, &
                  0.925741332048584396825109562213_dp, 0.891855739004632216794937031974_dp, &
                  0.852035021932362188859649711247_dp, 0.806544167605316815551565375980_dp, &
                  0.755685903753970680737738096779_dp, 0.699798680379184355912825840627_dp, &
                  0.639254415829681707180344869500_dp, 0.574456021047807081132927500924_dp, &
                  0.505834717927931103240533716468_dp, 0.433847169432376484373256162518_dp, &
                  0.358972440479435013256717621457_dp, 0.281708809790165261360114071419_dp, &
                  0.202570453892116703203977781515_dp, 0.122084025337867419869612381329_dp, &
                  0.040785147904578239913316432330_dp /

    REAL (KIND=dp), DIMENSION(18)              :: xmui18  ! L = 35

    DATA xmui18 / 0.997830462484085836198828345581_dp, 0.988586478902212238073395107877_dp, &
                  0.972027691049697949335605048431_dp, 0.948272984399507545202474114945_dp, &
                  0.917497774515659066075867223026_dp, 0.879929800890397131982427337466_dp, &
                  0.835847166992475306418820872499_dp, 0.785576230132206512827768965080_dp, &
                  0.729489171593556582090260479006_dp, 0.668001236585521062097191344341_dp, &
                  0.601567658135980535079944975074_dp, 0.530680285926245161640709080122_dp, &
                  0.455863944433420267207217669327_dp, 0.377672547119689216322736389670_dp, &
                  0.296684995344028270503239047745_dp, 0.213500892316865578943226228903_dp, &
                  0.128736103809384788651993388005_dp, 0.043018198473708607226968978228_dp /

    REAL (KIND=dp), DIMENSION(17)              :: xmui17  ! L = 33

    DATA xmui17 / 0.997571753790841919243372437455_dp, 0.987227816406309485049750431100_dp, &
                  0.968708262533344281764646573056_dp, 0.942162397405107091631676025461_dp, &
                  0.907809677718324468800899889019_dp, 0.865934638334564469263572090671_dp, &
                  0.816884227900933664591578906587_dp, 0.761064876629873014187408968979_dp, &
                  0.698939113216262907933000106576_dp, 0.631021727080528545317775755519_dp, &
                  0.557875500669746642736459886216_dp, 0.480106545190327034194102680507_dp, &
                  0.398359277758645940631494752932_dp, 0.313311081339463247458316765651_dp, &
                  0.225666691616449483868641180934_dp, 0.136152357259182975894428824331_dp, &
                  0.045509821953102542749075670852_dp /

    REAL (KIND=dp), DIMENSION(16)              :: xmui16  ! L = 31

    DATA xmui16 / 0.997263861849481563544981128665_dp, 0.985611511545268335400175044631_dp, &
                  0.964762255587506430773811928118_dp, 0.934906075937739689170919134835_dp, &
                  0.896321155766052123965307243720_dp, 0.849367613732569970133693004968_dp, &
                  0.794483795967942406963097298970_dp, 0.732182118740289680387426665091_dp, &
                  0.663044266930215200975115168663_dp, 0.587715757240762329040745476402_dp, &
                  0.506899908932229390023747474378_dp, 0.421351276130635345364119436172_dp, &
                  0.331868602282127649779916805730_dp, 0.239287362252137074544603209166_dp, &
                  0.144471961582796493485186373599_dp, 0.048307665687738316234812570441_dp /

    REAL (KIND=dp), DIMENSION(15)              :: xmui15  ! L = 29

    DATA xmui15 / 0.996893484074649540271630050918_dp, 0.983668123279747209970032581605_dp, &
                  0.960021864968307512216871025581_dp, 0.926200047429274325879324277080_dp, &
                  0.882560535792052681543116462530_dp, 0.829565762382768397442898119732_dp, &
                  0.767777432104826194917977340974_dp, 0.697850494793315796932292388026_dp, &
                  0.620526182989242861140477556431_dp, 0.536624148142019899264169793311_dp, &
                  0.447033769538089176780609900322_dp, 0.352704725530878113471037207089_dp, &
                  0.254636926167889846439805129817_dp, 0.153869913608583546963794672743_dp, &
                  0.051471842555317695833025213166_dp /

    REAL (KIND=dp), DIMENSION(14)              :: xmui14  ! L = 27

    DATA xmui14 / 0.996442497573954449950436390483_dp, 0.981303165370872753694559945808_dp, &
                  0.954259280628938197254101839705_dp, 0.915633026392132073869689423330_dp, &
                  0.865892522574395048942254567380_dp, 0.805641370917179171447885955425_dp, &
                  0.735610878013631772028144510293_dp, 0.656651094038864961219898176507_dp, &
                  0.569720471811401719308003283356_dp, 0.475874224955118261034411847667_dp, &
                  0.376251516089078710221357209556_dp, 0.272061627635178077676826356126_dp, &
                  0.164569282133380771281471777891_dp, 0.055079289884034270426516527342_dp /

    REAL (KIND=dp), DIMENSION(13)              :: xmui13  ! L = 25

    DATA xmui13 / 0.995885701145616929003216959323_dp, 0.978385445956470991100580354312_dp, &
                  0.947159066661714250135915283518_dp, 0.902637861984307074217665599231_dp, &
                  0.845445942788498018797507061468_dp, 0.776385948820678856192967247242_dp, &
                  0.696427260419957264863813913729_dp, 0.606692293017618063231978746917_dp, &
                  0.508440714824505717695703064726_dp, 0.403051755123486306481077377099_dp, &
                  0.292004839485956895142835382078_dp, 0.176858820356890183969057748418_dp, &
                  0.059230093429313207093718575198_dp /

    REAL (KIND=dp), DIMENSION(12)              :: xmui12  ! L = 23

    DATA xmui12 / 0.995187219997021360179997409700_dp, 0.974728555971309498198391993008_dp, &
                  0.938274552002732758523649001709_dp, 0.886415527004401034213154341982_dp, &
                  0.820001985973902921953949872670_dp, 0.740124191578554364243828103100_dp, &
                  0.648093651936975569252495786911_dp, 0.545421471388839535658375617218_dp, &
                  0.433793507626045138487084231913_dp, 0.315042679696163374386793291320_dp, &
                  0.191118867473616309158639820757_dp, 0.064056892862605626085043082625_dp /

    REAL (KIND=dp), DIMENSION(11)              :: xmui11  ! L = 21

    DATA xmui11 / 0.994294585482399292073031421161_dp, 0.970060497835428727123950986765_dp, &
                  0.926956772187174000520692939259_dp, 0.865812577720300136536425637019_dp, &
                  0.787816805979208162004277955408_dp, 0.694487263186682780050689835762_dp, &
                  0.587640403506911592958876927639_dp, 0.469355837986757026406330710966_dp, &
                  0.341935820892084225158147420427_dp, 0.207860426688221285478846533920_dp, &
                  0.069739273319722221213841796119_dp /

    REAL (KIND=dp), DIMENSION(10)              :: xmui10  ! L = 19

    DATA xmui10 / 0.993128599185094924786122388471_dp, 0.963971927277913791267666131197_dp, &
                  0.912234428251325905867752441203_dp, 0.839116971822218823394529061702_dp, &
                  0.746331906460150792614305070356_dp, 0.636053680726515025452836696226_dp, &
                  0.510867001950827098004364050955_dp, 0.373706088715419560672548177025_dp, &
                  0.227785851141645078080496195369_dp, 0.0765265211334973337546404093988_dp /

    REAL (KIND=dp), DIMENSION(9)              :: xmui09  ! L = 17

    DATA xmui09 / 0.991565168420930946730016004706_dp, 0.9558239495713977551811958929298_dp, &
                  0.892602466497555739206060591127_dp, 0.8037049589725231156824174550146_dp, &
                  0.691687043060353207874891081289_dp, 0.5597708310739475346078715485253_dp, &
                  0.411751161462842646035931793833_dp, 0.2518862256915055095889728548779_dp, &
                  0.084775013041735301242261852935784_dp /

    REAL (KIND=dp), DIMENSION(8)              :: xmui08  ! L = 15

    DATA xmui08 / 0.989400934991649932596154173450_dp, 0.944575023073232576077988415535_dp, &
                  0.865631202387831743880467897712_dp, 0.755404408355003033895101194847_dp, &
                  0.617876244402643748446671764049_dp, 0.458016777657227386342419442984_dp, &
                  0.281603550779258913230460501460_dp, 0.095012509837637440185319335425_dp /

    REAL (KIND=dp), DIMENSION(7)              :: xmui07  ! L = 13

    DATA xmui07 / 0.986283808696812338841597266704_dp, 0.928434883663573517336391139378_dp, &
                  0.827201315069764993189794742650_dp, 0.687292904811685470148019803019_dp, &
                  0.515248636358154091965290718551_dp, 0.319112368927889760435671824168_dp, &
                  0.108054948707343662066244650219835 /

    REAL (KIND=dp), DIMENSION(6)              :: xmui06  ! L = 11

    DATA xmui06 / 0.981560634246719250690549090149_dp, 0.904117256370474856678465866119_dp, &
                  0.769902674194304687036893833213_dp, 0.587317954286617447296702418941_dp, &
                  0.367831498998180193752691536644_dp, 0.125233408511468915472441369464_dp /

    REAL (KIND=dp), DIMENSION(5)              :: xmui05  ! L = 09

    DATA xmui05 / 0.973906528517171720077964012084_dp, 0.865063366688984510732096688423_dp, &
                  0.679409568299024406234327365115_dp, 0.433395394129247190799265943166_dp, &
                  0.148874338981631210884826001130_dp /

    REAL (KIND=dp), DIMENSION(4)              :: xmui04  ! L = 07

    DATA xmui04 / 0.960289856497536231683560868569_dp, 0.796666477413626739591553936476_dp, &
                  0.525532409916328985817739049189_dp, 0.183434642495649804939476142360_dp /

    REAL (KIND=dp), DIMENSION(3)              :: xmui03  ! L = 05

    DATA xmui03 / 0.932469514203152027812301554494_dp, 0.6612093864662645136613995950199_dp, &
                  0.238619186083196908630501721681_dp /

    REAL (KIND=dp), DIMENSION(2)              :: xmui02  ! L = 03

    DATA xmui02 / 0.861136311594052575223946488892_dp, 0.339981043584856264802665759103_dp /

    REAL (KIND=dp), DIMENSION(1)              :: xmui01  ! L = 01

    DATA xmui01 / 0.577350269189625764509148780501_dp /

    REAL (KIND=dp)                            :: ymui

    !==================================================================================
    ! Compute Legendre polynomials for the radiative transfer algorithm
    !==================================================================================
    ALLOCATE (xmui_M(nmui))
    ALLOCATE (AngleLeg_deg(nmui))

    SELECT CASE (nmui)
       CASE (1)
          xmui_M = xmui01
       CASE (2)
          xmui_M = xmui02
       CASE (3)
          xmui_M = xmui03
       CASE (4)
          xmui_M = xmui04
       CASE (5)
          xmui_M = xmui05
       CASE (6)
          xmui_M = xmui06
       CASE (7)
          xmui_M = xmui07
       CASE (8)
          xmui_M = xmui08
       CASE (9)
          xmui_M = xmui09
       CASE (10)
          xmui_M = xmui10
       CASE (11)
          xmui_M = xmui11
       CASE (12)
          xmui_M = xmui12
       CASE (13)
          xmui_M = xmui13
       CASE (14)
          xmui_M = xmui14
       CASE (15)
          xmui_M = xmui15
       CASE (16)
          xmui_M = xmui16
       CASE (17)
          xmui_M = xmui17
       CASE (18)
          xmui_M = xmui18
       CASE (19)
          xmui_M = xmui19
       CASE (20)
          xmui_M = xmui20
       CASE (21)
          xmui_M = xmui21
       CASE (22)
          xmui_M = xmui22
       CASE (23)
          xmui_M = xmui23
       CASE (24)
          xmui_M = xmui24
       CASE (25)
          xmui_M = xmui25
       CASE (26)
          xmui_M = xmui26
       CASE (27)
          xmui_M = xmui27
       CASE (28)
          xmui_M = xmui28
       CASE (29)
          xmui_M = xmui29
       CASE (30)
          xmui_M = xmui30
       CASE (31)
          xmui_M = xmui31
       CASE (32)
          xmui_M = xmui32
       CASE DEFAULT
          PRINT *, " nmui =", nmui, " is not allowed"
    END SELECT

    ! Angle in degrees of Legendre decomposition
    DO i = 1, nmui ! angle
       AngleLeg_deg(i) = ACOS(xmui_M(i)) * 180.0_dp / xpi
    ENDDO

    ! STEP 1 - Compute P_l(mu_i) - see recurence in Num Rec
    ! This is just (-1)**l
    ALLOCATE (xmunl(0:lleg))
    xmunl(0) = 1.0_dp
    DO l = 1, lleg
       xmunl(l) = - xmunl(l-1)
    ENDDO

    ALLOCATE (pli(0:lleg,-nmui:nmui))
    pli(0,:) = 1.0_dp
    DO i = 1, nmui
       ymui = xmui_M(i)
       pli(1,-i) = -ymui
       pli(1,i) = ymui
       DO l = 1, lleg-1
          pli(l+1,-i) = (-ymui*(2.0_dp*l+1.0_dp) * pli(l,-i) - l * pli(l-1,-i)) / (l+1.0_dp)
          pli(l+1,i) = xmunl(l+1) * pli(l+1,-i)
       ENDDO
    ENDDO

    !==================================================================================
    ! Information written in the .def file
    !==================================================================================
    WRITE(iwrit2, '("  ------------------------------------------------------")')
    WRITE(iwrit2, '("  Statistics of the rf_wl grid (number of points)")')
    nwlg_incr = 0

    !==================================================================================
    ! INITIALIZATION of the radiation field grid
    !==================================================================================
    ! ---------------------------------------
    ! Set initial grid step - BG 04/2013
    ! * 100 points (cf nwlg in PXDR_RF_TOOLS)
    !   from wlth0 to wlthm regardless of the
    !   inclusion of X ray transfer
    ! ---------------------------------------
    wlthm = 2.2e9_dp
    nwlg = 100
    pwlg = EXP(LOG(wlthm / wlth0) / DBLE(nwlg))
    PRINT *, "  wlth0 =", wlth0, " wlthm =", wlthm, " pwlg =", pwlg
    ! ---------------------------------------
    ! Initialisation of wavelengths
    ! * rf_wl%Ord is a permutation of
    !   (rf_wl%lower : rf_wl%upper) giving
    !   the order in which rf_wl%Val have
    !   been entered - see doc RF_Tools (FLP)
    ! ---------------------------------------
    IF (.NOT. ASSOCIATED(rf_wl%Val)) THEN
       CALL RF_CREATE (rf_wl, 0, nwlg)
       rf_wl%Val = wlth0 * ( pwlg**(/ (ii, ii=0,nwlg) /) )
       ! The last wavelength must be equal to wlthm and not an approximation of it
       rf_wl%Val(nwlg) = wlthm
       rf_wl%Ord = (/ (ii, ii=0,nwlg) /)
    ENDIF
    nwlg_incr = nwlg - nwlg_incr
    WRITE(iwrit2, '("     - default initial grid ...................... ",I5)') nwlg_incr
    nwlg_incr = nwlg

    ! Radiation field read in a file
    IF (srcpp(1:2) == 'F_') THEN
       fichier = TRIM(data_dir)//'Astrodata/'//TRIM(ADJUSTL(srcpp))
       WRITE(iscrn,*) "Opening file : ", TRIM(fichier)
       OPEN(iread2, file = fichier, status = 'old')
       READ (iread2,*) rstar
       READ (iread2,*) teff
       rstar = rstar * rsun
       READ (iread2,*) nb_pt
       READ (iread2,*)
       PRINT *, nb_pt

       ! Warning: for historical reasons data in file are per nanometers
       DO ii = 1, nb_pt
          READ (iread2,*) wl, rf
          wl = wl * 10.0_dp      ! convert to Angstrom
          IF (wl < wlth0) CYCLE
          CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)   ! add wl in grid
          nwlg = nwlg + add_1
       ENDDO
       CLOSE (iread2)
    ENDIF
    nwlg_incr = nwlg - nwlg_incr
    WRITE(iwrit2, '("     - input external radiation field ............ ",I5)') nwlg_incr
    nwlg_incr = nwlg

    !================================================================================
    ! Add all "special values"
    !================================================================================

    !---  Habing field limit --------------------------------------------------------
    CALL RF_EXTAND (rf_wl, wl_hab, add_1, iloc, iord)
    nwlg = nwlg + add_1

    !--- 6 eV field limit -----------------------------------------------------------
    CALL RF_EXTAND (rf_wl, wl_6ev, add_1, iloc, iord)
    nwlg = nwlg + add_1

    !--- Draine Field cut-off ------------------------------------------------------
    CALL RF_EXTAND (rf_wl, wl_cut_D, add_1, iloc, iord)
    nwlg = nwlg + add_1

    !--- Fitzpatrick & Massa extinction curve ---------------------------------------
    wl_cut_F = 1.0e4_dp / 5.9_dp
    CALL RF_EXTAND (rf_wl, wl_cut_F, add_1, iloc, iord)
    nwlg = nwlg + add_1

    ! Special : If the name of the line of sight is sig1.9, we force grains cross section
    !           to be 1.9E-21 cm-2 below 1070 Ang (9.346 microm-1) and then use the Galactic curve
    IF (los_ext == 'sig1.9') THEN
       wl_cut_F2 = 1.0e4_dp / 9.346_dp
       CALL RF_EXTAND (rf_wl, wl_cut_F2, add_1, iloc, iord)
       nwlg = nwlg + add_1
    ENDIF

    !--- Upper limit of Fitzpatrick and Massa fit
    CALL RF_EXTAND (rf_wl, FP_up_wl, add_1, iloc, iord)
    nwlg = nwlg + add_1

    !--- B and V Photometric band value
    CALL RF_EXTAND (rf_wl, B_band_wl, add_1, iloc, iord)
    nwlg = nwlg + add_1
    CALL RF_EXTAND (rf_wl, V_band_wl, add_1, iloc, iord)
    nwlg = nwlg + add_1

    nwlg_incr = nwlg - nwlg_incr
    WRITE(iwrit2, '("     - special values (Habing, Draine, ...) ...... ",I5)') nwlg_incr
    nwlg_incr = nwlg

    !--- Wavelength of transitions for detailed balance species
    DO k = 1, nspl
       DO ii = 1, spec_lin(k)%ntr
          wlth = spec_lin(k)%lac(ii) * 1.0e8_dp
          IF (wlth >= wlth0) THEN
             CALL RF_EXTAND (rf_wl, wlth, add_1, iloc, iord)
             nwlg = nwlg + add_1
          ENDIF
       ENDDO
       nwlg_incr = nwlg - nwlg_incr
       WRITE(iwrit2, '("     - detailed balance: ",A8," ................ ",I5)') spec_lin(k)%nam, nwlg_incr
       nwlg_incr = nwlg
    ENDDO

    !================================================================================
    ! Add the photodestruction cross sections (in UV domain) - files in data/Sections
    !================================================================================
    ! Loop on all photodestruction reactions computed with cross sections
    DO k = 1, nphot_i - i_mandat
       fichier = TRIM(data_dir)//'Sections/'//TRIM(csect_files(k))//".dat"
       WRITE(iscrn,*) "Opening file : ",TRIM(fichier)
       OPEN(iread2, file = fichier, status = 'old')

       ! -----------------------------------------------
       ! Count number of point above threshold
       ! -----------------------------------------------
       READ(iread2,"(2/)")
       READ(iread2,*) nb_pt  ! number of points in the file
       READ(iread2,*)

       ! -----------------------------------------------
       ! Remove points below Lyman limit
       ! -----------------------------------------------
       np_before = 0
       DO ii = 1, nb_pt
          READ (iread2,*) wl
          IF (wl < wl_lym) THEN
             np_before = np_before + 1
          ENDIF
       ENDDO
       nb_pt = nb_pt - np_before

       ! -----------------------------------------------
       ! Create the table containing the cross sections
       ! -----------------------------------------------
       IF (ASSOCIATED(phdest(k)%rf_esp%Val)) THEN
         CALL RF_FREE(phdest(k)%rf_esp)
       ENDIF
       CALL RF_CREATE (phdest(k)%rf_esp,1,nb_pt+1)

       ! -----------------------------------------------
       ! Jump all the points below wl_lym except for the
       ! last one (is used for interpolation at wl_lym)
       ! -----------------------------------------------
       REWIND(iread2)
       READ(iread2,"(4/)")
       IF (np_before > 1) THEN
          DO ii = 1, np_before-1
             READ(iread2,*)
          ENDDO
       ENDIF

       ! -----------------------------------------------
       ! Set the lower limit for integration
       ! Add the two first points in rf_wl
       !   1-wl_lym -> interpolation of c.s. if possible
       !            -> or c.s. equals the following data
       !   2-first point above wl_lym
       ! Cross section are stored in phdest%rf_esp
       !
       ! note: I really dislike the idea of adding a
       !       point at 912 Å regardless of the first
       !       position of ionization data - BG 04/2013
       !       e.g. : look at CH file to see the problem
       !              of extrapolating from 912 to 1458
       !       It was even worse when the first point
       !       was systematically removed (PDR v1.5.2)
       ! -----------------------------------------------
       phdest(k)%wl_cut_l_esp = wl_lym
       IF (np_before > 0) THEN
          READ (iread2,*) wl1, sig1
          READ (iread2,*) wl2, sig2
          CALL RF_EXTAND (rf_wl, wl_lym, add_1, iloc, iord)
          nwlg = nwlg + add_1
          phdest(k)%rf_esp%Val(1) = (sig1*(wl2-wl_lym)+sig2*(wl_lym-wl1))/(wl2-wl1)
          phdest(k)%rf_esp%Ord(1) = iord
          CALL RF_EXTAND (rf_wl, wl2, add_1, iloc, iord)
          nwlg = nwlg + add_1
          phdest(k)%rf_esp%Val(2) = sig2
          phdest(k)%rf_esp%Ord(2) = iord
       ELSE
          READ (iread2,*) wl2, sig2
          CALL RF_EXTAND (rf_wl, wl_lym, add_1, iloc, iord)
          nwlg = nwlg + add_1
          phdest(k)%rf_esp%Val(1) = sig2
          phdest(k)%rf_esp%Ord(1) = iord
          CALL RF_EXTAND (rf_wl, wl2, add_1, iloc, iord)
          nwlg = nwlg + add_1
          phdest(k)%rf_esp%Val(2) = sig2
          phdest(k)%rf_esp%Ord(2) = iord
       ENDIF

       ! -----------------------------------------------
       ! Read all the following data
       ! Extend rf_wl and store the cross sections
       ! -----------------------------------------------
       IF(nb_pt > 1) THEN
          DO ii = 3, nb_pt+1
             READ (iread2,*) wl1, sig1
             CALL RF_EXTAND (rf_wl, wl1, add_1, iloc, iord)
             nwlg = nwlg + add_1
             phdest(k)%rf_esp%Val(ii) = sig1
             phdest(k)%rf_esp%Ord(ii) = iord
          ENDDO
          phdest(k)%wl_cut_u_esp = wl1
       ELSE
          phdest(k)%wl_cut_u_esp = wl2
       ENDIF
       CLOSE(iread2)
       CALL RF_REVERS (wl_pointer)

       nwlg_incr = nwlg - nwlg_incr
       WRITE(iwrit2, '("     - photodestruction: ",A8," ................ ",I5)') phdest(k)%r1, nwlg_incr
       nwlg_incr = nwlg
    ENDDO

    !================================================================================
    ! Add an external radiation field
    !================================================================================
    ! In addition to an isotropic contribution (scales by radm and/or radp)
    ! we add a normal source. Three cases are possible:
    !    1 - srcpp = "F_XXXXX" : Name of a file containing the RF (e.g. Kurucz)
    !    2 - srcpp = "PP"      : Stands for "Plane Parallel" => No isotropic
    !                            contribution, but same energy density at the edge
    !    3 - srcpp = Star type : Analytic Black Body with prescribed spectral type
    !
    ! This normal RF subtends a solid angle such that "geom" = 1 - cos(theta)
    !
    ! Flag "F_Norm_RF" indicates where the radiations comes from:
    !    F_Norm_RF = 1 : from left side (or "m" - minus)
    !    F_Norm_RF = 2 : from right side (or "p" - plus)

    geom = 0.0_dp
    IF (F_Norm_RF /= 0) THEN
       ! Radiation field read in a file
       IF (srcpp(1:2) == 'F_') THEN
          PRINT *, srcpp
          fichier = TRIM(data_dir)//'Astrodata/'//TRIM(ADJUSTL(srcpp))
          WRITE(iscrn,*) "Opening file : ", TRIM(fichier)
          OPEN(iread2, file = fichier, status = 'old')
          READ (iread2,*) rstar
          READ (iread2,*) teff
          PRINT *, rstar, teff
          rstar = rstar * rsun
          READ (iread2,*) nb_pt
          READ (iread2,*)
          PRINT *, nb_pt

          IF (ALLOCATED(star_rf)) THEN
             DEALLOCATE(star_rf)
          ENDIF
          ALLOCATE (star_rf(nb_pt,2))

          ! Warning: for historical reasons data in file are per nanometers
          DO ii = 1, nb_pt
             READ (iread2,*) wl, rf
             wl = wl * 10.0_dp      ! convert to Angstrom
             rf = rf * 0.1_dp       ! convert to Angstr-1
             star_rf(ii,1) = wl
             star_rf(ii,2) = rf
          ENDDO
          CLOSE (iread2)

       ELSE IF (srcpp(1:2) /= 'PP') THEN

          CALL STAR

       ENDIF

       ! Dilution factor of the star emissivity as seen from the cloud
       ! geom is (1 - cos(theta_0)) so cos(theta_0) is (1.0 - geom)
       ! A second order development is adequate in most cases.
       !
       ! geom = 1 in special case "PP": see analytical derivation in notes
       IF (F_RF_PP == 1) THEN
          geom = 1.0_dp
       ELSE
          geom = 0.5_dp * (rstar/(ABS(d_sour)*pccm))**2
       ENDIF

    ENDIF

    WRITE (iscrn,*) " F_Norm_RF = ", F_Norm_RF, " F_RF_PP = ", F_RF_PP

    !================================================================================
    ! Read the dust properties from DUSTEM
    !================================================================================
    IF (F_DUST_P == 1) THEN
       ! -----------------------------------------------
       ! First, read wavelength grid (always useful)
       ! -----------------------------------------------
       fichier = "../dustem40/oprop/LAMBDA.DAT"
       OPEN(iread2, file = fichier, status = 'old')
       READ(iread2,'(///,i20)') nwl_ir
       DO ii = 1, nwl_ir
          READ (iread2,*) wl
          wl = wl * 1.0e4_dp
          IF (wl < wlth0) THEN
             CYCLE
          ENDIF
          CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
          nwlg = nwlg + add_1
       ENDDO
       CLOSE(iread2)

       ! Initial call to DUSTEM to create an input file
       PRINT *, " Run initial DUSTEM:"
       CALL EXECUTE_COMMAND_LINE(lancer_dustem)

       PRINT *, " READ DUSTEM absorption, diffusion and emissivities"
       fichier = TRIM(data_dir)//"EXTINCTION_DUSTEM.RES"
       WRITE(iscrn,*) "Opening file : ", TRIM(fichier)
       OPEN(iread2, file = fichier, status = 'old')
       READ (iread2,*) nwl_ir
       PRINT *, " nwl_ir:", nwl_ir
       IF (ALLOCATED(wl_ir)) THEN
          DEALLOCATE (wl_ir)
          DEALLOCATE (wl_ir_l)
       ENDIF
       ALLOCATE (wl_ir(0:nwl_ir+1))
       ALLOCATE (wl_ir_l(0:nwl_ir+1))

       IF (ALLOCATED (dustem_in%abso)) DEALLOCATE (dustem_in%abso)
       IF (ALLOCATED (dustem_in%scat)) DEALLOCATE (dustem_in%scat)
       IF (ALLOCATED (dustem_in%emis)) DEALLOCATE (dustem_in%emis)
       IF (ALLOCATED (dustem_in%gcos)) DEALLOCATE (dustem_in%gcos)
       IF (ALLOCATED (dustem_in%albe)) DEALLOCATE (dustem_in%albe)
       ALLOCATE (dustem_in%abso(0:nwl_ir+1,0:1))
       ALLOCATE (dustem_in%scat(0:nwl_ir+1,0:1))
       ALLOCATE (dustem_in%emis(0:nwl_ir+1,0:1))
       ALLOCATE (dustem_in%gcos(0:nwl_ir+1,0:1))
       ALLOCATE (dustem_in%albe(0:nwl_ir+1,0:1))

       IF (ALLOCATED (dustem_Qabs)) DEALLOCATE (dustem_Qabs)
       IF (ALLOCATED (dustem_Qsca)) DEALLOCATE (dustem_Qsca)
       ALLOCATE (dustem_Qabs(0:nwl_ir+1,npg,0:1))
       ALLOCATE (dustem_Qsca(0:nwl_ir+1,npg,0:1))
       dustem_Qabs = 0.0_dp
       dustem_Qsca = 0.0_dp

       DO ii = 1, nwl_ir
          READ (iread2,*) wl, r1, r2, rf, r3, r4
          wl = wl * 1.0e4_dp            ! convert to Angstrom
          wl_ir(ii) = wl
          wl_ir_l(ii) = LOG(wl)
          IF (wl > wlth0) THEN
             CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
             nwlg = nwlg + add_1
          ENDIF
          ! bug correction - dustem_in tables should
          ! be in LOG scale for extrapolation by
          ! the EXTEND function - BG 04/2013
          ! see PDR_DUSTEM function
          dustem_in%abso(ii,0:1) = LOG(r1)
          dustem_in%scat(ii,0:1) = LOG(r2)
          dustem_in%emis(ii,0:1) = LOG(rf / (4.0_dp * xpi * wl))
          dustem_in%albe(ii,0:1) = LOG(r3)
          dustem_in%gcos(ii,0:1) = r4
       ENDDO
       CLOSE (iread2)
       PRINT *, "  End reading DUSTEM file"
       wl_ir  (0)        = wlth0
       wl_ir_l(0)        = LOG(wlth0)
       wl_ir  (nwl_ir+1) = wlthm
       wl_ir_l(nwl_ir+1) = LOG(wlthm)
       CALL EXTEND_DUST(1)
       nwlg_incr = nwlg - nwlg_incr
       WRITE(iwrit2, '("     - DUSTEM dust coefficients grid ............. ",I5)') nwlg_incr
       nwlg_incr = nwlg
    ENDIF

    !================================================================================
    ! Grains absorption cross sections (interpolated from Draines data for each rad.)
    !================================================================================
    IF (ALLOCATED(sigabs)) THEN
       DO mm = 1, npg
          CALL RF_FREE(sigabs(mm))
       ENDDO
       DEALLOCATE(sigabs)
    ENDIF
    ALLOCATE(sigabs(npg))
    DO mm = 1, npg
       NULLIFY (sigabs(mm)%val,sigabs(mm)%ord)                     !SG
    ENDDO
    DO mm = 1, npg
       IF (ASSOCIATED(sigabs(mm)%Val)) THEN
          CALL RF_FREE(sigabs(mm))
       ENDIF
       CALL RF_CREATE(sigabs(mm),1,jwlmax_gr-jwlmin_gr+2)
    ENDDO

    ! ------------------------------------------------
    ! Modification of jwlmin_gr compared to PDR v1.5.2
    ! See MIXGRSI in PXDR_DRAINE_DUST.f90 - BG 04/2013
    ! ------------------------------------------------
    DO ii = jwlmin_gr, jwlmax_gr+1
       IF (ii == jwlmin_gr) THEN
          wl = wlth0
       ELSE IF (ii == jwlmax_gr+1) THEN
          wl = wlthm
       ELSE
          wl = wavgrd(ii)
       ENDIF
       CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
       nwlg = nwlg + add_1
       DO mm = 1, npg
          sigabs(mm)%Ord(ii-jwlmin_gr+1) = iord
       ENDDO
    ENDDO
    CALL RF_REVERS (wl_pointer)
    nwlg_incr = nwlg - nwlg_incr
    WRITE(iwrit2, '("     - Draine dust coefficients grid ............. ",I5)') nwlg_incr
    nwlg_incr = nwlg

    !================================================================================
    ! Efficiency of the photoelectric effect - fraction of energy that goes in ph-e-
    !================================================================================
    DO mm = 1, npg
       DO i = 0, noptm
          NULLIFY (eff_phel(mm,i)%val,eff_phel(mm,i)%ord)
       ENDDO
    ENDDO
    DO mm = 1, npg
       DO i = 0, noptm
          IF (ASSOCIATED(eff_phel(mm,i)%Val)) THEN
             CALL RF_FREE(eff_phel(mm,i))
          ENDIF
          CALL RF_CREATE(eff_phel(mm,i),1,jwlmax_gr-jwlmin_gr+2)
       ENDDO
    ENDDO
    DO ii = jwlmin_gr, jwlmax_gr+1
       DO mm = 1, npg
          DO i = 0, noptm
             eff_phel(mm,i)%Ord(ii-jwlmin_gr+1) = sigabs(mm)%Ord(ii-jwlmin_gr+1)
             eff_phel(mm,i)%Val(ii-jwlmin_gr+1) = 1.0_dp / xpi * ( xpi / 2.0_dp &
             - ATAN(2.e-2_dp*(rf_wl%Val(wl_pointer(eff_phel(mm,i)%Ord(ii-jwlmin_gr+1))) &
             -5.e2_dp)))
          ENDDO
       ENDDO
    ENDDO

    !================================================================================
    ! ADD LINE WAVELENGTH IN THE GRID
    !================================================================================
    ! JLB 19 II 2013 - In initial WL grid, use only line center and not profiles
    !--- Atomic hydrogen (Lyman lines) -----------------------------------------
    IF (jlyman == 1) THEN
       DO ii = 1, nraih
          wl0 = praih(ii,2)
          CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
          nwlg = nwlg + add_1
       ENDDO
    ENDIF
    nwlg_incr = nwlg - nwlg_incr
    WRITE(iwrit2, '("     - HI Lyman lines ............................ ",I5)') nwlg_incr
    nwlg_incr = nwlg

    !--- H2 lines, only lowest levels ---------------------------------

    DO ii = 1, nrh2bt
       nvl = irah2b(ii,1)
       njl = irah2b(ii,2)
       lev = lev_h2(nvl,njl)
       IF (lev >= lfgkh2) CYCLE
       wl0 = prah2b(ii,2)
       CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1
    ENDDO

    DO ii = 1, nrh2ct
       nvl = irah2c(ii,1)
       njl = irah2c(ii,2)
       lev = lev_h2(nvl,njl)
       IF (lev >= lfgkh2) CYCLE
       wl0 = prah2c(ii,2)
       CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1
    ENDDO

    nwlg_incr = nwlg - nwlg_incr
    WRITE(iwrit2, '("     - H2 B / C lines ............................ ",I5)') nwlg_incr
    nwlg_incr = nwlg

    !--- HD lines, only lowest levels ---------------------------------

    DO ii = 1, nrhdbt
       nvl = irahdb(ii,1)
       njl = irahdb(ii,2)
       lev = lev_hd(nvl,njl)
       IF (lev >= lfgkhd) CYCLE
       wl0 = prahdb(ii,2)
       CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1
    ENDDO

    DO ii = 1, nrhdct
       nvl = irahdc(ii,1)
       njl = irahdc(ii,2)
       lev = lev_hd(nvl,njl)
       IF (lev >= lfgkhd) CYCLE
       wl0 = prahdc(ii,2)
       CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1
    ENDDO

    nwlg_incr = nwlg - nwlg_incr
    WRITE(iwrit2, '("     - HD B / C lines ............................ ",I5)') nwlg_incr
    nwlg_incr = nwlg

    !---  CO lines, only lowest J --------------------------------

    DO ii = 1, nrco
       njl = iraco(ii,1)
       IF (njl >= jfgkco) CYCLE
       wl0 = praco(ii,2)
       CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1
    ENDDO
    nwlg_incr = nwlg - nwlg_incr
    WRITE(iwrit2, '("     - CO lines .................................. ",I5)') nwlg_incr
    nwlg_incr = nwlg

    !---  13CO lines, only lowest J -------------------------------

    DO ii = 1, nrc13o
       njl = irac13o(ii,1)
       IF (njl >= jfgkc13o) CYCLE
       wl0 = prac13o(ii,2)
       CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1
    ENDDO
    nwlg_incr = nwlg - nwlg_incr
    WRITE(iwrit2, '("     - 13CO lines ................................ ",I5)') nwlg_incr
    nwlg_incr = nwlg

    !---  CO18 lines, only lowest J -------------------------------

    DO ii = 1, nrco18
       njl = iraco18(ii,1)
       IF (njl >= jfgkco18) CYCLE
       wl0 = praco18(ii,2)
       CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1
    ENDDO
    nwlg_incr = nwlg - nwlg_incr
    WRITE(iwrit2, '("     - C18O lines ................................ ",I5)') nwlg_incr
    nwlg_incr = nwlg

    !----------------------------------------------------
    ! Add X-A UV transitions of CO for pumping excitation
    !----------------------------------------------------

    IF (F_CO_UV_pum == 1 .AND. i_co /= 0) THEN
       DO il2 = 1, spec_lin(in_sp(i_co))%use
          DO il1 = 1, lm_CO_A(il2,0)
             wl = lam_CO_A(lm_CO_A(il2,il1),il2)
             IF (wl == 0.0_dp) THEN
                PRINT *, il2, il1, wl
             ENDIF
             CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
             nwlg = nwlg + add_1
          ENDDO
       ENDDO
    ENDIF
    nwlg_incr = nwlg - nwlg_incr
    WRITE(iwrit2, '("     - X-A UV lines of CO ........................ ",I5)') nwlg_incr
    nwlg_incr = nwlg

    WRITE(iwrit2, '("     *** Total initial wavelength grid ........... ",I5)') nwlg_incr
    WRITE(iwrit2, '("  ------------------------------------------------------")')

    PRINT *, " 1st pass:  nwlg =", nwlg

    IF (rf_wl%upper /= nwlg) THEN
       PRINT *, "Probleme"
    ENDIF
    CALL RF_REVERS (wl_pointer)

    !================================================================================
    !****From here, the set of wavelength is complete and should not be modified*****
    !================================================================================

    !-------------------------------------------
    ! Initialisation of external radiation Field
    !-------------------------------------------
    IF (ASSOCIATED(rf_incIm%Val)) THEN
       CALL RF_FREE(rf_incIm)
    ENDIF
    CALL RF_CREATE (rf_incIm, 0, nwlg)
    IF (ASSOCIATED(rf_incIp%Val)) THEN
       CALL RF_FREE(rf_incIp)
    ENDIF
    CALL RF_CREATE (rf_incIp, 0, nwlg)

    !-------------------------------------------
    ! Initialisation of internal radiation Field
    !-------------------------------------------
    IF (ASSOCIATED(rf_u%Val)) THEN
       CALL RF_FREE(rf_u)
    ENDIF
    CALL RF_CREATE (rf_u, 0, nwlg)

    !----------------------------------------------------------------------------------
    ! Fitzpatrick and Massa extinction curves
    ! We use the Fitzpatrick and Massa expressions for the grains extinction
    ! Their expression gives E(lambda-V) / E(B-V)
    ! rf_expou%Val = [1 - 1/Rv * E(lambda-V) / E(B-V)]
    !
    ! Modification 24th june 2010
    ! If the name of the line of sight is sig1.9, we use :
    ! - from 912 to 1070 Ang (9.35 mum-1) : sigma_dust = 1.9E-21 (=> E(lambda-V) / E(B-V) = 8.8648)
    ! - above 1070 Ang                    : Galactic extinction curve
    !----------------------------------------------------------------------------------

    IF (ASSOCIATED(rf_extpou%Val)) THEN
       CALL RF_FREE(rf_extpou)
    ENDIF
    CALL RF_CREATE (rf_extpou, 0, nwlg)
    IF (ASSOCIATED(rf_tmp%Val)) THEN
       CALL RF_FREE(rf_tmp)
    ENDIF
    CALL RF_CREATE (rf_tmp, 0, nwlg)
    rf_tmp%Val = 1.0e4_dp / rf_wl%Val

    ! Create extinction curve from analytical fits
    ! 1 - UV slope and 2200 A bump
    !     lowest cut off at the Lyman limit (BG 04/2013)
    CALL RF_LOCATE (rf_wl, wl_lym, I_cut_L)
    wl_cut = FP_up_wl
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut)
    rf_extpou%Val(I_cut_L:I_cut) = cc1 + cc2 * rf_tmp%Val(I_cut_L:I_cut) &
                                 + cc3 * (rf_tmp%Val(I_cut_L:I_cut) * rf_tmp%Val(I_cut_L:I_cut) &
                                 / ((rf_tmp%Val(I_cut_L:I_cut) * rf_tmp%Val(I_cut_L:I_cut) - y0 * y0)**2 &
                                 + xgam * xgam * rf_tmp%Val(I_cut_L:I_cut) * rf_tmp%Val(I_cut_L:I_cut)))

    ! 2 - Far UV rise
    !     lowest cut off at the Lyman limit (BG 04/2013)
    wl_cut_F = 1.0e4_dp / 5.9_dp
    CALL RF_LOCATE (rf_wl, wl_cut_F, I_cut_F)
    rf_extpou%Val(I_cut_L:I_cut_F) = rf_extpou%Val(I_cut_L:I_cut_F) &
                                   + cc4 * (0.5392_dp * (rf_tmp%Val(I_cut_L:I_cut_F)-5.9_dp)**2 &
                                   + 0.0564_dp * (rf_tmp%Val(I_cut_L:I_cut_F)-5.9_dp)**3)

    ! 3 - IR analytical fit (using Galactic values from FP (2009) ApJ 699, 1209)
    !     Equation (5), used with alpha from BD+61 2365
    wl_cut = 1.0e4_dp
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut)
    wl_cut_F = 4.4e4_dp
    CALL RF_LOCATE (rf_wl, wl_cut_F, I_cut_F)
    rf_extpou%Val(I_cut:I_cut_F) = (0.349_dp + 2.087_dp * rrv) / (1.0_dp &
                                 + (1.0_dp / (0.507_dp*rf_tmp%Val(I_cut:I_cut_F))**2.28_dp)) - rrv

    ! 4 - IR rise at 18 microns as obtained with dustem and with Draine's data
    wl_cut = 4.4e4_dp
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut)
    wl_cut_F = 1.82e5_dp
    CALL RF_LOCATE (rf_wl, wl_cut_F, I_cut_F)
    rf_extpou%Val(I_cut:I_cut_F) = rf_extpou%Val(I_cut) + 0.015_dp * ( (1.0_dp / &
                                 & rf_tmp%Val(I_cut:I_cut_F) - wl_cut/1.0e4_dp) / 10.0_dp )**2.0

    ! 5 - IR bump at 10 microns as obtained with dustem and with Draine's data
    wl_cut = 1.0e4_dp
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut)
    rf_extpou%Val(I_cut:) = rf_extpou%Val(I_cut:) + 2.0_dp*rf_tmp%Val(I_cut:) &
                          * DEXP(-((DLOG(1.0_dp/rf_tmp%Val(I_cut:))-2.3_dp)/0.18_dp)**2.0_dp)

    ! 6 - FIR decreasing queue as obtained with dustem and with Draine's data
    wl_cut = 1.82e5_dp
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut)
    rf_extpou%Val(I_cut:) = (rf_extpou%Val(I_cut) + rrv) * (1.0_dp + rf_tmp%Val(I_cut)**(-1.75_dp)) / &
                          & (1.0_dp + rf_tmp%Val(I_cut:)**(-1.75_dp)) - rrv

    ! 7 - Visible: use 3 linear interpolation from 1 micron to V, from V to B and from B to UV cutoff (Ext = 0 at V and 1 at B)
    wl_cut = 1.0e4_dp
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut)
    wl_cut = V_band_wl
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut_F)
    rf_extpou%Val(I_cut_F:I_cut) = rf_extpou%Val(I_cut) - rf_extpou%Val(I_cut) &
                                 * (rf_tmp%Val(I_cut_F:I_cut) - 1.0_dp) &
                                 / ( 1.0e4_dp / V_band_wl -  1.0_dp)

    I_cut = I_cut_F
    wl_cut = B_band_wl
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut_F)
    rf_extpou%Val(I_cut_F:I_cut) = rf_extpou%Val(I_cut) + (1.0_dp - rf_extpou%Val(I_cut)) &
                                 * (rf_tmp%Val(I_cut_F:I_cut) - 1.0e4_dp / V_band_wl) &
                                 / ( 1.0e4_dp / B_band_wl -  1.0e4_dp / V_band_wl)

    I_cut = I_cut_F
    wl_cut = FP_up_wl
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut_F)
    rf_extpou%Val(I_cut_F:I_cut) = rf_extpou%Val(I_cut) + (rf_extpou%Val(I_cut_F) - 1.0_dp) &
                                 * (rf_tmp%Val(I_cut_F:I_cut) - 1.0e4_dp / B_band_wl) &
                                 / ( 1.0e4_dp / FP_up_wl -  1.0e4_dp / B_band_wl)

    IF (los_ext == "sig1.9") THEN
       wl_cut_F2 = 1.0e4_dp / 9.346_dp
       CALL RF_LOCATE (rf_wl, wl_cut_F2, I_cut_F2)
       rf_extpou%Val(I_cut_L:I_cut_F2) = 8.8648_dp ! value to get grain abs. cross section = 1.9E-21 cm2
    ENDIF

    rf_extpou%Val = 1.0_dp + rf_extpou%Val / rrv
    rf_extpou%Ord = (/ (ii, ii=rf_extpou%lower,rf_extpou%upper) /)

    CALL RF_FREE (rf_tmp)

    !--------------------------------------
    ! Variables used in PDR_DUSTEM - part 1
    !--------------------------------------

    IF (F_DUST_P == 1) THEN
       ! Initialise points where DUSTEM is called
       print *, " nav_dec =", nav_dec
       IF (MOD(npo,nav_dec) == 0) THEN
          nav_dust = 1 + npo / nav_dec
       ELSE
          nav_dust = 2 + npo / nav_dec
       ENDIF
       IF (ALLOCATED(iav_dust)) THEN
          DEALLOCATE (iav_dust)
          DEALLOCATE (av_dust)
       ENDIF
       ALLOCATE (iav_dust(nav_dust))
       ALLOCATE (av_dust(nav_dust))
       DO ii = 1, nav_dust-1
          iav_dust(ii) = nav_dec * (ii-1)
          av_dust(ii) = tau_o(nav_dec * (ii-1))
       ENDDO
       iav_dust(nav_dust) = npo
       av_dust(nav_dust) = tau_o(npo)
    ENDIF

    !--------------------------------------------------------------------------
    ! Compute Planck function over a fixed dust temperature grid
    !--------------------------------------------------------------------------
    DO i = 1, nTd
       IF(ASSOCIATED(rf_Plint(i)%Val)) CALL RF_FREE(rf_Plint(i))
       CALL RF_CREATE (rf_Plint(i), 0, nwlg)
       DO ii = 0, nwlg
          rf_Plint(i)%Val(ii) = BBL (rf_wl%Val(ii), Tdposs(i))
          rf_Plint(i)%Ord(ii) = rf_wl%Ord(ii)
       ENDDO
    ENDDO

    !----------------------------------------------------------------
    ! Set dust properties (absorption, scattering, emission, g factor
    !----------------------------------------------------------------

    IF (ALLOCATED(J_du_pro)) DEALLOCATE(J_du_pro)
    PRINT *, " Enter DUST_PROPERTIES - 1"
    CALL DUST_PROPERTIES

    !--------------------------------------------------------------------------------
    ! A first approximation of distances is needed for the following initialisations
    !--------------------------------------------------------------------------------

    IF (ALLOCATED(d_cm)) THEN
       DEALLOCATE(d_cm)
       DEALLOCATE(dd_cm)
    ENDIF
    ALLOCATE(d_cm(0:npo))
    ALLOCATE(dd_cm(0:npo))
    d_cm(:) = 0.0_dp
    dd_cm(:) = 0.0_dp
    CALL RF_LOCATE (rf_wl, V_band_wl, iloc)
    IF (F_DU_PROP == 0) THEN
       DO i = 1, npo ! d_cm(0) = 0
          dd_cm(i) = (   1.0_dp / ( (du_prop%abso(iloc,0) + du_prop%scat(iloc,0)) * dens_o(i)   ) &
                       + 1.0_dp / ( (du_prop%abso(iloc,0) + du_prop%scat(iloc,0)) * dens_o(i-1) ) &
                     ) * 0.5_dp * dtau_o(i)
          d_cm(i) = d_cm(i-1) + dd_cm(i)
       ENDDO
    ELSE
       DO i = 1, npo ! d_cm(0) = 0
          dd_cm(i) = (   1.0_dp / ( (du_prop%abso(iloc,i)   + du_prop%scat(iloc,i))   * dens_o(i)   ) &
                       + 1.0_dp / ( (du_prop%abso(iloc,i-1) + du_prop%scat(iloc,i-1)) * dens_o(i-1) ) &
                     ) * 0.5_dp * dtau_o(i)
          d_cm(i) = d_cm(i-1) + dd_cm(i)
       ENDDO
    ENDIF

    !----------------------------------------
    ! Set radiation field boundary conditions
    !----------------------------------------

    IF (ALLOCATED(tmp_wl)) DEALLOCATE(tmp_wl)
    ALLOCATE(tmp_wl(0:nwlg))
    tmp_wl(0:nwlg) = rf_wl%Val(0:nwlg)
    IF (ALLOCATED(intJmp)) THEN
       DEALLOCATE (intJmp)
    ENDIF
    ALLOCATE (intJmp(0:npo,0:nwlg))
    intJmp(0:npo,0:nwlg) = 0.0_dp

    CALL RF_BOUND

    !================================================================================
    ! Create a first very crude approximation to the radiation field
    !================================================================================

    IF (ALLOCATED(intJmp)) THEN
       DEALLOCATE (intJmp)
    ENDIF
    IF (ALLOCATED(intJm)) THEN
       DEALLOCATE (intJm, intJp)
    ENDIF
    ALLOCATE (intJm(0:npo,0:nwlg))
    ALLOCATE (intJp(0:npo,0:nwlg))
    ALLOCATE (intJmp(0:npo,0:nwlg))
    ! -----------------------------------------
    ! This loop takes some code from subroutine
    ! DUST_PROPERTIES to compute a dust extinc-
    ! tion coefficient from Draine's data. It
    ! also takes a code from subroutine ABSGAZ
    ! to compute gas extinction coefficients.
    ! This is used to compute a wavelength
    ! dependant optical depth. BG 04/2013
    ! -----------------------------------------
    DO j = 0, nwlg
       wlth = rf_wl%Val(j)
       DO i = jwlmin_gr, jwlmax_gr+1
          IF (wavgrd(i) > wlth) THEN
             il1 = i-1
             EXIT
          ENDIF
       ENDDO
       IF (wavgrd(jwlmax_gr) < wlth) THEN
          il1 = jwlmax_gr
       ENDIF
       ! -----------------------------------------
       ! 1 - absor  = absorption, summed over size
       ! 2 - scater = scattering, summed over size
       !
       ! Extrapolations are required at small and
       ! large wavelengths - BG 04/2013
       ! ** power laws except for le_g
       ! ** linear interpolation otherwise
       ! -----------------------------------------
       absor  = 0.0_dp
       scater = 0.0_dp
       IF (il1 == 0) THEN
          fact = (LOG(wlth) - LOG(wavgrd(il1))) / (LOG(wavgrd(il1+1)) - LOG(wavgrd(il1)))
          DO mm = 1, npg
             Dr_int1 = LOG(Qabs_Dr(il1,mm)) + (LOG(Qabs_Dr(il1+1,mm)) - LOG(Qabs_Dr(il1,mm))) * fact
             absor   = absor + wwmg(mm) * EXP(Dr_int1) * xpi * adust(mm)**2
             Dr_int2 = LOG(Qsca_Dr(il1,mm)) + (LOG(Qsca_Dr(il1+1,mm)) - LOG(Qsca_Dr(il1,mm))) * fact
             scater  = scater + wwmg(mm) * EXP(Dr_int2) * xpi * adust(mm)**2
          ENDDO
       ELSE IF(il1 == jwlmax_gr) THEN
          fact = (LOG(wlth) - LOG(wavgrd(il1))) / (LOG(wavgrd(il1+1)) - LOG(wavgrd(il1)))
          DO mm = 1, npg
             Dr_int1 = LOG(Qabs_Dr(il1,mm)) + (LOG(Qabs_Dr(il1+1,mm)) - LOG(Qabs_Dr(il1,mm))) * fact
             absor   = absor + wwmg(mm) * EXP(Dr_int1) * xpi * adust(mm)**2
             Dr_int2 = LOG(Qsca_Dr(il1,mm)) + (LOG(Qsca_Dr(il1+1,mm)) - LOG(Qsca_Dr(il1,mm))) * fact
             scater  = scater + wwmg(mm) * EXP(Dr_int2) * xpi * adust(mm)**2
          ENDDO
       ELSE
          fact = (wlth - wavgrd(il1)) / (wavgrd(il1+1) - wavgrd(il1))
          DO mm = 1, npg
             Dr_int1 = Qabs_Dr(il1,mm) + (Qabs_Dr(il1+1,mm) - Qabs_Dr(il1,mm)) * fact
             absor   = absor + wwmg(mm) * Dr_int1 * xpi * adust(mm)**2
             Dr_int2 = Qsca_Dr(il1,mm) + (Qsca_Dr(il1+1,mm) - Qsca_Dr(il1,mm)) * fact
             scater  = scater + wwmg(mm) * Dr_int2 * xpi * adust(mm)**2
          ENDDO
       ENDIF
       extinct = (absor + scater) * dens_o(0)

       ! -----------------------------------------
       ! Initial approximation of the radiation
       ! field and the escape probability
       ! -----------------------------------------
       DO ii = 0, npo
          intJm(ii,j) = (rf_incIm%Val(j) + I_incr_m(j)) * EXP(-extinct*d_cm(ii))
          intJp(ii,j) = (rf_incIp%Val(j) + I_incr_p(j)) * EXP(-extinct*(d_cm(npo) - d_cm(ii)))
          intJmp(ii,j) = (intJm(ii,j) + intJp(ii,j)) * 0.5_dp
       ENDDO
    ENDDO

    !--------------------------------------
    ! Variables used in PDR_DUSTEM - part 2
    !--------------------------------------

    IF (ALLOCATED(J_du_pro)) DEALLOCATE(J_du_pro)
    ALLOCATE (J_du_pro(0:npo,0:nwlg))
    J_du_pro(0:npo,0:nwlg) = intJmp(0:npo,0:nwlg)

    IF (F_DUST_P == 1) THEN
       IF (ALLOCATED(J_dust)) THEN
          DEALLOCATE(J_dust)
       ENDIF
       ALLOCATE (J_dust(nav_dust,0:nwlg))
       DO ii = 1, nav_dust
          J_dust(ii,:) = intJmp(iav_dust(ii),:)
       ENDDO
    ENDIF

    !---------------------------------------------------------------------
    ! Create RF_TYPE objects and tables used in FBTH (stv5c.f)
    !---------------------------------------------------------------------

    IF (ALLOCATED (si_)) THEN
       DEALLOCATE(si_)
    ENDIF
    ALLOCATE (si_(rf_wl%lower:rf_wl%upper))
    IF (ALLOCATED (si_QU)) THEN
       DEALLOCATE(si_QU)
    ENDIF
    ALLOCATE (si_QU(npg,rf_wl%lower:rf_wl%upper))
    IF (ALLOCATED (si_QUl)) THEN
       DEALLOCATE(si_QUl)
    ENDIF
    ALLOCATE (si_QUl(npg,rf_wl%lower:rf_wl%upper))
    IF (ALLOCATED (si_QUl2)) THEN
       DEALLOCATE(si_QUl2)
    ENDIF
    ALLOCATE (si_QUl2(npg,rf_wl%lower:rf_wl%upper))

    PRINT *, "##########   nwlg =", nwlg

END SUBROUTINE RF_INIT

!******************
SUBROUTINE RF_BOUND
!******************

    USE PXDR_FLAGS
    USE PXDR_CONSTANTES
    USE PXDR_CHEM_DATA
    USE PXDR_PROFIL
    USE PXDR_AUXILIAR
    USE PXDR_STR_DATA
    USE PXDR_DRAINE_DUST
    USE PXDR_RF_TOOLS
    USE PXDR_TRANSFER

    IMPLICIT NONE

    REAL (KIND=dp)                             :: wl_cut_D =1.0e5_dp   ! Draine field cutoff wavelength (extended)
    INTEGER                                    :: I_cut_D              ! Draine field cutoff position in grid
    REAL (KIND=dp)                             :: wl_cut_V = 8000.0_dp ! Visible cutoff wavelength
    INTEGER                                    :: I_cut_V              ! Index in wavelength Grid of visible cutoff
    INTEGER                                    :: I_cut_L              ! Index in wavelength Grid of Lyman cutoff
    INTEGER                                    :: ii, j, k
    TYPE (RF_TYPE)                             :: rf_tmp, rf_lamI

    REAL(KIND=dp), DIMENSION (:), ALLOCATABLE  :: perp_RF_m, perp_RF_p
    REAL(KIND=dp)                              :: wlth
    INTEGER                                    :: i1, i2, i_l, iwlg

    ! Declarations for intergalactic radiation field -------------------------------------------------------------
    ! Note : To use an intergalactic radiation field, the user has to provide the 3 following parameters
    !        Intergalactic spectra are READ from haardt96.dat in the data directory
    !        To not use an intergalactic radiation field, program has to be compiled with F_haardt96 = 0
    ! Other quantities used to initialize the intergalactic spectrum
    INTEGER                                    :: nredshift           ! number of redshifts with intergalactic spectra provided
    INTEGER, PARAMETER                         :: nlignmax_haardt96=1000 ! Max number of wavelength in haardt96.dat
    INTEGER                                    :: nlign_haardt96      ! number of wavelength in haardt96.dat
    REAL (KIND=dp), ALLOCATABLE                :: tab_redshift(:)     ! table of known redshifts
    REAL (KIND=dp), ALLOCATABLE                :: lam_haardt(:)       ! table of the wavelenght in Ang. in datafile
    REAL (KIND=dp), ALLOCATABLE                :: I_haardt(:,:)       ! spectra in function of wavelength and redshift
                                                                      ! READ in haardt96.dat in erg s-1 cm-2 A-1 str-1
    REAL (KIND=dp), ALLOCATABLE                :: I_intergal(:)       ! spectrum at the proper redshift
    INTEGER                                    :: num_redshift        ! number of the CLOSEst redshift in data to the one searched
    REAL (KIND=dp)                             :: fact                ! factor used in linear interpolation

    TYPE (RF_TYPE)                             :: rf_incI_c1          ! ISRF component : VUV and UV (scaled by radm and radp)
    TYPE (RF_TYPE)                             :: rf_incI_c2          ! ISRF component : Old stars  (unscaled)
    TYPE (RF_TYPE)                             :: rf_incI_c3          ! ISRF component : CMB        (scaled by redshift)
    TYPE (RF_TYPE)                             :: rf_incI_c4          ! ISRF component : Warm dust and PAHs (unscaled)

    ! Variables pour ajouter le champs IR cree par DUSTEM.
    REAL (KIND=dp)                              :: a_IR        ! Coefficients to fit the amount of pah
    REAL (KIND=dp)                              :: b_IR        ! Coefficients to fit the amount of vsg
    REAL (KIND=dp)                              :: c_IR        ! Coefficients to fit the amount of bg
    REAL (KIND=dp)                              :: NH_IR       ! Column density to fit the IR data.
    REAL (KIND=dp)                              :: pah_IR      ! 4*pi*I_lambda des PAH (erg s-1 H-1 A-1 sr-1)
    REAL (KIND=dp)                              :: vsg_IR      ! 4*pi*I_lambda des VSG (erg s-1 H-1 A-1 sr-1)
    REAL (KIND=dp)                              :: bg_IR       ! 4*pi*I_lambda des BG  (erg s-1 H-1 A-1 sr-1)
    REAL (KIND=dp), ALLOCATABLE, DIMENSION(:)   :: wl_ir_cpn   ! Longueurs d'onde pour lesquels on a des donnees (A)
    REAL (KIND=dp), ALLOCATABLE, DIMENSION(:)   :: IR_ir_cpn   ! 4*pi*I_lambda totale (erg s-1 H-1 A-1 sr-1)
    INTEGER, PARAMETER                          :: nwlg_ir_cpn = 1500 ! Nombre de points en longueur d'onde
    INTEGER                                     :: j_IR        ! Integer pour faire les boucles.

    ! These constants fit the mean intensity from the outside of the clump
    ! in the case of a perpendicular contribution
    ! Computed with L = 19
    ! Check that it is independant of L (JLB - 25 XII 2013)
    REAL (KIND=dp)                              :: d_pp = 0.0088161_dp
    REAL (KIND=dp)                              :: c_pp = 0.358332_dp
    REAL (KIND=dp)                              :: x_pp = 4.07632_dp
    REAL (KIND=dp)                              :: bet_pp

    !===================================================================================
    ! CREATION OF THE INCIDENT RADIATION FIELD
    !===================================================================================
    ! The radiation field from Far-UV to submillimeter is build by 4 components
    ! - component 1 : Far UV
    ! - component 2 : Near UV, Visible, near I.R.
    ! - component 3 : CMB
    ! - component 4 : I.R.
    ! Draine Formula: Specific intensity
    ! Radiation field in erg cm-2 s-1 A-1 sterad-1
    !===================================================================================
    IF (ASSOCIATED(rf_incI_c1%Val)) THEN
       CALL RF_FREE (rf_incI_c1)
    ENDIF
    CALL RF_CREATE (rf_incI_c1, 0, nwlg)
    IF (ASSOCIATED(rf_incI_c2%Val)) THEN
       CALL RF_FREE (rf_incI_c2)
    ENDIF
    CALL RF_CREATE (rf_incI_c2, 0, nwlg)
    IF (ASSOCIATED(rf_incI_c3%Val)) THEN
       CALL RF_FREE (rf_incI_c3)
    ENDIF
    CALL RF_CREATE (rf_incI_c3, 0, nwlg)
    IF (ASSOCIATED(rf_incI_c4%Val)) THEN
       CALL RF_FREE (rf_incI_c4)
    ENDIF
    CALL RF_CREATE (rf_incI_c4, 0, nwlg)

    !--------------------------------------------------------------------------------
    ! Component 1 : Far-UV
    !--------------------------------------------------------------------------------
    CALL RF_LOCATE (rf_wl, wl_lym, I_cut_L)
    IF (F_ISRF == 1) THEN  ! Mathis
       ISRF_name = "Mathis et al. 1983 FUV radiation field (modified to extend it)"
       ! Far UV (fit JLB Jan 2008)
       CALL RF_LOCATE (rf_wl, wl_cut_V, I_cut_V)
       rf_incI_c1%Val(I_cut_L:i_cut_V) = (TANH(4.07e-3_dp*rf_wl%Val(I_cut_L:i_cut_V)-4.5991_dp) + 1.0_dp) &
                                       * 107.192_dp * rf_wl%Val(I_cut_L:i_cut_V)**(-2.89_dp)
       rf_incI_c1%Val(i_cut_V+1:nwlg)  = 2.0_dp * 107.192_dp * rf_wl%Val(i_cut_V+1:nwlg)**(-2.89_dp)
       rf_incI_c1%Ord = (/ (ii, ii=rf_incI_c1%lower,rf_incI_c1%upper) /)

    ELSE IF (F_ISRF == 2) THEN ! Draine

       ISRF_name = "Draine FUV radiation field "
       ! Draine's formula extended to 10000.0 A to give a smooth transition to Visible
       ! 2 Fev 2012 - JLB - Still not enough for Rad = 1.0e5 (required by Leiden Workshop)
       !                    Extend to 1.0e5 A = 10 micron
       CALL RF_LOCATE (rf_wl, wl_cut_D, I_cut_D)
       ! Draine 1978 radiation field
       ! Expression from M. Kopp PhD thesis page 284, from Sternberg and Dalgarno 1985.
       rf_incI_c1%Val(I_cut_L:I_cut_D) = 6.36E7_dp / rf_wl%Val(I_cut_L:I_cut_D)**4 &
                                       - 1.0237E11_dp / rf_wl%Val(I_cut_L:I_cut_D)**5 &
                                       + 4.0812E13_dp / rf_wl%Val(I_cut_L:I_cut_D)**6
       rf_incI_c1%Val = rf_incI_c1%Val / (4.0 * xpi) ! rf_incI is now in erg cm-2 s-1 str-1 A-1
       rf_incI_c1%Ord = (/ (ii, ii=rf_incI_c1%lower,rf_incI_c1%upper) /)
    ENDIF

    !--------------------------------------------------------------------------------
    ! Component 2 : Near-UV, visible, near I.R
    !--------------------------------------------------------------------------------
    ! 3 diluted Black Bodies (from Mathis, Mezger and Panagia 1983, A&A, 128, 212)
    ! Values revised by JLB to take account of far UV.
    !    6 IV 2010
    ! Fit redone by JLB to bring into accordance with DIRBE data provided by Francois Boulanger
    ! Note: still 3 BB component although the first 2 are almost degenerate. Suppress one?
    !--------------------------------------------------------------------------------
    DO ii = 0, nwlg
    !  rf_incI_c2%Val(ii) = 1.05e-14_dp * BBL(rf_wl%Val(ii),7127.0_dp) &
    !                     + 1.28e-13_dp * BBL(rf_wl%Val(ii),4043.0_dp) &
    !                     + 3.30e-13_dp * BBL(rf_wl%Val(ii),2930.0_dp)
       rf_incI_c2%Val(ii) = 2.07e-14_dp * BBL(rf_wl%Val(ii),6184.0_dp) &
                          + 1.09e-14_dp * BBL(rf_wl%Val(ii),6123.0_dp) &
                          + 1.52e-12_dp * BBL(rf_wl%Val(ii),2539.0_dp)
    ENDDO
    rf_incI_c2%Ord = (/ (ii, ii=rf_incI_c1%lower,rf_incI_c1%upper) /)

    !--------------------------------------------------------------------------------
    ! Component 3 : CMB
    !--------------------------------------------------------------------------------
    rf_incI_c3%Ord = (/ (ii, ii=rf_incI_c1%lower,rf_incI_c1%upper) /)
    DO ii = 0, nwlg
       rf_incI_c3%Val(ii) = BBL(rf_wl%Val(ii),tcmb)
    ENDDO

    !--------------------------------------------------------------------------------
    ! Component 4 : Dust contribution to the IR
    !--------------------------------------------------------------------------------
    ! Manuel Gonzalez-Garcia, 20 november 2008
    ! This component has been estimated thanks to the DUSTEM code
    !--------------------------------------------------------------------------------
    ! Coefficients to fit IR Interstellar field from DUSTEM code are:
    a_IR = 52.5593_dp
    b_IR = 75.0854_dp
    c_IR = 16.1213_dp
    NH_IR = 1.0e20_dp
    ! Allocate the variable IR, which will contain data of the IR interstellar field.
    ALLOCATE(IR_ir_cpn(nwlg_ir_cpn))
    ALLOCATE(wl_ir_cpn(nwlg_ir_cpn))
    OPEN(iwrtmp,file='./../data/Astrodata/IR_field_dustem.dat',status='old')
    DO ii = 1, nwlg_ir_cpn
       READ(iwrtmp,*) wl_ir_cpn(ii), pah_IR, vsg_IR, bg_IR
       IR_ir_cpn(ii) = (a_IR*pah_IR + b_IR*vsg_IR + c_IR*bg_IR) * NH_IR
    ENDDO
    CLOSE(iwrtmp)

    rf_incI_c4%Ord = (/ (ii, ii=rf_incI_c1%lower,rf_incI_c1%upper) /)
    DO ii = 0,nwlg
       ! 31 Oct 2011 - Change upper value
       ! IF (rf_wl%Val(ii) < 1.0e4_dp .OR. rf_wl%Val(ii) > 8.0e6_dp ) THEN
       IF (rf_wl%Val(ii) < 1.0e4_dp) THEN
          rf_incI_c4%Val(ii) = 0.0_dp
       ELSE
          DO j_IR = 1, nwlg_ir_cpn-1
             IF (rf_wl%Val(ii) >= wl_ir_cpn(j_IR) .AND. rf_wl%Val(ii) < wl_ir_cpn(j_IR+1)) THEN
                rf_incI_c4%Val(ii) = IR_ir_cpn(j_IR) &
                                   + (IR_ir_cpn(j_IR+1) - IR_ir_cpn(j_IR)) &
                                   * (rf_wl%Val(ii)-wl_ir_cpn(j_IR)) &
                                   / (wl_ir_cpn(j_IR+1)-wl_ir_cpn(j_IR))
             ENDIF
          ENDDO
       ENDIF
    ENDDO

    DEALLOCATE(IR_ir_cpn)
    DEALLOCATE(wl_ir_cpn)

    !===================================================================================
    ! BUILD FULL SPECTRUM (unscaled)
    !===================================================================================
    rf_incIm%Ord = rf_incI_c1%Ord
    rf_incIm%Val = rf_incI_c1%Val + rf_incI_c2%Val + rf_incI_c3%Val + rf_incI_c4%Val
    ! This file is written in the unit used by DUSTEM:
    !           x: Lambda in micrometer
    !           y: J_nu in erg s-1 cm-2 Hz-1 (note that integration on sterad is done)
    OPEN (iwrtmp,FILE=TRIM(dustem_dat)//'ISRF.DAT',STATUS="unknown")
    WRITE(iwrtmp,'("#",/,"#",/,"#",/,"#",/,"#",/,"#",/,i8)') nwlg
    DO ii = 0, nwlg
       WRITE (iwrtmp,*) 1.0e-4_dp*rf_wl%Val(ii), J_to_u*1.0e-8_dp*rf_incIm%Val(ii)*(rf_wl%Val(ii))**2
    ENDDO
    CLOSE (iwrtmp)

    !===================================================================================
    ! COMPUTE THE RADIATION FIELD INTEGRATED ENERGY IN FREE SPACE (TAKEN FROM STV2Z.F90)
    !===================================================================================

     CALL RF_CREATE(rf_lamI,0,nwlg)
     rf_lamI%Val = rf_wl%Val * rf_incIm%Val

    !--------------------------------------------------------------------------------
    ! Integration of UV radiation field from Lyman up to Habing limit
    ! (wl_lym = 911 Å ; wl_hab = 2400 Å) - BG 04/2013
    !     change to wl_6ev to compute up to 6 eV
    !--------------------------------------------------------------------------------
     CALL RF_CREATE(rf_tmp,1,2)
     CALL RF_LOCATE(rf_wl, wl_lym, i1)
     rf_tmp%Val(1) = 1.0_dp
     rf_tmp%Ord(1) = rf_wl%Ord(i1)
     CALL RF_LOCATE(rf_wl, wl_hab, i2)
     rf_tmp%Val(2) = 1.0_dp
     rf_tmp%Ord(2) = rf_wl%Ord(i2)
     int_rad0 = SECT_INT(rf_tmp, rf_incIm, wl_pointer, wl_hab) * J_to_u
     flu_rad0 = SECT_INT(rf_tmp, rf_lamI,  wl_pointer, wl_hab) * J_to_u / (1.0e8_dp * xhp)
     CALL RF_FREE (rf_lamI)
     CALL RF_FREE (rf_tmp)

     PRINT *, " "
     PRINT *, " Standard field in free space:"
     WRITE (iscrn, '("  U0_UV:",1pe12.4," erg cm-3,     ","from",0pf9.2," to",0pf9.2," Angstrom")') int_rad0,  wl_lym,  wl_hab
     WRITE (iscrn, '("  F0_UV:",1pe12.4," phot cm-2 s-1,","from",0pf9.2," to",0pf9.2," Angstrom")') flu_rad0,  wl_lym,  wl_hab
     WRITE (iwrit2,'("  U0_UV:",1pe12.4," erg cm-3,     ","from",0pf9.2," to",0pf9.2," Angstrom")') int_rad0,  wl_lym,  wl_hab
     WRITE (iwrit2,'("  F0_UV:",1pe12.4," phot cm-2 s-1,","from",0pf9.2," to",0pf9.2," Angstrom")') flu_rad0,  wl_lym,  wl_hab

    ! Now scale left and right side differently:
    rf_incIm%Val = radm_ini * rf_incI_c1%Val + rf_incI_c2%Val + rf_incI_c3%Val + rf_incI_c4%Val
    rf_incIp%Val = radp_ini * rf_incI_c1%Val + rf_incI_c2%Val + rf_incI_c3%Val + rf_incI_c4%Val
    ! 30 oct 2011: try enhancing also the IR field to pump CO
    !rf_incIm%Val = radm_ini * rf_incI_c1%Val + rf_incI_c2%Val + rf_incI_c3%Val + radm_ini * rf_incI_c4%Val
    !rf_incIp%Val = radp_ini * rf_incI_c1%Val + rf_incI_c2%Val + rf_incI_c3%Val + radp_ini * rf_incI_c4%Val

    ! 27 June 2011 write scaled radiation field
    ! This file is written in the unit used by DUSTEM:
    !           x: Lambda in micrometer
    !           y: J_nu in erg s-1 cm-2 Hz-1 (note that integration on sterad is done)
    OPEN (iwrtmp,FILE=TRIM(dustem_dat)//'ISRF.DAT',STATUS="unknown")
    WRITE(iwrtmp,'("#",/,"#",/,"#",/,"#",/,"#",/,"#",/,i8)') nwlg
    DO ii = 0, nwlg
       WRITE (iwrtmp,*) 1.0e-4_dp*rf_wl%Val(ii), J_to_u*1.0e-8_dp*rf_incIm%Val(ii)*(rf_wl%Val(ii))**2
    ENDDO
    CLOSE (iwrtmp)

    !===================================================================================
    ! NORMAL - EXTERNAL RADIATION FIELD
    ! An external radiation field, e.g. from a star, is added if d_sour .NE. 0 or srcpp is set to a flag
    ! d_sour > 0 : the star is on the back side of the cloud
    ! d_sour < 0 : the star is on the observer side of the cloud
    !===================================================================================

    ! Everywhere in this block: tag "m" for left side, "p" for right side
    IF (ALLOCATED(I_incr_m)) DEALLOCATE (I_incr_m)
    IF (ALLOCATED(I_incr_p)) DEALLOCATE (I_incr_p)
    ALLOCATE (I_incr_m(0:nwlg))
    ALLOCATE (I_incr_p(0:nwlg))
    I_incr_m = 0.0_dp
    I_incr_p = 0.0_dp
    IF (F_Norm_RF /= 0) THEN

       ALLOCATE (perp_RF_m(0:nwlg))
       ALLOCATE (perp_RF_p(0:nwlg))

       IF (srcpp(1:2) == 'F_') THEN
          i_l = 1
          DO iwlg = 0, nwlg
             wlth = rf_wl%Val(iwlg)
             CALL INTERP_X (wlth, perp_RF_m(iwlg), i_l)
          ENDDO
          perp_RF_p = perp_RF_m
       ELSE IF (F_RF_PP == 1) THEN
          DO iwlg = 0, nwlg
             perp_RF_m(iwlg) = rf_incIm%Val(iwlg)
             perp_RF_p(iwlg) = rf_incIp%Val(iwlg)
          ENDDO
       ELSE
          DO iwlg = 0, nwlg
             perp_RF_m(iwlg) = BBL(rf_wl%Val(iwlg),teff)
          ENDDO
          perp_RF_p = perp_RF_m
       ENDIF

       ! Add here computation of I_incr_m and I_incr_p from note
       !------------------------------------------------------------------------
       ! Perpendicalar RF on the observer side of the cloud (F_Norm_RF == 1)
       !------------------------------------------------------------------------
       IF (F_Norm_RF == 1) THEN
          IF (F_RF_PP == 0) THEN
             DO iwlg = 0, nwlg
                IF (rf_incIm%Val(iwlg) /= 0.0_dp) THEN
                   bet_pp = LOG(0.5_dp * ((1.0_dp - geom) + geom * perp_RF_m(iwlg) / rf_incIm%Val(iwlg)))
                   IF (bet_pp > LOG(0.5_dp)) THEN
                      I_incr_m(iwlg) = rf_incIm%Val(iwlg) &
                                     * EXP(0.5_dp * (x_pp*x_pp + 4.0_dp*c_pp - &
                                       (x_pp - LOG(0.25_dp) + 2.0_dp*bet_pp)**2 ) &
                                     / (LOG(0.25_dp) - 2.0_dp * bet_pp))
                   ENDIF
                ENDIF
             ENDDO
          ELSE IF (F_RF_PP == 1) THEN !--- Case of a beamed ISRF
             DO iwlg = 0, nwlg
                I_incr_m(iwlg) = perp_RF_m(iwlg) * 0.5_dp * geom / d_pp
             ENDDO
          ENDIF
       !------------------------------------------------------------------------
       ! Perpendicalar RF on the back side of the cloud (F_Norm_RF == 2)
       !------------------------------------------------------------------------
       ELSE IF (F_Norm_RF == 2) THEN
          IF (F_RF_PP == 0) THEN !-- Star
             DO iwlg = 0, nwlg
                IF (rf_incIp%Val(iwlg) /= 0.0_dp) THEN
                   bet_pp = LOG(0.5_dp * ((1.0_dp - geom) + geom * perp_RF_p(iwlg) / rf_incIp%Val(iwlg)))
                   IF (bet_pp > LOG(0.5_dp)) THEN
                      I_incr_p(iwlg) = rf_incIp%Val(iwlg) &
                                     * EXP(0.5_dp * (x_pp*x_pp + 4.0_dp*c_pp - &
                                       (x_pp - LOG(0.25_dp) + 2.0_dp*bet_pp)**2 ) &
                                     / (LOG(0.25_dp) - 2.0_dp * bet_pp))
                   ENDIF
                ENDIF
             ENDDO
          ELSE IF (F_RF_PP == 1) THEN !--- Case of a beamed ISRF
             DO iwlg = 0, nwlg
                I_incr_p(iwlg) = perp_RF_p(iwlg) * 0.5_dp * geom / d_pp
             ENDDO
          ENDIF
       ENDIF

       !------------------------------------------------------------------------
       ! Switch off isotropic contribution for beamed ISRF
       !------------------------------------------------------------------------
       IF (F_RF_PP == 1) THEN
          DO iwlg = 0, nwlg
             rf_incIm%Val(iwlg) = 0.0_dp
             rf_incIp%Val(iwlg) = 0.0_dp
          ENDDO
       ENDIF
    ENDIF

    !===================================================================================
    ! INTERGALACTIC RADIATION FIELD (Haardt & Madau 1996, ApJ, 461, 20)
    !===================================================================================

    IF (F_haardt96 == 1) THEN

       IF (F_haardt96_only == 1) THEN
          ! On ne prend en compte que Haardt & Madau 96
          ! aussi on commence par "effacer" Draine
          rf_incIm%Val(0:nwlg) = 0.0_dp
          rf_incIp%Val(0:nwlg) = 0.0_dp
          ! Sinon, le champ de intergalactique viendra s'ajouter a Draine
       ENDIF

       fichier = TRIM(data_dir)//TRIM("Astrodata/haardt96.dat")
       WRITE(iscrn,"('Opening file :',A30)") TRIM(fichier)
       OPEN (iwrtmp, file = fichier, status="old", action="READ")
       READ(iwrtmp,*)           ! Comment line
       READ(iwrtmp,*)           ! Comment line
       READ(iwrtmp,*)           ! Comment line
       READ(iwrtmp,*)           ! Comment line
       READ(iwrtmp,*)           ! Comment line
       READ(iwrtmp,*) nredshift ! Number of redshifts in table
       ! Allocation de memoire pour stocker les donnees de la table
       ALLOCATE(tab_redshift(nredshift))
       ALLOCATE (lam_haardt(nlignmax_haardt96))
       ALLOCATE (I_haardt(nlignmax_haardt96,nredshift))
       ALLOCATE(I_intergal(nlignmax_haardt96))
       ! Cherche la liste des redshifts disponibles dans la table
       READ(iwrtmp,*) (tab_redshift(ii), ii=1, nredshift)

       !Lecture des donnees
       DO ii = 1, nlignmax_haardt96
          READ(iwrtmp,*,END=110) lam_haardt(ii), (I_haardt(ii,k), k=1, nredshift)
!          PRINT *,ii,lam_haardt(ii),I_haardt(ii,26)
       ENDDO
110    CONTINUE
       nlign_haardt96 = ii - 1 ! number of wavelength in haardt96.dat
       CLOSE(iwrtmp)

       ! Search in haardt96.dat for an interval of redshift containing the whished redshift
       ! If searched redshift outside of data :
       IF (redshift < tab_redshift(1)) THEN
          PRINT *,"Redshift too small compared to data"
          PRINT *,"Switch to smallest redshift in data :", tab_redshift(1)
          redshift = tab_redshift(1)
       ENDIF
       IF (redshift > tab_redshift(nredshift)) THEN
          PRINT *,"Redshift too big compared to data"
          PRINT *,"Switch to highest redshift in data :",tab_redshift(nredshift)
          redshift = tab_redshift(nredshift)
       ENDIF
       ! Searched redshift inside data range
       ! On traite les cas particuliers :
       ! Si le redshift cherche est le premier de la liste
       IF (redshift == tab_redshift(1)) THEN
          I_intergal(:) = I_haardt(:,1)
       ENDIF
       ! Si le redshift cherche est le dernier de la liste
       IF (redshift == tab_redshift(nredshift)) THEN
          I_intergal(:) = I_haardt(:,nredshift)
       ENDIF
       ! Cas general
       IF ( tab_redshift(1) < redshift .AND. redshift < tab_redshift(nredshift)) THEN
          num_redshift = 1
          DO WHILE (redshift > tab_redshift(num_redshift))
             num_redshift = num_redshift + 1
             PRINT *,num_redshift, redshift, tab_redshift(num_redshift), tab_redshift(num_redshift+1)
          ENDDO
          ! Interpolation lineaire du redshift
          IF (redshift == tab_redshift(num_redshift-1)) THEN
             I_intergal(:) = I_haardt(:,num_redshift-1)
             PRINT *,"Exact redshift found in data :",num_redshift,redshift, tab_redshift(num_redshift-1)
          ELSE IF (redshift == tab_redshift(num_redshift)) THEN
             I_intergal(:) = I_haardt(:,num_redshift)
             PRINT *,"Exact redshift found in data :",num_redshift,redshift, tab_redshift(num_redshift)
          ELSE IF ((tab_redshift(num_redshift-1) < redshift) .AND. (redshift < tab_redshift(num_redshift))) THEN
             fact = (redshift - tab_redshift(num_redshift-1)) / (tab_redshift(num_redshift) - tab_redshift(num_redshift-1))
             I_intergal(:) = I_haardt(:,num_redshift-1) * (1.0_dp-fact) + I_haardt(:,num_redshift) * fact
             PRINT *,"Interpolation of intergal. spect. between z :",tab_redshift(num_redshift-1), tab_redshift(num_redshift)
          ENDIF
       ENDIF

       ! Remplissage du tableau des intensites avec interpolation lineaire aux lam. necessaires
       k = 0
       DO WHILE (rf_wl%Val(k) < lam_haardt(1))
          PRINT *,"Warning : Data for intergal. spectrum do not begin at a small enough lambda"
          PRINT *,"          Extrapolation"
          rf_incIm%Val(k) = rf_incIm%Val(k) + I_intergal(1)
          rf_incIp%Val(k) = rf_incIp%Val(k) + I_intergal(1)
          k = k + 1
       ENDDO
       j = 1
       DO ii = k, nwlg
          DO WHILE (rf_wl%Val(ii) > lam_haardt(j))
             j = j + 1
          ENDDO
          IF (lam_haardt(j) == rf_wl%Val(ii)) THEN
             rf_incIm%Val(ii) = rf_incIm%Val(ii) + I_intergal(j)
             rf_incIp%Val(ii) = rf_incIp%Val(ii) + I_intergal(j)
          ELSE IF ((lam_haardt(j-1) < rf_wl%Val(ii)) .AND. (rf_wl%Val(ii) < lam_haardt(j))) THEN
             ! On interpole en lambda dans la table haardt.dat
             ! interpolation pour le premier redshift
             fact = (rf_wl%Val(ii) - lam_haardt(j-1)) / (lam_haardt(j) - lam_haardt(j-1))
             rf_incIm%Val(ii) = rf_incIm%Val(ii) + I_intergal(j-1) * (1.0_dp - fact) &
                                                 + fact * I_intergal(j)
             rf_incIp%Val(ii) = rf_incIp%Val(ii) + I_intergal(j-1) * (1.0_dp - fact) &
                                                 + fact * I_intergal(j)
          ELSE IF (rf_wl%Val(ii) > lam_haardt(nlign_haardt96)) THEN
             rf_incIm%Val(ii) = rf_incIm%Val(ii) + I_intergal(nlign_haardt96)
             rf_incIp%Val(ii) = rf_incIp%Val(ii) + I_intergal(nlign_haardt96)
             PRINT *,"WARNING : Not enough point in lambda in haardt96.dat"
             PRINT *,"          Lambda searched :", rf_wl%Val(ii)
          ENDIF
       ENDDO

       ! Remarque : Pas besoin de toucher a rf_incIm%Ord. Il a ete correctement construit lors
       !            du remplissage du champ de Draine.

       DEALLOCATE(tab_redshift)
       DEALLOCATE(lam_haardt)
       DEALLOCATE(I_haardt)
       DEALLOCATE(I_intergal)

    ENDIF

!   VERIFICATION DU CHAMP
!     OPEN (iwrtmp,file="toto.dat",status="unknown",action="WRITE")
!     DO ii = 0,nwlg
!        WRITE(iwrtmp,*) rf_wl%Val(ii), rf_incIm%Val(ii)
!     ENDDO
!     CLOSE(iwrtmp)
!     PRINT *,"Fin d'ecriture du champ initial dans toto.dat"
!     STOP

    CALL RF_FREE (rf_incI_c1)
    CALL RF_FREE (rf_incI_c2)
    CALL RF_FREE (rf_incI_c3)
    CALL RF_FREE (rf_incI_c4)
    IF (ALLOCATED(perp_RF_m)) THEN
       DEALLOCATE (perp_RF_m)
       DEALLOCATE (perp_RF_p)
    ENDIF

END SUBROUTINE RF_BOUND

!**********************************************************************************************************
SUBROUTINE ADD_WL_LINES
!**********************************************************************************************************
! purpose : Add lines profiles wavelength after a first approximation of all quantities has been computed

    USE PXDR_FLAGS
    USE PXDR_CONSTANTES
    USE PXDR_CHEM_DATA
    USE PXDR_PROFIL
    USE PXDR_AUXILIAR
    USE PXDR_STR_DATA
    USE PXDR_DRAINE_DUST
    USE PXDR_RF_TOOLS
    USE PXDR_READCHIM
    USE PXDR_DUSTEM

    IMPLICIT NONE

    INTEGER                                    :: I_cut_L              !  Index of Lyman limit
    REAL (KIND=dp)                             :: wl_cut_F             !  Fitzpatrick cutoff wavelength (Ang)
    INTEGER                                    :: I_cut_F              !  Index of Fitzpatrick cutoff position in grid
    REAL (KIND=dp)                             :: wl_cut_F2            !  WL cutoff below which grains cross section is constant
    INTEGER                                    :: I_cut_F2             !  Index of the wavelength defined above
    REAL (KIND=dp)                             :: wl_cut               !  Dummy WL cutoff
    INTEGER                                    :: I_cut                !  Index of dummy cutoff position in grid

    INTEGER                                    :: i, j, il1, ii, lev
    INTEGER                                    :: add_1, iloc, iord    !  Used by RF_EXTAND
    REAL (KIND=dp)                             :: wl, fact
    TYPE (RF_TYPE)                             :: rf_tmp

    REAL(KIND=dp)                              :: wl0, wl1, dopl, dwl, wl0cm
    INTEGER                                    :: njl, nvl

    ! Auxiliary variables for DUSTEM field

    REAL(KIND=dp)                              :: N_crit_L, N_crit_D, N_crit, N_spec
    REAL(KIND=dp)                              :: tau_crit
    REAL(KIND=dp)                              :: Tg_max, dop_wid, vt2
    REAL(KIND=dp)                              :: Del_nu_L, Del_nu_D, Del_nu
    REAL(KIND=dp)                              :: Del_lam, wl_max
    REAL(KIND=dp)                              :: fosc, dvie

    REAL (KIND=dp),ALLOCATABLE, DIMENSION(:,:) :: J_du_tmp
    REAL (KIND=dp),ALLOCATABLE, DIMENSION(:,:) :: J_tmp

    !---------------------------------------------------------------------------
    ! SAVE THE RADIATION FIELDS ON PREVIOUS RF_WL GRID
    !---------------------------------------------------------------------------
    ALLOCATE (J_du_tmp(0:npo,0:nwlg))
    J_du_tmp(0:npo,0:nwlg) = J_du_pro(0:npo,0:nwlg)
    IF (F_DUST_P == 1) THEN
       ALLOCATE (J_tmp(nav_dust,0:nwlg))
       J_tmp(1:nav_dust,0:nwlg) = J_dust(1:nav_dust,0:nwlg)
    ENDIF

    !---------------------------------------------------------------------------
    ! ADD LINE WAVELENGTH IN THE RF_WL GRID
    !---------------------------------------------------------------------------

    ! Save previous wavelength grid
    IF (ALLOCATED (tmp_wl)) DEALLOCATE (tmp_wl)
    ALLOCATE (tmp_wl(0:nwlg))
    tmp_wl(0:nwlg) = rf_wl%Val(0:nwlg)

    ! Add lines only up to tau_crit
    tau_crit = 1.0e-3_dp
    Tg_max = MAXVAL(tgaz_o)
    vt2 = vturb * vturb
    wl_max = rf_wl%Val(nwlg)
    print *, " +++ Entering ADD_WL_LINES"
    print *, " Tg_max = ", Tg_max, " wl_max = ", wl_max
    print *, " nwlg = ", nwlg

    !--- Atomic hydrogen (Lyman lines) -----------------------------------------
    !     PARAMETER "dopl" is the step in lambda mesured in Doppler width
    IF (jlyman == 1) THEN
       N_spec = codesp(i_h,npo)
       DO ii = 1, nraih
          fosc = praih(ii,1)
          wl0  = praih(ii,2)
          dvie = praih(ii,3) / (4.0_dp * xpi)
          dopl = dopl_scal * wl0 * vturb / clum
          CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
          nwlg = nwlg + add_1

          wl0cm = wl0 * 1.0e-8_dp
          dop_wid = SQRT(vt2 + Tg_max * deksamu) / wl0cm
          N_crit_L = tau_crit * xpi * dvie * mecspie2 / fosc
          N_crit_D = tau_crit * mecspie2 * sqpi * dop_wid / fosc
          N_crit = MIN(N_crit_L, N_crit_D)

          IF (N_spec < N_crit) CYCLE
          iraih(ii,5) = 1
          Del_nu_D = 0.0_dp
          Del_nu_L = 0.0_dp
          IF (N_spec > N_crit_L) THEN
             Del_nu_L = SQRT(N_spec * fosc * dvie /(tau_crit * mecspie2 * xpi) - dvie*dvie)
          ENDIF
          IF (N_spec > N_crit_D) THEN
             Del_nu_D = dop_wid * SQRT(-LOG(tau_crit * mecspie2 * sqpi * dop_wid / (N_spec * fosc)))
          ENDIF
          Del_nu = MAX(Del_nu_L, Del_nu_D)
          Del_lam = 1.0e8_dp * Del_nu * wl0cm*wl0cm / clum

          dwl = 0.2_dp * dopl
          wl = wl0
          DO
             IF ((wl-wl0) < Del_lam .AND. wl < wl_max) THEN
                wl1 = rf_wl%Val(iloc+1)
                IF ((wl1-wl) > 2.0_dp*dwl) THEN
                   wl = wl + dwl
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                ELSE IF ((wl1-wl) > dwl) THEN
                   wl = wl + (wl1-wl) * 0.5_dp
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                   wl = wl1
                   iloc = iloc + 1
                ELSE
                   wl = wl1
                   iloc = iloc + 1
                ENDIF
                dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl-wl0)
             ELSE
                EXIT
             ENDIF
          ENDDO
          praih(ii,6) = rf_wl%Val(iloc)

          CALL RF_LOCATE (rf_wl, wl0, iloc)

          dwl = 0.2_dp * dopl
          wl = wl0
          DO
             IF ((wl0-wl) < Del_lam .AND. wl > wlth0) THEN
                wl1 = rf_wl%Val(iloc-1)
                IF ((wl-wl1) > 2.0_dp*dwl) THEN
                   wl = wl - dwl
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                ELSE IF ((wl-wl1) > dwl) THEN
                   wl = wl - (wl-wl1) * 0.5_dp
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                   wl = wl1
                   iloc = iloc - 1
                ELSE
                   wl = wl1
                   iloc = iloc - 1
                ENDIF
                dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl0-wl)
             ELSE
                EXIT
             ENDIF
          ENDDO
          praih(ii,5) = rf_wl%Val(iloc)

       ENDDO
    ENDIF
    print *, " ++ H done"

    !--- H2 lines, only lowest levels ---------------------------------

    DO ii = 1, nrh2bt
       nvl = irah2b(ii,1)
       njl = irah2b(ii,2)
       lev = lev_h2(nvl,njl)
       IF (lev >= lfgkh2) CYCLE

       lev = lev_h2(nvl,njl)
       N_spec = codesph2(lev,npo)
       fosc = prah2b(ii,1)
       wl0  = prah2b(ii,2)
       IF (wl0 < wl_lym) CYCLE
       dvie = prah2b(ii,3) / (4.0_dp * xpi)
       dopl = dopl_scal * wl0 * vturb / clum
       CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1

       wl0cm = wl0 * 1.0e-8_dp
       dop_wid = SQRT(vt2 + Tg_max * deksamu / spec_lin(in_sp(i_h2))%mom) / wl0cm
       N_crit_L = tau_crit * xpi * dvie * mecspie2 / fosc
       N_crit_D = tau_crit * mecspie2 * sqpi * dop_wid / fosc
       N_crit = MIN(N_crit_L, N_crit_D)

       IF (N_spec < N_crit) CYCLE
       irah2b(ii,5) = 1
       Del_nu_D = 0.0_dp
       Del_nu_L = 0.0_dp
       IF (N_spec > N_crit_L) THEN
          Del_nu_L = SQRT(N_spec * fosc * dvie /(tau_crit * mecspie2 * xpi) - dvie*dvie)
       ENDIF
       IF (N_spec > N_crit_D) THEN
          Del_nu_D = dop_wid * SQRT(-LOG(tau_crit * mecspie2 * sqpi * dop_wid / (N_spec * fosc)))
       ENDIF
       Del_nu = MAX(Del_nu_L, Del_nu_D)
       Del_lam = 1.0e8_dp * Del_nu * wl0cm*wl0cm / clum

       dwl = 0.2_dp * dopl
       wl = wl0
       DO
          IF ((wl-wl0) < Del_lam .AND. wl < wl_max) THEN
             wl1 = rf_wl%Val(iloc+1)
             IF ((wl1-wl) > 2.0_dp*dwl) THEN
                wl = wl + dwl
                CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
             ELSE IF ((wl1-wl) > dwl) THEN
                wl = wl + (wl1-wl) * 0.5_dp
                CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
                wl = wl1
                iloc = iloc + 1
             ELSE
                wl = wl1
                iloc = iloc + 1
             ENDIF
             dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl-wl0)
          ELSE
             EXIT
          ENDIF
       ENDDO
       prah2b(ii,6) = rf_wl%Val(iloc)

       CALL RF_LOCATE (rf_wl, wl0, iloc)

       dwl = 0.2_dp * dopl
       wl = wl0
       DO
          IF ((wl0-wl) < Del_lam .AND. wl > wlth0) THEN
             wl1 = rf_wl%Val(iloc-1)
             IF ((wl-wl1) > 2.0_dp*dwl) THEN
                wl = wl - dwl
                CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
             ELSE IF ((wl-wl1) > dwl) THEN
                wl = wl - (wl-wl1) * 0.5_dp
                CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
                wl = wl1
                iloc = iloc - 1
             ELSE
                wl = wl1
                iloc = iloc - 1
             ENDIF
             dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl0-wl)
          ELSE
             EXIT
          ENDIF
       ENDDO
       prah2b(ii,5) = rf_wl%Val(iloc)
    ENDDO
    print *, " ++ H2 B done"

    DO ii = 1, nrh2ct
       nvl = irah2c(ii,1)
       njl = irah2c(ii,2)
       lev = lev_h2(nvl,njl)
       IF (lev >= lfgkh2) CYCLE

       lev = lev_h2(nvl,njl)
       N_spec = codesph2(lev,npo)
       fosc = prah2c(ii,1)
       wl0  = prah2c(ii,2)
       IF (wl0 < wl_lym) CYCLE
       dvie = prah2c(ii,3) / (4.0_dp * xpi)
       dopl = dopl_scal * wl0 * vturb / clum
       CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
       nwlg = nwlg + add_1

       wl0cm = wl0 * 1.0e-8_dp
       dop_wid = SQRT(vt2 + Tg_max * deksamu / spec_lin(in_sp(i_h2))%mom) / wl0cm
       N_crit_L = tau_crit * xpi * dvie * mecspie2 / fosc
       N_crit_D = tau_crit * mecspie2 * sqpi * dop_wid / fosc
       N_crit = MIN(N_crit_L, N_crit_D)

       IF (N_spec < N_crit) CYCLE
       irah2c(ii,5) = 1
       Del_nu_D = 0.0_dp
       Del_nu_L = 0.0_dp
       IF (N_spec > N_crit_L) THEN
          Del_nu_L = SQRT(N_spec * fosc * dvie /(tau_crit * mecspie2 * xpi) - dvie*dvie)
       ENDIF
       IF (N_spec > N_crit_D) THEN
          Del_nu_D = dop_wid * SQRT(-LOG(tau_crit * mecspie2 * sqpi * dop_wid / (N_spec * fosc)))
       ENDIF
       Del_nu = MAX(Del_nu_L, Del_nu_D)
       Del_lam = 1.0e8_dp * Del_nu * wl0cm*wl0cm / clum

       dwl = 0.2_dp * dopl
       wl = wl0
       DO
          IF ((wl-wl0) < Del_lam .AND. wl < wl_max) THEN
             wl1 = rf_wl%Val(iloc+1)
             IF ((wl1-wl) > 2.0_dp*dwl) THEN
                wl = wl + dwl
                CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
             ELSE IF ((wl1-wl) > dwl) THEN
                wl = wl + (wl1-wl) * 0.5_dp
                CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
                wl = wl1
                iloc = iloc + 1
             ELSE
                wl = wl1
                iloc = iloc + 1
             ENDIF
             dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl-wl0)
          ELSE
             EXIT
          ENDIF
       ENDDO
       prah2c(ii,6) = rf_wl%Val(iloc)

       CALL RF_LOCATE (rf_wl, wl0, iloc)

       dwl = 0.2_dp * dopl
       wl = wl0
       DO
          IF ((wl0-wl) < Del_lam .AND. wl > wlth0) THEN
             wl1 = rf_wl%Val(iloc-1)
             IF ((wl-wl1) > 2.0_dp*dwl) THEN
                wl = wl - dwl
                CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
             ELSE IF ((wl-wl1) > dwl) THEN
                wl = wl - (wl-wl1) * 0.5_dp
                CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                nwlg = nwlg + add_1
                wl = wl1
                iloc = iloc - 1
             ELSE
                wl = wl1
                iloc = iloc - 1
             ENDIF
             dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl0-wl)
          ELSE
             EXIT
          ENDIF
       ENDDO
       prah2c(ii,5) = rf_wl%Val(iloc)
    ENDDO
    print *, " nwlg = ", nwlg
    print *, " ++ H2 C done"

    !--- HD lines, only lowest levels ---------------------------------

    IF (i_hd /= 0) THEN
       DO ii = 1, nrhdbt
          nvl = irahdb(ii,1)
          njl = irahdb(ii,2)
          lev = lev_hd(nvl,njl)
          IF (lev >= lfgkhd) CYCLE

          lev = lev_hd(nvl,njl)
          N_spec = codesphd(lev,npo)
          fosc = prahdb(ii,1)
          wl0  = prahdb(ii,2)
          IF (wl0 < wl_lym) CYCLE
          dvie = prahdb(ii,3) / (4.0_dp * xpi)
          dopl = dopl_scal * wl0 * vturb / clum
          CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
          nwlg = nwlg + add_1

          wl0cm = wl0 * 1.0e-8_dp
          dop_wid = SQRT(vt2 + Tg_max * deksamu / spec_lin(in_sp(i_hd))%mom) / wl0cm
          N_crit_L = tau_crit * xpi * dvie * mecspie2 / fosc
          N_crit_D = tau_crit * mecspie2 * sqpi * dop_wid / fosc
          N_crit = MIN(N_crit_L, N_crit_D)

          IF (N_spec < N_crit) CYCLE
          irahdb(ii,5) = 1
          Del_nu_D = 0.0_dp
          Del_nu_L = 0.0_dp
          IF (N_spec > N_crit_L) THEN
             Del_nu_L = SQRT(N_spec * fosc * dvie /(tau_crit * mecspie2 * xpi) - dvie*dvie)
          ENDIF
          IF (N_spec > N_crit_D) THEN
             Del_nu_D = dop_wid * SQRT(-LOG(tau_crit * mecspie2 * sqpi * dop_wid / (N_spec * fosc)))
          ENDIF
          Del_nu = MAX(Del_nu_L, Del_nu_D)
          Del_lam = 1.0e8_dp * Del_nu * wl0cm*wl0cm / clum

          dwl = 0.2_dp * dopl
          wl = wl0
          DO
             IF ((wl-wl0) < Del_lam .AND. wl < wl_max) THEN
                wl1 = rf_wl%Val(iloc+1)
                IF ((wl1-wl) > 2.0_dp*dwl) THEN
                   wl = wl + dwl
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                ELSE IF ((wl1-wl) > dwl) THEN
                   wl = wl + (wl1-wl) * 0.5_dp
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                   wl = wl1
                   iloc = iloc + 1
                ELSE
                   wl = wl1
                   iloc = iloc + 1
                ENDIF
                dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl-wl0)
             ELSE
                EXIT
             ENDIF
          ENDDO
          prahdb(ii,6) = rf_wl%Val(iloc)

          CALL RF_LOCATE (rf_wl, wl0, iloc)

          dwl = 0.2_dp * dopl
          wl = wl0
          DO
             IF ((wl0-wl) < Del_lam .AND. wl > wlth0) THEN
                wl1 = rf_wl%Val(iloc-1)
                IF ((wl-wl1) > 2.0_dp*dwl) THEN
                   wl = wl - dwl
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                ELSE IF ((wl-wl1) > dwl) THEN
                   wl = wl - (wl-wl1) * 0.5_dp
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                   wl = wl1
                   iloc = iloc - 1
                ELSE
                   wl = wl1
                   iloc = iloc - 1
                ENDIF
                dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl0-wl)
             ELSE
                EXIT
             ENDIF
          ENDDO
          prahdb(ii,5) = rf_wl%Val(iloc)
       ENDDO
       print *, " ++ HD B done"

       DO ii = 1, nrhdct
          nvl = irahdc(ii,1)
          njl = irahdc(ii,2)
          lev = lev_hd(nvl,njl)
          IF (lev >= lfgkhd) CYCLE

          lev = lev_hd(nvl,njl)
          N_spec = codesphd(lev,npo)
          fosc = prahdc(ii,1)
          wl0  = prahdc(ii,2)
          IF (wl0 < wl_lym) CYCLE
          dvie = prahdc(ii,3) / (4.0_dp * xpi)
          dopl = dopl_scal * wl0 * vturb / clum
          CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
          nwlg = nwlg + add_1

          wl0cm = wl0 * 1.0e-8_dp
          dop_wid = SQRT(vt2 + Tg_max * deksamu / spec_lin(in_sp(i_hd))%mom) / wl0cm
          N_crit_L = tau_crit * xpi * dvie * mecspie2 / fosc
          N_crit_D = tau_crit * mecspie2 * sqpi * dop_wid / fosc
          N_crit = MIN(N_crit_L, N_crit_D)

          IF (N_spec < N_crit) CYCLE
          irahdc(ii,5) = 1
          Del_nu_D = 0.0_dp
          Del_nu_L = 0.0_dp
          IF (N_spec > N_crit_L) THEN
             Del_nu_L = SQRT(N_spec * fosc * dvie /(tau_crit * mecspie2 * xpi) - dvie*dvie)
          ENDIF
          IF (N_spec > N_crit_D) THEN
             Del_nu_D = dop_wid * SQRT(-LOG(tau_crit * mecspie2 * sqpi * dop_wid / (N_spec * fosc)))
          ENDIF
          Del_nu = MAX(Del_nu_L, Del_nu_D)
          Del_lam = 1.0e8_dp * Del_nu * wl0cm*wl0cm / clum

          dwl = 0.2_dp * dopl
          wl = wl0
          DO
             IF ((wl-wl0) < Del_lam .AND. wl < wl_max) THEN
                wl1 = rf_wl%Val(iloc+1)
                IF ((wl1-wl) > 2.0_dp*dwl) THEN
                   wl = wl + dwl
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                ELSE IF ((wl1-wl) > dwl) THEN
                   wl = wl + (wl1-wl) * 0.5_dp
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                   wl = wl1
                   iloc = iloc + 1
                ELSE
                   wl = wl1
                   iloc = iloc + 1
                ENDIF
                dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl-wl0)
             ELSE
                EXIT
             ENDIF
          ENDDO
          prahdc(ii,6) = rf_wl%Val(iloc)

          CALL RF_LOCATE (rf_wl, wl0, iloc)

          dwl = 0.2_dp * dopl
          wl = wl0
          DO
             IF ((wl0-wl) < Del_lam .AND. wl > wlth0) THEN
                wl1 = rf_wl%Val(iloc-1)
                IF ((wl-wl1) > 2.0_dp*dwl) THEN
                   wl = wl - dwl
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                ELSE IF ((wl-wl1) > dwl) THEN
                   wl = wl - (wl-wl1) * 0.5_dp
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                   wl = wl1
                   iloc = iloc - 1
                ELSE
                   wl = wl1
                   iloc = iloc - 1
                ENDIF
                dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl0-wl)
             ELSE
                EXIT
             ENDIF
          ENDDO
          prahdc(ii,5) = rf_wl%Val(iloc)
       ENDDO
       print *, " ++ HD C done"
    ENDIF

    !---  CO lines, only lowest J --------------------------------

    IF (i_co /= 0) THEN
       DO ii = 1, nrco
          njl = iraco(ii,1)
          IF (njl >= jfgkco) THEN
             CYCLE
          ENDIF

          N_spec = codespco(njl+1,npo)
          fosc = praco(ii,1)
          wl0  = praco(ii,2)
          IF (wl0 < wl_lym) CYCLE
          dvie = praco(ii,3) / (4.0_dp * xpi)
          dopl = dopl_scal * wl0 * vturb / clum
          CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
          nwlg = nwlg + add_1

          wl0cm = wl0 * 1.0e-8_dp
          dop_wid = SQRT(vt2 + Tg_max * deksamu / spec_lin(in_sp(i_co))%mom) / wl0cm
          N_crit_L = tau_crit * xpi * dvie * mecspie2 / fosc
          N_crit_D = tau_crit * mecspie2 * sqpi * dop_wid / fosc
          N_crit = MIN(N_crit_L, N_crit_D)

          IF (N_spec < N_crit) CYCLE
          iraco(ii,5) = 1
          Del_nu_D = 0.0_dp
          Del_nu_L = 0.0_dp
          IF (N_spec > N_crit_L) THEN
             Del_nu_L = SQRT(N_spec * fosc * dvie /(tau_crit * mecspie2 * xpi) - dvie*dvie)
          ENDIF
          IF (N_spec > N_crit_D) THEN
             Del_nu_D = dop_wid * SQRT(-LOG(tau_crit * mecspie2 * sqpi * dop_wid / (N_spec * fosc)))
          ENDIF
          Del_nu = MAX(Del_nu_L, Del_nu_D)
          Del_lam = 1.0e8_dp * Del_nu * wl0cm*wl0cm / clum

          dwl = 0.2_dp * dopl
          wl = wl0
          DO
             IF (ABS(wl-wl0)/dopl < wlrwid) THEN
                wl1 = rf_wl%Val(iloc+1)
                IF ((wl1-wl) > 2.0_dp*dwl) THEN
                   wl = wl + dwl
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                ELSE IF ((wl1-wl) > dwl) THEN
                   wl = wl + (wl1-wl) * 0.5_dp
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                   wl = wl1
                   iloc = iloc + 1
                ELSE
                   wl = wl1
                   iloc = iloc + 1
                ENDIF
                dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl-wl0)
             ELSE
                EXIT
             ENDIF
          ENDDO
          praco(ii,6) = rf_wl%Val(iloc)

          CALL RF_LOCATE (rf_wl, wl0, iloc)

          dwl = 0.2_dp * dopl
          wl = wl0
          DO
             IF (ABS(wl-wl0)/dopl < wlrwid .AND. &
                wl > wlth0) THEN
                wl1 = rf_wl%Val(iloc-1)
                IF ((wl-wl1) > 2.0_dp*dwl) THEN
                   wl = wl - dwl
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                ELSE IF ((wl-wl1) > dwl) THEN
                   wl = wl - (wl-wl1) * 0.5_dp
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                   wl = wl1
                   iloc = iloc - 1
                ELSE
                   wl = wl1
                   iloc = iloc - 1
                ENDIF
                dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl0-wl)
             ELSE
                EXIT
             ENDIF
          ENDDO
          praco(ii,5) = rf_wl%Val(iloc)

       ENDDO
       print *, " ++ CO done"
    ENDIF

    !---  13CO lines, only lowest J -------------------------------

    IF (i_c13o /= 0) THEN
       DO ii = 1, nrc13o
          njl = irac13o(ii,1)
          IF (njl >= jfgkc13o) THEN
             CYCLE
          ENDIF

          N_spec = codespc13o(njl+1,npo)
          fosc = prac13o(ii,1)
          wl0  = prac13o(ii,2)
          IF (wl0 < wl_lym) CYCLE
          dvie = prac13o(ii,3) / (4.0_dp * xpi)
          dopl = dopl_scal * wl0 * vturb / clum
          CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
          nwlg = nwlg + add_1

          wl0cm = wl0 * 1.0e-8_dp
          dop_wid = SQRT(vt2 + Tg_max * deksamu / spec_lin(in_sp(i_c13o))%mom) / wl0cm
          N_crit_L = tau_crit * xpi * dvie * mecspie2 / fosc
          N_crit_D = tau_crit * mecspie2 * sqpi * dop_wid / fosc
          N_crit = MIN(N_crit_L, N_crit_D)

          IF (N_spec < N_crit) CYCLE
          irac13o(ii,5) = 1
          Del_nu_D = 0.0_dp
          Del_nu_L = 0.0_dp
          IF (N_spec > N_crit_L) THEN
             Del_nu_L = SQRT(N_spec * fosc * dvie /(tau_crit * mecspie2 * xpi) - dvie*dvie)
          ENDIF
          IF (N_spec > N_crit_D) THEN
             Del_nu_D = dop_wid * SQRT(-LOG(tau_crit * mecspie2 * sqpi * dop_wid / (N_spec * fosc)))
          ENDIF
          Del_nu = MAX(Del_nu_L, Del_nu_D)
          Del_lam = 1.0e8_dp * Del_nu * wl0cm*wl0cm / clum

          dwl = 0.2_dp * dopl
          wl = wl0
          DO
             IF (ABS(wl-wl0)/dopl < wlrwid) THEN
                wl1 = rf_wl%Val(iloc+1)
                IF ((wl1-wl) > 2.0_dp*dwl) THEN
                   wl = wl + dwl
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                ELSE IF ((wl1-wl) > dwl) THEN
                   wl = wl + (wl1-wl) * 0.5_dp
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                   wl = wl1
                   iloc = iloc + 1
                ELSE
                   wl = wl1
                   iloc = iloc + 1
                ENDIF
                dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl-wl0)
             ELSE
                EXIT
             ENDIF
          ENDDO
          prac13o(ii,6) = rf_wl%Val(iloc)

          CALL RF_LOCATE (rf_wl, wl0, iloc)

          dwl = 0.2_dp * dopl
          wl = wl0
          DO
             IF (ABS(wl-wl0)/dopl < wlrwid .AND. &
                wl > wlth0) THEN
                wl1 = rf_wl%Val(iloc-1)
                IF ((wl-wl1) > 2.0_dp*dwl) THEN
                   wl = wl - dwl
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                ELSE IF ((wl-wl1) > dwl) THEN
                   wl = wl - (wl-wl1) * 0.5_dp
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                   wl = wl1
                   iloc = iloc - 1
                ELSE
                   wl = wl1
                   iloc = iloc - 1
                ENDIF
                dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl0-wl)
             ELSE
                EXIT
             ENDIF
          ENDDO
          prac13o(ii,5) = rf_wl%Val(iloc)

       ENDDO
       print *, " ++ 13CO done"
    ENDIF

    !---  CO18 lines, only lowest J -------------------------------

    IF (i_co18 /= 0) THEN
       DO ii = 1, nrco18
          njl = iraco18(ii,1)
          IF (njl >= jfgkco18) THEN
             CYCLE
          ENDIF

          N_spec = codespco18(njl+1,npo)
          fosc = praco18(ii,1)
          wl0  = praco18(ii,2)
          IF (wl0 < wl_lym) CYCLE
          dvie = praco18(ii,3) / (4.0_dp * xpi)
          dopl = dopl_scal * wl0 * vturb / clum
          CALL RF_EXTAND (rf_wl, wl0, add_1, iloc, iord)
          nwlg = nwlg + add_1

          wl0cm = wl0 * 1.0e-8_dp
          dop_wid = SQRT(vt2 + Tg_max * deksamu / spec_lin(in_sp(i_co18))%mom) / wl0cm
          N_crit_L = tau_crit * xpi * dvie * mecspie2 / fosc
          N_crit_D = tau_crit * mecspie2 * sqpi * dop_wid / fosc
          N_crit = MIN(N_crit_L, N_crit_D)

          IF (N_spec < N_crit) CYCLE
          iraco18(ii,5) = 1
          Del_nu_D = 0.0_dp
          Del_nu_L = 0.0_dp
          IF (N_spec > N_crit_L) THEN
             Del_nu_L = SQRT(N_spec * fosc * dvie /(tau_crit * mecspie2 * xpi) - dvie*dvie)
          ENDIF
          IF (N_spec > N_crit_D) THEN
             Del_nu_D = dop_wid * SQRT(-LOG(tau_crit * mecspie2 * sqpi * dop_wid / (N_spec * fosc)))
          ENDIF
          Del_nu = MAX(Del_nu_L, Del_nu_D)
          Del_lam = 1.0e8_dp * Del_nu * wl0cm*wl0cm / clum

          dwl = 0.2_dp * dopl
          wl = wl0
          DO
             IF (ABS(wl-wl0)/dopl < wlrwid) THEN
                wl1 = rf_wl%Val(iloc+1)
                IF ((wl1-wl) > 2.0_dp*dwl) THEN
                   wl = wl + dwl
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                ELSE IF ((wl1-wl) > dwl) THEN
                   wl = wl + (wl1-wl) * 0.5_dp
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                   wl = wl1
                   iloc = iloc + 1
                ELSE
                   wl = wl1
                   iloc = iloc + 1
                ENDIF
                dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl-wl0)
             ELSE
                EXIT
             ENDIF
          ENDDO
          praco18(ii,6) = rf_wl%Val(iloc)

          CALL RF_LOCATE (rf_wl, wl0, iloc)

          dwl = 0.2_dp * dopl
          wl = wl0
          DO
             IF (ABS(wl-wl0)/dopl < wlrwid .AND. &
                wl > wlth0) THEN
                wl1 = rf_wl%Val(iloc-1)
                IF ((wl-wl1) > 2.0_dp*dwl) THEN
                   wl = wl - dwl
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                ELSE IF ((wl-wl1) > dwl) THEN
                   wl = wl - (wl-wl1) * 0.5_dp
                   CALL RF_EXTAND (rf_wl, wl, add_1, iloc, iord)
                   nwlg = nwlg + add_1
                   wl = wl1
                   iloc = iloc - 1
                ELSE
                   wl = wl1
                   iloc = iloc - 1
                ENDIF
                dwl = 0.2_dp * dopl + (wlrinc - 1.0_dp) * (wl0-wl)
             ELSE
                EXIT
             ENDIF
          ENDDO
          praco18(ii,5) = rf_wl%Val(iloc)

       ENDDO
       print *, " ++ CO18 done"
    ENDIF

    PRINT *, " 2nd pass:   nwlg =", nwlg, SIZE(rf_wl%Val)
    PRINT *, " Line profiles grid completed"

    ! Compute WL limits for all lines in full transfer
    ! Full data for lines computed are stored. They are used
    ! for information on any run
    fichier = TRIM(out_dir)//'full_lines.dat'
    OPEN (44, FILE=fichier, STATUS="unknown")
    DO ii = 1, nraih
       IF (iraih(ii,5) == 0) CYCLE
       CALL RF_LOCATE (rf_wl, praih(ii,5), iloc)
       iraih(ii,6) = iloc
       CALL RF_LOCATE (rf_wl, praih(ii,6), iloc)
       iraih(ii,7) = iloc
       WRITE (44,*) ii, iraih(ii,:), praih(ii,:)
       if (iraih(ii,6) == iraih(ii,7)) stop
    ENDDO
    WRITE (44,*) "           0"

    DO ii = 1, nrh2bt
       IF (irah2b(ii,5) == 0) CYCLE
       CALL RF_LOCATE (rf_wl, prah2b(ii,5), iloc)
       irah2b(ii,6) = iloc
       CALL RF_LOCATE (rf_wl, prah2b(ii,6), iloc)
       irah2b(ii,7) = iloc
       WRITE (44,*) ii, irah2b(ii,:), prah2b(ii,:)
       if (irah2b(ii,6) == irah2b(ii,7)) stop
    ENDDO
    WRITE (44,*) "           0"

    DO ii = 1, nrh2ct
       IF (irah2c(ii,5) == 0) CYCLE
       CALL RF_LOCATE (rf_wl, prah2c(ii,5), iloc)
       irah2c(ii,6) = iloc
       CALL RF_LOCATE (rf_wl, prah2c(ii,6), iloc)
       irah2c(ii,7) = iloc
       WRITE (44,*) ii, irah2c(ii,:), prah2c(ii,:)
       if (irah2c(ii,6) == irah2c(ii,7)) stop
    ENDDO
    WRITE (44,*) "           0"

    DO ii = 1, nrhdbt
       IF (irahdb(ii,5) == 0) CYCLE
       CALL RF_LOCATE (rf_wl, prahdb(ii,5), iloc)
       irahdb(ii,6) = iloc
       CALL RF_LOCATE (rf_wl, prahdb(ii,6), iloc)
       irahdb(ii,7) = iloc
       WRITE (44,*) ii, irahdb(ii,:), prahdb(ii,:)
       if (irahdb(ii,6) == irahdb(ii,7)) stop
    ENDDO
    WRITE (44,*) "           0"

    DO ii = 1, nrhdct
       IF (irahdc(ii,5) == 0) CYCLE
       CALL RF_LOCATE (rf_wl, prahdc(ii,5), iloc)
       irahdc(ii,6) = iloc
       CALL RF_LOCATE (rf_wl, prahdc(ii,6), iloc)
       irahdc(ii,7) = iloc
       WRITE (44,*) ii, irahdc(ii,:), prahdc(ii,:)
       if (irahdc(ii,6) == irahdc(ii,7)) stop
    ENDDO
    WRITE (44,*) "           0"

    IF (i_co /= 0) THEN
       DO ii = 1, nrco
          IF (iraco(ii,5) == 0) CYCLE
          CALL RF_LOCATE (rf_wl, praco(ii,5), iloc)
          iraco(ii,6) = iloc
          CALL RF_LOCATE (rf_wl, praco(ii,6), iloc)
          iraco(ii,7) = iloc
          WRITE (44,*) ii, iraco(ii,:), praco(ii,:)
          if (iraco(ii,6) == iraco(ii,7)) stop
       ENDDO
       WRITE (44,*) "           0"
    ENDIF

    IF (i_c13o /= 0) THEN
       DO ii = 1, nrc13o
          IF (irac13o(ii,5) == 0) CYCLE
          CALL RF_LOCATE (rf_wl, prac13o(ii,5), iloc)
          irac13o(ii,6) = iloc
          CALL RF_LOCATE (rf_wl, prac13o(ii,6), iloc)
          irac13o(ii,7) = iloc
          WRITE (44,*) ii, irac13o(ii,:), prac13o(ii,:)
          if (irac13o(ii,6) == irac13o(ii,7)) stop
       ENDDO
       WRITE (44,*) "           0"
    ENDIF

    IF (i_co18 /= 0) THEN
       DO ii = 1, nrco18
          IF (iraco18(ii,5) == 0) CYCLE
          CALL RF_LOCATE (rf_wl, praco18(ii,5), iloc)
          iraco18(ii,6) = iloc
          CALL RF_LOCATE (rf_wl, praco18(ii,6), iloc)
          iraco18(ii,7) = iloc
          WRITE (44,*) ii, iraco18(ii,:), praco18(ii,:)
          if (iraco18(ii,6) == iraco18(ii,7)) stop
       ENDDO
       WRITE (44,*) "           0"
    ENDIF
    CLOSE (44)
    CALL RF_REVERS (wl_pointer)

    !-------------------------------------------------------------------------------------
    !****    From here, the set of wavelength is complete and should not be modified ****
    !-------------------------------------------------------------------------------------
    ! All initialisations have to be redone with the new grid

    !-------------------------------------------
    ! Initialisation of external radiation Field
    !-------------------------------------------

    IF (ASSOCIATED(rf_incIm%Val)) THEN
       CALL RF_FREE(rf_incIm)
    ENDIF
    CALL RF_CREATE (rf_incIm, 0, nwlg)
    IF (ASSOCIATED(rf_incIp%Val)) THEN
       CALL RF_FREE(rf_incIp)
    ENDIF
    CALL RF_CREATE (rf_incIp, 0, nwlg)

    IF (ASSOCIATED(rf_u%Val)) THEN
       CALL RF_FREE(rf_u)
    ENDIF
    CALL RF_CREATE (rf_u, 0, nwlg)

    !----------------------------------------------------------------------------------
    ! Fitzpatrick and Massa extinction curves
    ! We use the Fitzpatrick and Massa expressions for the grains extinction
    ! Their expression gives E(lambda-V) / E(B-V)
    ! rf_expou%Val = [1 - 1/Rv * E(lambda-V) / E(B-V)]
    !
    ! Modification 24th june 2010
    ! If the name of the line of sight is sig1.9, we use :
    ! - from 912 to 1070 Ang (9.35 mum-1) : sigma_dust = 1.9E-21 (=> E(lambda-V) / E(B-V) = 8.8648)
    ! - above 1070 Ang                    : Galactic extinction curve
    !----------------------------------------------------------------------------------

    IF (ASSOCIATED(rf_extpou%Val)) THEN
       CALL RF_FREE(rf_extpou)
    ENDIF
    CALL RF_CREATE (rf_extpou, 0, nwlg)
    IF (ASSOCIATED(rf_tmp%Val)) THEN
       CALL RF_FREE(rf_tmp)
    ENDIF
    CALL RF_CREATE (rf_tmp, 0, nwlg)
    rf_tmp%Val = 1.0e4_dp / rf_wl%Val

    ! Create extinction curve from analytical fits
    ! 1 - UV slope and 2200 A bump
    !     lowest cut off at the Lyman limit (BG 04/2013)
    wl_cut = wl_lym
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut_L)
    wl_cut = FP_up_wl
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut)
    rf_extpou%Val(I_cut_L:I_cut) = cc1 + cc2 * rf_tmp%Val(I_cut_L:I_cut) &
                                 + cc3 * (rf_tmp%Val(I_cut_L:I_cut) * rf_tmp%Val(I_cut_L:I_cut) &
                                 / ((rf_tmp%Val(I_cut_L:I_cut) * rf_tmp%Val(I_cut_L:I_cut) - y0 * y0)**2 &
                                 + xgam * xgam * rf_tmp%Val(I_cut_L:I_cut) * rf_tmp%Val(I_cut_L:I_cut)))

    ! 2 - Far UV rise
    !     lowest cut off at the Lyman limit (BG 04/2013)
    wl_cut_F = 1.0e4_dp / 5.9_dp
    CALL RF_LOCATE (rf_wl, wl_cut_F, I_cut_F)
    rf_extpou%Val(I_cut_L:I_cut_F) = rf_extpou%Val(I_cut_L:I_cut_F) &
                                   + cc4 * (0.5392_dp * (rf_tmp%Val(I_cut_L:I_cut_F)-5.9_dp)**2 &
                                   + 0.0564_dp * (rf_tmp%Val(I_cut_L:I_cut_F)-5.9_dp)**3)

    ! 3 - IR analytical fit (using Galactic values from FP (2009) ApJ 699, 1209)
    !     Equation (5), used with alpha from BD+61 2365
    wl_cut = 1.0e4_dp
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut)
    wl_cut_F = 4.4e4_dp
    CALL RF_LOCATE (rf_wl, wl_cut_F, I_cut_F)
    rf_extpou%Val(I_cut:I_cut_F) = (0.349_dp + 2.087_dp * rrv) &
                                 / (1.0_dp + (1.0_dp / (0.507_dp*rf_tmp%Val(I_cut:I_cut_F))**2.28_dp)) - rrv

    ! 4 - IR rise at 18 microns as obtained with dustem and with Draine's data
    wl_cut = 4.4e4_dp
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut)
    wl_cut_F = 1.82e5_dp
    CALL RF_LOCATE (rf_wl, wl_cut_F, I_cut_F)
    rf_extpou%Val(I_cut:I_cut_F) = rf_extpou%Val(I_cut) + 0.015_dp * ( (1.0_dp / &
                                 & rf_tmp%Val(I_cut:I_cut_F) - wl_cut/1.0e4_dp) / 10.0_dp )**2.0

    ! 5 - IR bump at 10 microns as obtained with dustem and with Draine's data
    wl_cut = 1.0e4_dp
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut)
    rf_extpou%Val(I_cut:) = rf_extpou%Val(I_cut:) + 2.0_dp*rf_tmp%Val(I_cut:) &
                          * DEXP(-((DLOG(1.0_dp/rf_tmp%Val(I_cut:))-2.3_dp)/0.18_dp)**2.0_dp)

    ! 6 - FIR decreasing queue as obtained with dustem and with Draine's data
    wl_cut = 1.82e5_dp
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut)
    rf_extpou%Val(I_cut:) = (rf_extpou%Val(I_cut) + rrv) * (1.0_dp + rf_tmp%Val(I_cut)**(-1.75_dp)) / &
                          & (1.0_dp + rf_tmp%Val(I_cut:)**(-1.75_dp)) - rrv

    ! 7 - Visible: use 3 linear interpolation from 1 micron to V, from V to B and from B to UV cutoff (Ext = 0 at V and 1 at B)
    wl_cut = 1.0e4_dp
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut)
    wl_cut = V_band_wl
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut_F)
    rf_extpou%Val(I_cut_F:I_cut) = rf_extpou%Val(I_cut) - rf_extpou%Val(I_cut) &
                                 * (rf_tmp%Val(I_cut_F:I_cut) - 1.0_dp) &
                                 / ( 1.0e4_dp / V_band_wl -  1.0_dp)

    I_cut = I_cut_F
    wl_cut = B_band_wl
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut_F)
    rf_extpou%Val(I_cut_F:I_cut) = rf_extpou%Val(I_cut) + (1.0_dp - rf_extpou%Val(I_cut)) &
                                 * (rf_tmp%Val(I_cut_F:I_cut) - 1.0e4_dp / V_band_wl) &
                                 / ( 1.0e4_dp / B_band_wl -  1.0e4_dp / V_band_wl)

    I_cut = I_cut_F
    wl_cut = FP_up_wl
    CALL RF_LOCATE (rf_wl, wl_cut, I_cut_F)
    rf_extpou%Val(I_cut_F:I_cut) = rf_extpou%Val(I_cut) + (rf_extpou%Val(I_cut_F) - 1.0_dp) &
                                 * (rf_tmp%Val(I_cut_F:I_cut) - 1.0e4_dp / B_band_wl) &
                                 / ( 1.0e4_dp / FP_up_wl -  1.0e4_dp / B_band_wl)

    IF (los_ext == "sig1.9") THEN
       wl_cut_F2 = 1.0e4_dp / 9.346_dp
       CALL RF_LOCATE (rf_wl, wl_cut_F2, I_cut_F2)
       rf_extpou%Val(I_cut_L:I_cut_F2) = 8.8648_dp ! value to get grain abs. cross section = 1.9E-21 cm2
    ENDIF

    rf_extpou%Val = 1.0_dp + rf_extpou%Val / rrv
    rf_extpou%Ord = (/ (ii, ii=rf_extpou%lower,rf_extpou%upper) /)

    CALL RF_FREE (rf_tmp)

    !--------------------------------------
    ! Variables used in PDR_DUSTEM - part 1
    !--------------------------------------

    IF (F_DUST_P == 1) THEN
       ! Initialise points where DUSTEM is called
       print *, " nav_dec =", nav_dec
       IF (MOD(npo,nav_dec) == 0) THEN
          nav_dust = 1 + npo / nav_dec
       ELSE
          nav_dust = 2 + npo / nav_dec
       ENDIF
       IF (ALLOCATED(iav_dust)) THEN
          DEALLOCATE (iav_dust)
          DEALLOCATE (av_dust)
       ENDIF
       ALLOCATE (iav_dust(nav_dust))
       ALLOCATE (av_dust(nav_dust))
       DO ii = 1, nav_dust-1
          iav_dust(ii) = nav_dec * (ii-1)
          av_dust(ii) = tau_o(nav_dec * (ii-1))
       ENDDO
       iav_dust(nav_dust) = npo
       av_dust(nav_dust) = tau_o(npo)
    ENDIF

    !--------------------------------------------------------------------------
    ! Interpolate the radiation field on the new rf_wl grid
    !--------------------------------------------------------------------------

    IF (ALLOCATED(J_du_pro)) DEALLOCATE(J_du_pro)
    ALLOCATE (J_du_pro(0:npo,0:nwlg))
    IF (F_DUST_P == 1) THEN
       IF (ALLOCATED(J_dust)) DEALLOCATE(J_dust)
       ALLOCATE (J_dust(nav_dust,0:nwlg))
    ENDIF
    DO j = 0, nwlg
       wl = rf_wl%Val(j)
       DO i= 1, SIZE(tmp_wl)-1
          IF (tmp_wl(i) >= wl) THEN
             il1 = i-1
             EXIT
          ENDIF
       ENDDO
       fact = (wl - tmp_wl(il1)) / (tmp_wl(il1+1) - tmp_wl(il1))
       DO i = 0, npo
          J_du_pro(i,j) = J_du_tmp(i,il1) + (J_du_tmp(i,il1+1) - J_du_tmp(i,il1)) * fact
       ENDDO
       IF (F_DUST_P == 1) THEN
          DO i = 1, nav_dust
             J_dust(i,j) = J_tmp(i,il1) + (J_tmp(i,il1+1) - J_tmp(i,il1)) * fact
          ENDDO
       ENDIF
    ENDDO

    !--------------------------------------------------------------------------
    ! Compute Planck function over a fixed dust temperature grid
    !--------------------------------------------------------------------------
    DO i = 1, nTd
       IF(ASSOCIATED(rf_Plint(i)%Val)) CALL RF_FREE(rf_Plint(i))
       CALL RF_CREATE (rf_Plint(i), 0, nwlg)
       DO ii = 0, nwlg
          rf_Plint(i)%Val(ii) = BBL (rf_wl%Val(ii), Tdposs(i))
          rf_Plint(i)%Ord(ii) = rf_wl%Ord(ii)
       ENDDO
    ENDDO

    !----------------------------------------------------------------
    ! Set dust properties (absorption, scattering, emission, g factor
    !----------------------------------------------------------------

    PRINT *, " Enter DUST_PROPERTIES - 3"
    CALL DUST_PROPERTIES

    !----------------------------------------
    ! Set radiation field boundary conditions
    !----------------------------------------

    CALL RF_BOUND

    !---------------------------------------------------------------------
    ! Create RF_TYPE objects and tables used in FBTH (stv5c.f)
    !---------------------------------------------------------------------

    IF (ALLOCATED (si_)) THEN
       DEALLOCATE(si_)
    ENDIF
    ALLOCATE (si_(rf_wl%lower:rf_wl%upper))
    IF (ALLOCATED (si_QU)) THEN
       DEALLOCATE(si_QU)
    ENDIF
    ALLOCATE (si_QU(npg,rf_wl%lower:rf_wl%upper))
    IF (ALLOCATED (si_QUl)) THEN
       DEALLOCATE(si_QUl)
    ENDIF
    ALLOCATE (si_QUl(npg,rf_wl%lower:rf_wl%upper))
    IF (ALLOCATED (si_QUl2)) THEN
       DEALLOCATE(si_QUl2)
    ENDIF
    ALLOCATE (si_QUl2(npg,rf_wl%lower:rf_wl%upper))

    print *, " +++ End of ADD_WL_LINES"
    PRINT *, "##########   nwlg =", nwlg

    DEALLOCATE(J_du_tmp)
    IF (F_DUST_P == 1) DEALLOCATE(J_tmp)

END SUBROUTINE ADD_WL_LINES

END MODULE PXDR_RF_INIT
